package gov.va.med.fee.tool;

import gov.va.med.domain.fee.AppUser;
import gov.va.med.domain.fee.ContactInfo;
import gov.va.med.domain.fee.VaFacility;
import gov.va.med.domain.fee.VaFacilityType;
import gov.va.med.domain.fee.Visn;
import gov.va.med.domain.fee.ZipCode;
import gov.va.med.fee.FeeMessages;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.hibernate.validator.Min;
import org.hibernate.validator.NotNull;
import org.jboss.seam.ScopeType;
import org.jboss.seam.annotations.In;
import org.jboss.seam.annotations.Name;
import org.jboss.seam.annotations.Transactional;
import org.jboss.seam.annotations.security.Restrict;
import org.jboss.seam.faces.FacesMessages;
import org.jboss.seam.framework.EntityHome;
import org.jboss.seam.international.StatusMessage.Severity;

/**
 * The Class VaFacilityHome.
 */
@Name("vaFacilityHome")
@Restrict("#{s:hasRole('FEE_SUPERVISOR') || s:hasRole('FEE_VISN_PGM_MANAGER') || s:hasRole('ADMIN')}")
public class VaFacilityHome extends EntityHome<VaFacility> {

    /** The visn id cd. */
    @Min(0)
    private Long visnIdCd;

    /** The va type cd. */
    @NotNull
    private String vaTypeCd;

    /** The faces messages. */
    @In
    private FacesMessages facesMessages;

    /** The app user. */
    @In(value = "appUser", required = true, scope = ScopeType.SESSION)
    private AppUser appUser;

    /**
     * Sets the va facility va facility cd.
     *
     * @param id the new va facility va facility cd
     */
    public void setVaFacilityVaFacilityCd(String id) {
        setId(id);
    }

    /**
     * Gets the va facility va facility cd.
     *
     * @return the va facility va facility cd
     */
    public String getVaFacilityVaFacilityCd() {
        return (String) getId();
    }

    /**
     * Gets the visn id cd.
     *
     * @return the visn id cd
     */
    public Long getVisnIdCd() {
        return visnIdCd;
    }

    /**
     * Sets the visn id cd.
     *
     * @param visnIdCd the new visn id cd
     */
    public void setVisnIdCd(Long visnIdCd) {
        this.visnIdCd = visnIdCd;
    }

    /**
     * Gets the va type cd.
     *
     * @return the va type cd
     */
    public String getVaTypeCd() {
        return vaTypeCd;
    }

    /**
     * Sets the va type cd.
     *
     * @param vaTypeCd the new va type cd
     */
    public void setVaTypeCd(String vaTypeCd) {
        this.vaTypeCd = vaTypeCd;
    }

    /**
     * Wire.
     */
    public void wire() {
        visnIdCd = getInstance().getVisn().getVisnIdCd();
        vaTypeCd = getInstance().getVaFacilityType().getVaTypeCd();
    }

    /**
     * Gets the defined instance.
     *
     * @return the defined instance
     */
    public VaFacility getDefinedInstance() {
        return isIdDefined() ? getInstance() : null;
    }

    /**
     * Gets the contact infos.
     *
     * @return the contact infos
     */
    public List<ContactInfo> getContactInfos() {
        return getInstance() == null ? null : new ArrayList<ContactInfo>(
                getInstance().getContactInfos());
    }

    /**
     * Gets the zip codes.
     *
     * @return the zip codes
     */
    public List<ZipCode> getZipCodes() {
        return getInstance() == null ? null : new ArrayList<ZipCode>(
                getInstance().getZipCodes());
    }

    /**
     * Check.
     *
     * @return true, if successful
     */
    public boolean check() {
        VaFacility v = getEntityManager().find(VaFacility.class, getInstance().getVaFacilityCd());
        if (v != null) {
            FeeMessages.addFromResourceBundle(
                    "vaFacility:vaFacilityCdField:vaFacilityCd",
                    javax.faces.application.FacesMessage.SEVERITY_ERROR,
                    "gov.va.med.fee.tools.StationMaint.STATION_EXIST");
            return false;
        }
        return true;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @Transactional
    public String persist() {
        if (check()) {
            getInstance().setCreatedBy(appUser.getUserName());
            getInstance().setDateCreated(new Date());
            prepareInstance();
            return super.persist();
        } else {
            facesMessages.addFromResourceBundle(
                    Severity.ERROR,
                    "gov.va.med.fee.tools.StationMaint.STATION_EXIST");
        }
        return null;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @Transactional
    @Restrict("#{s:hasRole('FEE_VISN_PGM_MANAGER') || s:hasRole('ADMIN')}")
    public String remove() {
        try {
            return super.remove();
        } catch (Throwable c) {
            facesMessages.addFromResourceBundle(
                    Severity.ERROR,
                    "gov.va.med.fee.tools.StationMaint.REMOVE_FAILED");
        }
        return null;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @Transactional
    public String update() {
        prepareInstance();
        return super.update();
    }

    /**
     * Overrides to return the correct persistent context name for this app.
     *
     * @return persistence context name
     */
    @Override
    protected final String getPersistenceContextName() {
       return "FPPSEntityManager";
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected VaFacility createInstance() {
        VaFacility vaFacility = new VaFacility();
        vaFacility.setVaFacilityType(new VaFacilityType());
        Visn visn = new Visn();
        visn.setVisnIdCd(-1);
        vaFacility.setVisn(visn);
        return vaFacility;
    }

    private void prepareInstance() {
        VaFacility instance = getInstance();

        VaFacilityType vaFacilityType = getEntityManager().find(VaFacilityType.class, vaTypeCd);
        instance.setVaFacilityType(vaFacilityType);

        Visn visn = getEntityManager().find(Visn.class, visnIdCd);
        instance.setVisn(visn);
    }

}
