package gov.va.med.fee.report;

import static org.testng.AssertJUnit.assertEquals;

import java.util.Calendar;
import java.util.Date;

import org.hibernate.validator.ClassValidator;
import org.hibernate.validator.InvalidValue;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

/**
 * The Class ReportDateTest.
 */
public class ReportDateTest {

    /** The date entry. */
    private DateEntry dateEntry;

    /** The date value. */
    private DateValue dateValue;

    /** The class validator. */
    private ClassValidator<DateEntry> classValidator;

    /** The invalid values. */
    private InvalidValue[] invalidValues;

    /**
    * Sets the up.
    */
    @BeforeClass
    public final void setUp() {
        dateEntry = new DateEntry();
        dateValue = new DateValue();
        dateEntry.setDateValue(dateValue);
        classValidator = new ClassValidator<DateEntry>(DateEntry.class);
    }

    /**
    * Test no entries.
    *
    * @throws Exception
    *             the exception
    */
    @Test
    public final void testNoEntries() throws Exception {
        dateValue.reset();
        invalidValues = classValidator.getInvalidValues(dateEntry);
        assertEquals(1, invalidValues.length);
    }

    /**
    * Test age.
    *
    * @throws Exception
    *             the exception
    */
    @Test
    public final void testAge() throws Exception {
        dateValue.reset();
        dateValue.ageTab();
        dateValue.setDays(1);
        invalidValues = classValidator.getInvalidValues(dateEntry);
        assertEquals(0, invalidValues.length);
    }

    /**
    * Test missing to date.
    *
    * @throws Exception
    *             the exception
    */
    @Test
    public final void testMissingToDate() throws Exception {
        dateValue.reset();
        dateValue.rangeTab();
        dateValue.setFromDate(new Date());
        invalidValues = classValidator.getInvalidValues(dateEntry);
        assertEquals(1, invalidValues.length); // missing toDate
    }

    /**
    * Test range.
    *
    * @throws Exception
    *             the exception
    */
    @Test
    public final void testRange() throws Exception {
        dateValue.reset();
        dateValue.rangeTab();
        Date d = new Date();
        dateValue.setFromDate(d);
        dateValue.setToDate(d);
        invalidValues = classValidator.getInvalidValues(dateEntry);
        assertEquals(0, invalidValues.length); // has both from and to
    }

    /**
    * Test valid range.
    *
    * @throws Exception
    *             the exception
    */
    @Test
    public final void testValidRange() throws Exception {
        // test valid date ranges
        Calendar c = Calendar.getInstance();
        Date d = new Date();
        c.setTime(d);
        setToMidnight(c);

        dateValue.reset();
        dateValue.rangeTab();
        dateValue.setFromDate(c.getTime());

        c.add(Calendar.MINUTE, 1);
        dateValue.setToDate(c.getTime());
        invalidValues = classValidator.getInvalidValues(dateEntry);
        assertEquals(0, invalidValues.length); // one minutes after - OK

        c.add(Calendar.MINUTE, -1);
        dateValue.setToDate(c.getTime());
        invalidValues = classValidator.getInvalidValues(dateEntry);
        assertEquals(0, invalidValues.length); // one minutes before - not OK
    }

    /**
    * Sets the to midnight.
    *
    * @param cal
    *            the new to midnight
    */
    private void setToMidnight(Calendar cal) {
        cal.set(Calendar.HOUR, 0);
        cal.set(Calendar.MINUTE, 0);
        cal.set(Calendar.SECOND, 0);
        cal.set(Calendar.MILLISECOND, 0);
        cal.set(Calendar.AM_PM, Calendar.AM);
    }
}
