package gov.va.med.hac.edi.era.jpa;

import gov.va.med.hac.edi.reposload.x12.jpa.X12eReposLoad;
import gov.va.med.hac.edi.utils.*;
import gov.va.med.hac.edi.utils.jpa.DbTransactionEnderException;
import gov.va.med.hac.edi.utils.jpa.JPAUtil;
import gov.va.med.hac.edi.utils.jpa.LogUtils;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.MissingOptionException;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.*;

import java.util.logging.*;

import javax.naming.NamingException;

import javax.persistence.EntityManager;


/**
 * This process deletes any data in the staging tables related to a
 * particular 835 file (identified by the file number).
 * This process is run to delete **all** the data related to the 835 file
 * that was previously generated.
 * All the data related to the file is deleted because one or more of the records
 * in the file contained non-compliant data.
 * All the data related to the file, rather than non-compliant data, is deleted
 * from this system because the data must be fixed at the source of
 * the error, an external system, and then re-generated.
 *
 *   The program may be invoked using the following command.
 *
 *   java DeleteStg835 <FILE_NUMBER>
 *
 *   Where file number is the number of the 835 file that contains
 *   compliance errors.
 *
 *   Dependencies
 *   dbhibernate.properties - for connection information
         persistence.xml - for the eResposExtract configuration
         (FYI - any persistence mapping could be used this code uses native sql and not the mappings,
         but we use this unit as a place holder)
         *** TODO: Make a 1 argument getEntityManager that takes a string (not file)
                                 i.e. don't require the use of a named persistence unit.
 *
 * @author HIPAA_READY
 */
public final class DeleteStg835 {
    static final Logger summaryLOGGER = LoggerFactory.getLogger(DeleteStg835.class);
    static final Logger traceLOGGER = LoggerFactory.getLogger("trace." +
            DeleteStg835.class);

    /**
     * This is the main method that is executed to delete data related to an 835
     * file from the staging tables.  The main method expects one argument -
     * the file number that contained non-compliant data.
     *
     * The program may be invoked using the following command.
     *
     *   java DeleteStg835 <FILE_NUMBER>
     * @param args 835 File Number
     */
    public static void main(final String[] args) {
        // Parse the command line
        CommandLine line = null;
        Options opts = getCliOptions();

        try {
            CommandLineParser parser = new GnuParser();
            line = parser.parse(opts, args);

            // If the "help" option is specified, just show usage options and
            // exit
            if (line.hasOption("h")) {
                printUsage(opts, args, false);
                System.exit(1);
            }

            // If any required command line parameter is missing, just show
            // usage options and exit
        } catch (MissingOptionException moe) {
            printUsage(opts, args, true);
            System.exit(1);
        } catch (ParseException pe) {
            System.out.println("Problem parsing the command line");
            summaryLOGGER.error("Problem parsing the command line");
            traceLOGGER.error("Problem parsing the command line");
            printUsage(opts, args, false);
            System.exit(1);
        }

        boolean processSuccess = false;

        // Catch any exception (outside of setting up for logging)
        try {
            LogUtils.stampStartingTime("Delete Stage 835 Start", summaryLOGGER);
            summaryLOGGER.info("Code Version: Refactoring Project");
            traceLOGGER.info("Code Version: Refactoring Project");

            String vFileNumber = line.getOptionValue("n");

            if (vFileNumber != null) {
                vFileNumber = vFileNumber.trim();
            }

            if ((vFileNumber == null) || (vFileNumber.length() < 1)) {
                summaryLOGGER.info(
                    "DeleteStg835: Please Pass 835 File Number as Parameter");
            } else {
                summaryLOGGER.info("File Number: " + vFileNumber);

                final DeleteStg835ProcessedData process = new DeleteStg835ProcessedData();

                File configFile = null;
                String configFilePath = line.getOptionValue("f");

                if ((configFilePath != null) && (configFilePath.length() > 0)) {
                    configFile = JPAUtil.getJpaConfigFile(configFilePath.trim());
                } else {
                    configFile = JPAUtil.getJpaConfigFile(
                            "dbhibernate.properties");
                }

                EntityManager dbEM = null;
                dbEM = JPAUtil.getEntityManager(configFile, "eResposExtract");

                process.processDelete835Data(vFileNumber, dbEM);

                // Put a completion message into the log file
                summaryLOGGER.info("Process completed successfully");
                traceLOGGER.info("Process completed successfully");

                processSuccess = true;
            }
        } catch (OutOfMemoryError e) {
            summaryLOGGER.error("!!!!! Out Of Memory Error !!!!!", e);
            traceLOGGER.error("!!!!! Out Of Memory Error !!!!!", e);
        } catch (Throwable t) {
            summaryLOGGER.error("Exception in processing", t);
            traceLOGGER.error("Exception in processing", t);
        } finally {
            if (!processSuccess) {
                summaryLOGGER.error("!!!!! Process seems to have failed !!!!!");
                traceLOGGER.error("!!!!! Process seems to have failed !!!!!");
            } else {
                summaryLOGGER.info("Delete Stage 835 End");
                traceLOGGER.info("Delete Stage 835 End");
                LogUtils.stampEndingTime("Delete Stage 835 End", summaryLOGGER);
            }
        }
    }

    private static void printUsage(Options opts, String[] inArgs,
        boolean missingArg) {
        final int helpwidth = 80;
        HelpFormatter fmtr = new HelpFormatter();
        System.out.println();

        if (missingArg) {
            System.out.println("Some required commandline argument was missing");
        }

        fmtr.printHelp(helpwidth, "DeleteStg835", "", opts, "", true);

        if (missingArg) {
            System.out.println(inArgs.length +
                " command line arguments were received: ");

            for (int i = 0; i < inArgs.length; i++) {
                System.out.println(inArgs[i]);
            }
        }
    }

    /**
     * Helper method to populate command line options objects.
     *
     * @return the <code>Options</code> object containing the set of command line option.
     */
    @SuppressWarnings("static-access")
    private static Options getCliOptions() {
        Options options = new Options();

        Option configfile = OptionBuilder.withArgName("jpaFile").hasArg()
                                         .withLongOpt("jpaFile")
                                         .withDescription("The JPA Configuration file, such as hibernate.properties.")
                                         .create("f");

        Option fileNumber = OptionBuilder.withArgName("fileNumber").hasArg()
                                         .isRequired().withLongOpt("fileNumber")
                                         .withDescription("The name of the 835 file whose data shoule be removed from the staging tables.")
                                         .create("n");

        Option help = OptionBuilder.withLongOpt("help")
                                   .withDescription("Print this message.")
                                   .create("h");

        options.addOption(help);
        options.addOption(configfile);
        options.addOption(fileNumber);

        return options;
    }
}
