package gov.va.med.hac.edi.era.jpa;

import gov.va.med.domain.erepos.Stg835Counter;
import gov.va.med.hac.edi.utils.StringUtils;
import gov.va.med.hac.edi.utils.jpa.DbTransactionEnderException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.UUID;

import javax.persistence.EntityManager;
import javax.persistence.Query;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class ProcessEdi835ErrorQueue {
  static final Logger summaryLOGGER = LoggerFactory.getLogger(ProcessEdi835ErrorQueue.class);
  static final Logger traceLOGGER = LoggerFactory.getLogger("trace."
      + ProcessEdi835ErrorQueue.class);

  final private static String COMPLIANCE_ERROR_STATUS = "CMP_ERROR";
  final private static String ENVELOPE_ERROR_STATUS = "ENV_ERROR";

  // This method will only be run if there was an error during compliance
  // checking. It will delete records and update statuses, putting the
  // repository and the staging tables into a state from which processing
  // can continue.
  // Also, this method will definitely run if there was an error during compliance
  // checking. Therefore, if within this method we find no transactions that
  // had errors, we conclude that there was an ENVELOPE_ERROR.
  protected void processErrorQueue(final String vRunId, EntityManager dbEM)
          throws DbTransactionEnderException {
    // These holds any transaction numbers that had errors
    final ArrayList transactionsWithErrors;

    final String fileNumber;

    try {

      // Get the list of transaction numbers for this run that had errors,
      // with transaction code HP
      transactionsWithErrors = getTransactionList(vRunId, dbEM);

      // Create a comma delimited list of transaction numbers that can be used
      // by database queries to select data if the transaction number is in this list
      String transactionsStringWithLeadingZeros = StringUtils.concatenateList(
          transactionsWithErrors,
                    false);

      // Get the file number for the run with this runId
      // This will be "" if there were no transactions with errors, or if we simply cannot
      // find a filename for this runId. (Don't know whether that can actually happen.)
      fileNumber = getFileNumber(vRunId, dbEM);

      // If there were transactions with errors, carry out four steps to
      // back the process out and get the offending transactions out of the
      // run, so the run can proceed without them
      if (!(transactionsWithErrors.isEmpty())) { // NOPMD - Easier to understand this way

        // For those transactions with errors, mark their status in 3 staging tables
        updateStg835ReposStatuses(COMPLIANCE_ERROR_STATUS,
                        fileNumber,
                        transactionsWithErrors, dbEM);

        dbEM.getTransaction().begin();
        
        // For the transactions with errors, put a suffix on the file number in 5 tables
        updateFileNumber(fileNumber, dbEM);

        // For the transactions without errors, delete their records from the target tables
        deleteFromEdiRepos(COMPLIANCE_ERROR_STATUS, fileNumber,
                    transactionsStringWithLeadingZeros, dbEM);

        // For those transactions that did not have errors, they are still marked
        // with the original file number. Mark the record in EDI_835_PROCESS for
        // that file number as 'NEW', so it will be reprocessed through the
        // extraction and compliance steps.
        updateEdiProcess(fileNumber, dbEM);

        // If there were no transactions with errors, we know there was an
        // error anyway, or we wouldn't be in this method. It must have been
        // an envelope error. If so, run 2 of the 4 steps
      } else {
        // If no file number was found, log this fact
        if ("".equals(fileNumber)) {
          summaryLOGGER.info("Couldn't find the File Number from Run ID");
          traceLOGGER.info("Couldn't find the File Number from Run ID");

        } else {
          // Mark the status of all transactions in 3 staging tables
          updateStg835ReposStatuses(ENVELOPE_ERROR_STATUS,
                            fileNumber,
                            transactionsWithErrors,
                            dbEM);
          dbEM.getTransaction().begin();
          // Delete all transactions' records from the target tables
          deleteFromEdiRepos(ENVELOPE_ERROR_STATUS, fileNumber,
                        transactionsStringWithLeadingZeros, dbEM);
        }
      }

      traceLOGGER.info("Committing both database transactions");
      dbEM.getTransaction().commit();
      traceLOGGER.info("Commits successful");
    }
    catch (Throwable ex) {

      if (dbEM != null) {
        if (dbEM.getTransaction().isActive()) {
          traceLOGGER.info("Rolling back both database transactions");
          dbEM.getTransaction().rollback();
          traceLOGGER.info("Rollbacks successful");
        }
      }
      throw new DbTransactionEnderException("Rolled Back both transactions", ex);
    }
  }

  private ArrayList getTransactionList(String vRunId, EntityManager dbEM) throws Exception {

    Query getTransNumbers = dbEM
        .createNativeQuery("SELECT TRANS_NO FROM TRLOG WHERE TRANS_NAME = '835' AND TRANS_CODE = 'HP' AND ERRORS > 0 AND RUN_ID = :runId GROUP BY TRANS_NO");
    getTransNumbers.setParameter("runId", vRunId);
    List transNumbers = getTransNumbers.getResultList();
    Iterator transNumbersIterator = transNumbers.iterator();

    ArrayList list = new ArrayList();
    String vTransNo = null;
    while (transNumbersIterator.hasNext()) {
      vTransNo = transNumbersIterator.next().toString();
      list.add(vTransNo);
    }
    return list;
  }

  // Update the statuses in the three leading staging tables
  private void updateStg835ReposStatuses(final String errStatus,
                                         final String inFileNumber,
                                         final ArrayList inTransactionsWithErrors,
                                         EntityManager dbEM)
            throws SQLException {

    dbEM.getTransaction().begin();
    
    // Not sure this transaction is necessary
    String vTransNos = "";
    String vLogStr = "";

    if (inFileNumber.startsWith("W84146") || inFileNumber.startsWith("C74161")) {
      vTransNos = StringUtils.concatenateList(inTransactionsWithErrors,
                    true);
    } else {
      vTransNos = StringUtils.concatenateList(inTransactionsWithErrors,
                    false);
    }

    // If there were transactions marked with errors, update statuses based on
    // the transaction numbers
    if (errStatus.equalsIgnoreCase(COMPLIANCE_ERROR_STATUS)) {
      // This method will do its own logging
      updateStatusesByTransactionNumber(errStatus, inFileNumber, dbEM, vTransNos);

      // If there were no transactions marked with errors, update statuses based
      // on the presence of "DELETE" as a status, for this file number
    } else {
      updateStatusesByDeleteStatus(errStatus, inFileNumber, dbEM);
      vLogStr = "Updated the following File Number to " + errStatus
          + " status instead of deleting: " + inFileNumber;
      traceLOGGER.debug(vLogStr);
    }
    dbEM.getTransaction().commit();
  }

  private void updateStatusesByTransactionNumber(final String errStatus,
                                                 final String inFileNumber, EntityManager dbEM,
                                                 String inTransactionsWithErrors)
      throws SQLException {

    String sHdr = "UPDATE STG_835_HDR SET STSE_STATUS = :errStatus " +
        "WHERE ST02TXN_CTL_NBR IN " + "( " + inTransactionsWithErrors
        + " ) AND FILE_NUMBER = :inFileNumber ";
    String sClp = "UPDATE STG_835_CLP SET CLP_STATUS = :errStatus " +
        "WHERE ST02TXN_CTL_NBR IN " + "( " + inTransactionsWithErrors
        + " ) AND FILE_NUMBER = :inFileNumber ";
    String sSvc = "UPDATE STG_835_SVC SET SVC_STATUS = :errStatus " +
        "WHERE CLP_KEY IN (SELECT CLP_KEY FROM STG_835_CLP WHERE ST02TXN_CTL_NBR IN " +
        "( " + inTransactionsWithErrors + " ) AND FILE_NUMBER = :inFileNumber) ";

    Query psHdr = dbEM.createNativeQuery(sHdr);
    psHdr.setParameter("errStatus", errStatus);
    psHdr.setParameter("inFileNumber", inFileNumber);

    Query psClp = dbEM.createNativeQuery(sClp);
    psClp.setParameter("errStatus", errStatus);
    psClp.setParameter("inFileNumber", inFileNumber);

    Query psSvc = dbEM.createNativeQuery(sSvc);
    psSvc.setParameter("errStatus", errStatus);
    psSvc.setParameter("inFileNumber", inFileNumber);

    psHdr.executeUpdate();
    psClp.executeUpdate();
    psSvc.executeUpdate();

  }

  private void updateStatusesByDeleteStatus(final String errStatus, final String inFileNumber,
                                            EntityManager dbEM)
      throws SQLException {

    String vDelStr = "SELECT ST02TXN_CTL_NBR FROM STG_835_HDR WHERE STSE_STATUS = 'DELETE' ";
    String sHdr = "UPDATE STG_835_HDR SET STSE_STATUS = :errStatus " +
        "where STSE_STATUS = 'DELETE' " + " AND FILE_NUMBER =  :inFileNumber ";
    String sClp = "UPDATE STG_835_CLP SET CLP_STATUS = :errStatus " +
        "WHERE ST02TXN_CTL_NBR IN " + "( " +
        "SELECT ST02TXN_CTL_NBR FROM STG_835_HDR WHERE STSE_STATUS = 'DELETE' " +
        "AND FILE_NUMBER = :inFileNumber) ";
    String sSvc = "UPDATE STG_835_SVC SET SVC_STATUS = :errStatus " +
        "WHERE CLP_KEY IN (SELECT CLP_KEY FROM STG_835_CLP WHERE " +
        "ST02TXN_CTL_NBR IN " + "( " +
        "SELECT ST02TXN_CTL_NBR FROM STG_835_HDR WHERE STSE_STATUS = 'DELETE' " +
        "AND FILE_NUMBER = :inFileNumber)) ";

    Query psHdr = dbEM.createNativeQuery(sHdr);
    psHdr.setParameter("errStatus", errStatus);
    psHdr.setParameter("inFileNumber", inFileNumber);

    Query psClp = dbEM.createNativeQuery(sClp);
    psClp.setParameter("errStatus", errStatus);
    psClp.setParameter("inFileNumber", inFileNumber);

    Query psSvc = dbEM.createNativeQuery(sSvc);
    psSvc.setParameter("errStatus", errStatus);
    psSvc.setParameter("inFileNumber", inFileNumber);

    // TWP - Order is very important in these next three statements
    // you need to be able to find records in 'DELETE' status in
    // the header table so don;t update that too early.
    psClp.executeUpdate();
    psSvc.executeUpdate();
    psHdr.executeUpdate();

  }

  // Update the file number column in several tables, giving it a suffix that
  // indicates how many times the file has been processed so far.
  private void updateFileNumber(final String inFileNumber, EntityManager dbEM)
            throws SQLException {
    String newCount;
    String newFileNumber = "";

    // Get a record out of table STG_835_COUNTER for this file_number, if there is one
    // SELECT sc FROM gov.va.med.domain.erepos.Stg835Counter sc WHERE sc.fileNumber =:fileNumber
    Query getStg835Counter = dbEM
        .createQuery("select sc from gov.va.med.domain.erepos.Stg835Counter sc where sc.fileNumber=:fileNumber");
    getStg835Counter.setParameter("fileNumber", inFileNumber);
    List stg835Counters = getStg835Counter.getResultList();
    Iterator stg835CounterIterator = stg835Counters.iterator();

    Stg835Counter crStg835Counter = null;
    if (stg835CounterIterator.hasNext()) {
      crStg835Counter = (Stg835Counter) stg835CounterIterator.next();
    }

    // If no such record was found, create a new one, with the counter = 1
    if (crStg835Counter == null) {
      newCount = "1";
      final Stg835Counter trStg835Counter = new Stg835Counter();
      trStg835Counter.setFileNumber(inFileNumber);
      trStg835Counter.setCounter(newCount);
      dbEM.persist(trStg835Counter);

      // If a record was found, increment its counter
    } else {
      final String oldCount = crStg835Counter.getCounter();
      newCount = Integer.toString(Integer.parseInt(oldCount) + 1);
      crStg835Counter.setCounter(newCount);
      dbEM.persist(crStg835Counter);
    }
    newFileNumber = inFileNumber + "_" + newCount;
    updateTables(inFileNumber, newFileNumber, dbEM);
    traceLOGGER.debug("Successfully Updated the  FileNumber from " +
                inFileNumber + " to " + newFileNumber);
  }

  private void updateTables(final String oldFileNumber, final String newFileNumber,
                            EntityManager dbEM)
            throws SQLException {

    final String sHdrSel = "(select ST02TXN_CTL_NBR from STG_835_HDR where "
        +
                "STSE_STATUS = 'CMP_ERROR' and FILE_NUMBER = :oldFileNumber ) and FILE_NUMBER =:oldFileNumber ";

    final String sClp = "Update STG_835_CLP set FILE_NUMBER = :newFileNumber where ST02TXN_CTL_NBR in "
          + sHdrSel; // NOPMD
    final Query psClp = dbEM.createNativeQuery(sClp);
    psClp.setParameter("oldFileNumber", oldFileNumber);
    psClp.setParameter("newFileNumber", newFileNumber);

    final String sHdrTrn = "Update STG_835_HDR_TRN set FILE_NUMBER = :newFileNumber where ST02TXN_CTL_NBR in "
          + sHdrSel;
    final Query psHdrTrn = dbEM.createNativeQuery(sHdrTrn);
    psHdrTrn.setParameter("oldFileNumber", oldFileNumber);
    psHdrTrn.setParameter("newFileNumber", newFileNumber);

    final String sHdrBpr = "Update STG_835_HDR_BPR set FILE_NUMBER = :newFileNumber where ST02TXN_CTL_NBR in "
          + sHdrSel;
    final Query psHdrBpr = dbEM.createNativeQuery(sHdrBpr);
    psHdrBpr.setParameter("oldFileNumber", oldFileNumber);
    psHdrBpr.setParameter("newFileNumber", newFileNumber);

    final String sPlb = "Update STG_835_PLB set FILE_NUMBER = :newFileNumber where ST02TXN_CTL_NBR in "
          + sHdrSel;
    final Query psPlb = dbEM.createNativeQuery(sPlb);
    psPlb.setParameter("oldFileNumber", oldFileNumber);
    psPlb.setParameter("newFileNumber", newFileNumber);

    final String sHdr = "Update STG_835_HDR set FILE_NUMBER = :newFileNumber where STSE_STATUS = 'CMP_ERROR' and FILE_NUMBER = :oldFileNumber ";
    final Query psHdr = dbEM.createNativeQuery(sHdr);
    psHdr.setParameter("oldFileNumber", oldFileNumber);
    psHdr.setParameter("newFileNumber", newFileNumber);

    psClp.executeUpdate();
    psHdrTrn.executeUpdate();
    psHdrBpr.executeUpdate();
    psPlb.executeUpdate();
    psHdr.executeUpdate();
  }

  //
  private void deleteFromEdiRepos(final String errStatus,
                                  final String inFileNumber,
                                  String inTransactionsStringWithLeadingZeros, EntityManager dbEM)
      throws SQLException {

    String sFileClause = "";
    String sClpClause = "";
    String sSvcClause = "";
    String sFileClauseEnv = "WHERE FILE_NUMBER = '" + inFileNumber + "' ";
    String vLogStr = "";

    if (errStatus.equalsIgnoreCase(COMPLIANCE_ERROR_STATUS)) {
      sFileClause = "WHERE FILE_NUMBER = '" + inFileNumber + "' AND "
          + "ST02TXN_CTL_NBR IN ("
          + inTransactionsStringWithLeadingZeros + ") ";
      sClpClause = "WHERE CLAIM_PYMT_INFO_KEY IN (SELECT CLAIM_PYMT_INFO_KEY FROM CLAIM_PYMT_INFO "
          + sFileClause + ") ";
      vLogStr = "Deleted data from eRepos 835 tables for the following FileNumber : "
          + inFileNumber
          + " and TransCtlNbr : "
          + inTransactionsStringWithLeadingZeros;
    } else {
      sFileClause = sFileClauseEnv;
      sClpClause = "WHERE CLAIM_PYMT_INFO_KEY IN (SELECT CLAIM_PYMT_INFO_KEY FROM CLAIM_PYMT_INFO "
          + sFileClauseEnv + ") ";
      vLogStr = "Deleted data from eRepos 835 tables for the following FileNumber : "
          + inFileNumber;
    }

    sSvcClause = "WHERE SVC_PYMT_INFO_KEY IN (SELECT SVC_PYMT_INFO_KEY FROM SVC_PYMT_INFO WHERE CLAIM_PYMT_INFO_KEY IN (SELECT CLAIM_PYMT_INFO_KEY FROM CLAIM_PYMT_INFO "
        + sFileClause + ")) ";
    final String sHcareRemarkCd = "DELETE FROM HCARE_REMARK_CD "
        + sSvcClause;
    final String sCases = "DELETE FROM CASES " + sClpClause;
    final String sClaimAmount = "DELETE FROM CLAIM_AMOUNT " + sClpClause;
    final String sClaimDt = "DELETE FROM CLAIM_DT " + sClpClause;
    final String sClaimDtm = "DELETE FROM CLAIM_DTM " + sClpClause;
    final String sMediInptAdj = "DELETE FROM MEDI_INPT_ADJDCT " + sClpClause;
    final String sMediOutptAdj = "DELETE FROM MEDI_OUTPT_ADJDCT " + sClpClause;
    final String sClaimProdDt = "DELETE FROM CLAIM_DT " + sFileClause;
    final String sCntcInfo = "DELETE FROM CNTC_INFO " + sFileClause;
    final String sEntityAddress = "DELETE FROM ENTITY_ADDRESS "
        + sFileClause;
    final String sEntityadtlid = "DELETE FROM ENTITY_ADTL_ID "
        + sFileClause;
    final String sEntityName = "DELETE FROM ENTITY_NAME " + sFileClause;
    final String sEntityId = "DELETE FROM ENTITY_ID " + sFileClause;
    final String sSvcPymtInfo = "DELETE FROM SVC_PYMT_INFO " + sClpClause;
    final String sClaimPymtInfo = "DELETE FROM CLAIM_PYMT_INFO "
        + sFileClause;
    final String sDetailHeadInfo = "DELETE FROM DETAIL_HEAD_INFO "
        + sFileClause;
    final String sTraceNumber = "DELETE FROM TRACE_NUMBER " + sFileClause;
    final String sFinancialInfo = "DELETE FROM FINANCIAL_INFO "
        + sFileClause;
    final String sPrvAdj = "DELETE FROM PRV_ADJ " + sFileClause;
    final String sTxnSet = "DELETE FROM TXN_SET " + sFileClause;
    final String sFunctionalGroup = "DELETE FROM FUNCTIONAL_GROUP "
        + sFileClause;
    final String sIntchgControl = "DELETE FROM INTCHG_CONTROL "
        + sFileClause;

    final Query psHcareRemarkCd = dbEM.createNativeQuery(sHcareRemarkCd);
    final Query psCases = dbEM.createNativeQuery(sCases);
    final Query psClaimAmount = dbEM.createNativeQuery(sClaimAmount);
    final Query psClaimDt = dbEM.createNativeQuery(sClaimDt);
    final Query psClaimDtm = dbEM.createNativeQuery(sClaimDtm);
    final Query psMediInptAdj = dbEM.createNativeQuery(sMediInptAdj);
    final Query psMediOutptAdj = dbEM.createNativeQuery(sMediOutptAdj);
    
    final Query psClaimProdDt = dbEM.createNativeQuery(sClaimProdDt);
    final Query psCntcInfo = dbEM.createNativeQuery(sCntcInfo);
    final Query psEntityAddress = dbEM.createNativeQuery(sEntityAddress);
    final Query psEntityadtlid = dbEM.createNativeQuery(sEntityadtlid);
    final Query psEntityName = dbEM.createNativeQuery(sEntityName);
    final Query psEntityId = dbEM.createNativeQuery(sEntityId);
    final Query psSvcPymtInfo = dbEM.createNativeQuery(sSvcPymtInfo);
    final Query psClaimPymtInfo = dbEM.createNativeQuery(sClaimPymtInfo);
    final Query psDetailHeadInfo = dbEM.createNativeQuery(sDetailHeadInfo);
    final Query psTraceNumber = dbEM.createNativeQuery(sTraceNumber);
    final Query psFinancialInfo = dbEM.createNativeQuery(sFinancialInfo);
    final Query psPrvAdj = dbEM.createNativeQuery(sPrvAdj);
    final Query psTxnSet = dbEM.createNativeQuery(sTxnSet);

    Query psIntchgControl = null, psFunctionalGroup = null;
    if (errStatus.equalsIgnoreCase(ENVELOPE_ERROR_STATUS)) {
      psIntchgControl = dbEM.createNativeQuery(sIntchgControl);
      psFunctionalGroup = dbEM.createNativeQuery(sFunctionalGroup);
    }

    

    // Create an array with all but the last two PreparedStatements in it
    // The last two are used only for special cases
    Query[] ps = { psHcareRemarkCd, psCases, psClaimAmount, psClaimDt,
    		psClaimDtm, psMediInptAdj, psMediOutptAdj, psClaimProdDt,
    		psCntcInfo, psEntityAddress, psEntityadtlid, psEntityName,
    		psEntityId, psSvcPymtInfo, psClaimPymtInfo, psDetailHeadInfo,
    		psTraceNumber, psFinancialInfo, psPrvAdj, psTxnSet };

    for (int i = 0; i < ps.length; i++) {
      ps[i].executeUpdate();
    }

    if (errStatus.equalsIgnoreCase(ENVELOPE_ERROR_STATUS)) {
      psIntchgControl.executeUpdate();
      psFunctionalGroup.executeUpdate();
    }

    traceLOGGER.debug(vLogStr);
  }

  private void updateEdiProcess(final String inFileNumber, EntityManager dbEM)
            throws SQLException {

    final String sEdiProcess = "UPDATE EDI_835_PROCESS SET STATUS = 'NEW' " +
                "WHERE FILE_NUMBER = :fileNumber";
    final Query psEdiProcess = dbEM.createNativeQuery(sEdiProcess);
    psEdiProcess.setParameter("fileNumber", inFileNumber);
    psEdiProcess.executeUpdate();
    traceLOGGER.debug("Updated the Status of File Number : " + inFileNumber + " to 'NEW' ");
  }

  private String getFileNumber(final String runId, EntityManager dbEM) {
    String fileName = "";
    String fileNumber;

    // Get the filenames for this run that have "E" in them
    final String strFileName = "SELECT DISTINCT FILENAME FROM TRLOG " +
                "WHERE RUN_ID = :runId AND FILENAME LIKE '%E%'";

    Query getFileNumbers = dbEM.createNativeQuery(strFileName);
    getFileNumbers.setParameter("runId", runId);
    List fileNumbers = getFileNumbers.getResultList();
    Iterator fileNumberIterator = fileNumbers.iterator();

    // Get the last such file name - there should only be one
    while (fileNumberIterator.hasNext()) {
      fileName = (String) fileNumberIterator.next().toString();
    }

    // Look for a "W8", an "M8", or a "W1" in the filename, in that order,
    // stopping when one of these strings is encountered
    int idx1 = fileName.lastIndexOf("W8");

    if (idx1 < 0) {
      idx1 = fileName.lastIndexOf("M8");
    }
    if (idx1 < 0) {
      idx1 = fileName.lastIndexOf("W1");
    }
    if (idx1 < 0) {
      idx1 = fileName.lastIndexOf("S8");
    }
    if (idx1 < 0) {
      idx1 = fileName.lastIndexOf("S5");
    }
    if (idx1 < 0) {
      idx1 = fileName.lastIndexOf("C7");
    }
    if (idx1 < 0) {
        idx1 = fileName.lastIndexOf("SXC");
    }
    // If any of the above three strings was found, use the filename, but
    // cut off anything before the found string
    if (idx1 > 0) {
      fileNumber = fileName.substring(idx1, fileName.length());
      // If none of the above three strings was found, return a blank string
    } else {
      fileNumber = "";
    }

    return fileNumber;
  }

}
