package gov.va.med.hac.edi.era.jpa;

import gov.va.med.domain.erepos.*;
import gov.va.med.hac.edi.utils.jpa.ConversionUtil;
import gov.va.med.hac.edi.utils.jpa.DateUtil;
import gov.va.med.hac.edi.utils.jpa.DbTransactionEnderException;
import gov.va.med.hac.edi.utils.jpa.FileUtils;
import gov.va.med.hac.edi.utils.jpa.JPAUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.*;

import java.math.BigDecimal;

import java.sql.SQLException;

import java.util.Iterator;
import java.util.List;
import java.util.UUID;

import javax.persistence.EntityManager;
import javax.persistence.Query;


/**
 * This class contains the functionality that reads data from a .DAT data file
 * containing stage data and stores that data in the staging tables/repository.
 *
 * The processLoad method is the main entry point of the class. This method
 * opens the data file, reads it line by line, identifies the type of line read
 * and, based on the type, passes the tokenized line of data to helpers method that map
 * data to the appropriate object/fields and finally, the helper methods store
 * the objects/data in the database.
 *
 * @author HIPAA_READY
 */
public class ProcessHacStg835Load {
    static final Logger summaryLOGGER = LoggerFactory.getLogger(ProcessHacStg835Load.class);
    static final Logger traceLOGGER = LoggerFactory.getLogger("trace." +
            ProcessHacStg835Load.class);

    // Constants
    final private static BigDecimal zeroBD = new BigDecimal("0");
    private String loadRec = ""; // stores a single line of the input data file
    private int cntHdr = 0; // Count of the header lines
    private int cntClp = 0; // Count of the claim payment lines
    private String vHdrKey = "";
    private String vClpKey = "";
    private String vSvcKey = "";
    private String vProviderId = "";
    private String transNumber = "";

    /**
     * This is the main entry point method of this class. It controls the
     * overall transaction related to this process -- The method opens the data
     * file reads the data line by line identifies each line type by the first
     * token in the line calls the appropriate helper method interpret the tokens read
     * The helper maps the data to an object and stores the object in the database.
     *
     * After all the data has been read - it commits the data in one bulk
     * commit.
     *
     * @param fileName
     *            - The full path and name of the data file
     * @throws FileNotFoundException
     *             - thrown when the data file can not be located
     * @throws IOException
     *             - thrown if there are any problems that occur while reading
     *             the data file
     * @throws DbTransactionEnderException
     *             - explictly thrown to stop the transaction
     */
    protected void processLoad(final String fileName, EntityManager dbEM)
        throws FileNotFoundException, IOException, DbTransactionEnderException {
    	
    	boolean DEBUG = true;
    	
        // Local variables
        String recType = "";
        BufferedReader in = null;

      
        if (!new File(fileName).exists()) {
        	
        	if(DEBUG) {
        		System.out.println("ProcessHacStg835Load.processLoad() input file NOT FOUND: " + fileName);
        	}
        	
        	
            final String message = "Can't find input data file:" + fileName;
            
            
            
            throw new FileNotFoundException(message);
        }

        try {
            // Open the input file for reading
            in = new BufferedReader(new FileReader(fileName));

            dbEM.getTransaction().begin();

            // This loop reads each line in the input file, determines the type
            // of line based on the first token in the line and then calls the
            // appropriate helper method to map the data in the line to an
            // object
            // so the object/data may be persisted to the database
            while ((loadRec = in.readLine()) != null) {
                traceLOGGER.debug(loadRec);

                // Split the line into multiple tokens, as separated by '^'
                final String[] token = loadRec.split("\\^", -1);

                // The first Field identifies the type of line in the file
                // i.e. each line has a different meaning that is indicated by
                // the
                // For example, a line starting with HDR contains data that is
                // part of the header
                recType = token[0];
                
                if(DEBUG) {
                	System.out.println("processLoad() - recType is: " + recType);
                }

                // The transaction number can be found in the third position
                transNumber = token[2];

                // Call a different loading method for each type of line
                if (recType.equalsIgnoreCase("HDR")) {
                    //Load a Transaction Set Header.
                    loadStg835Hdr(token, dbEM);
                    System.out.println("Stg835Hdr loaded");
                } else if (recType.equalsIgnoreCase("BPR")) {
                    //Load a billing provider line - Financial Information
                    loadStg835HdrBpr(token, dbEM);
                    System.out.println("Stg835HdrBpr loaded");
                } else if (recType.equalsIgnoreCase("TRN")) {
                    //Load the Reassociation Trace Number line
                    loadStg835HdrTrn(token, dbEM);
                    System.out.println("Stg835HdrTrn loaded");
                } else if (recType.equalsIgnoreCase("CLP")) {
                    // There are two different layouts for claim payment records
                    if (token[3].equalsIgnoreCase("PHARM CLAIM")) {
                        //Load a pharm payment line
                        loadStg835PharmClp(token, dbEM);
                        System.out.println("Stg835PharmClp loaded");
                    } else {
                        //load a normal claim payment line
                        loadStg835Clp(token, dbEM);
                        System.out.println("Stg835Clp loaded");
                    }
                } else if (recType.equalsIgnoreCase("CLPNAME")) {
                    loadStg835PharmClpName(token, dbEM);
                    System.out.println("Stg835PharmClpName loaded");
                } else if (recType.equalsIgnoreCase("CLPCAS")) {
                     //Load a Claim Payment Claim Adjustment line
                    loadStg835ClpCas(token, dbEM);
                    System.out.println("Stg835ClpCas loaded");
                } else if (recType.equalsIgnoreCase("SVC")) {
                    //Load a Service Payment Information line
                    loadStg835Svc(token, dbEM);
                    System.out.println("Stg835Svc loaded");
                } else if (recType.equalsIgnoreCase("SVCCAS")) {
                    //Load a service payment claim adjustment line
                    loadStg835SvcCas(token, dbEM);
                    System.out.println("Stg835SvcCas loaded");
                } else if (recType.equalsIgnoreCase("SVCLQ")) {
                    //Load a service payment Health Care Remark Codes line
                    loadStg835SvcLq(token, dbEM);
                    System.out.println("Stg835SvcLq loaded");
                } else if (recType.equalsIgnoreCase("PLB")) {
                    //Load a Provider Adjustment line
                    loadStg835Plb(token, dbEM);
                    System.out.println("Stg835Plb loaded");
                }
            }

            System.out.println("No. of Transaction Sets Loaded : " + cntHdr);
            traceLOGGER.info("No. of Claims Loaded : " + cntClp);
            summaryLOGGER.info("No. of Transaction Sets Loaded : " + cntHdr);
            summaryLOGGER.info("No. of Claims Loaded : " + cntClp);
            System.out.println("Committing the database transaction");
            dbEM.getTransaction().commit();
            System.out.println("Commit successful");

            // // Remove the input data file, which would have been archived
            // previously
            in.close();
            FileUtils.deleteFile(fileName);

            // If any unexpected exception was thrown, rollback and log it, and
            // rethrow
        } catch (Throwable ex) {
        	System.out.println("processLoad() - error at transaction#: " + transNumber + "rolling back transaction");
            traceLOGGER.error("Rolling back the database transaction " +
                "when at data trans number " + transNumber);
            
            System.out.println("message: " + ex.getMessage());
            
            

            if ((dbEM != null) && (dbEM.getTransaction() != null) &&
                    dbEM.getTransaction().isActive()) {
                dbEM.getTransaction().rollback();
            }

            System.out.println("Rollback successful");
            throw new DbTransactionEnderException("Rolled Back the Transaction",
                ex);
        } finally {
            if (in != null) {
                in.close();
            }
        }
    }

    /**
     * This method is passed a sting of tokens that has been read from a staging data file.
     * The method expects the tokens to correspond to the Transaction Set Header -
     * identified by the first token which should be "HDR".  It takes the tokens
     * and maps them into the Stg835Hdr object so that the Transaction Set Header data
     * may be stored in the database. The method uses the entity manager to store
     * the data in the database.  The entity manager in turn depends on the mappings
     * in META-INF/stg835LoadMappings.xml to store the objects in the database.
     *
     * @param tokens - the line that represents the Transaction Set Header data.
     * @param dbEM - the entity manager used to persist the data to the database.
     * @throws SQLException
     */

    // Load a line that starts with "HDR" -a Transaction Set Header.
    private void loadStg835Hdr(final String[] tokens, EntityManager dbEM)
        throws SQLException {
        String vClshId = "";

        // TWP: Make this configurable - performance code
        if ((cntHdr % 100) == 0) {
            dbEM.flush();
            dbEM.clear();
        }

        System.out.println("Processing header #: " + transNumber + " at: " +
            DateUtil.getCurrentTimeFormatted("HH:mm:ss:SSS"));

        // Create a new object to map the data to
        final Stg835Hdr crStg835Hdr = new Stg835Hdr();

        // Generate a new primary key for the object/data so that
        // it can be stored in the database
        vHdrKey = UUID.randomUUID().toString().toUpperCase();
        crStg835Hdr.setHdrKey(vHdrKey);

        // Assign the other fields
        crStg835Hdr.setFileNumber(tokens[1]);
        crStg835Hdr.setSt02txnCtlNbr(tokens[2]);
        crStg835Hdr.setFacilityId(tokens[5]);
        crStg835Hdr.setPayerId(tokens[6]);
        vProviderId = tokens[7];
        crStg835Hdr.setProviderId(vProviderId);
        vClshId = tokens[8];
        crStg835Hdr.setClearinghouseId(vClshId);
        crStg835Hdr.setUsageIndicator(tokens[9]);
        crStg835Hdr.setStseStatus(tokens[10]);

        // added provider NPI field to HDR record not all records will have it
        // check for length to prevent errors. Currently only non-
        // pharmacy claims will return the NPI sent in by the 837.
        String vNpi = null;

        if ((tokens.length >= 13) && !"".equals(tokens[12])) {
            vNpi = tokens[12];
            crStg835Hdr.setNpi(vNpi);
        }

        // Set the Provider Name in STG_835_HDR table
        // If the clearinghouse is Medical Matrix, the provider's name will
        // be in the 12th field
        if (vClshId.equalsIgnoreCase("403400000")) {
            crStg835Hdr.setProviderName(tokens[11]);

            // Otherwise, we get it from the EntityName table
        } else if (vClshId.equalsIgnoreCase("752578509")) {
            crStg835Hdr.setProviderName("SXC HEALTH SOLUTIONS");
        } else {
            // the NM109 field should be the NPI unless the provider cannot
            // get an NPI and then this will be the TIN
            String provid = "";

            if (vNpi == null) {
                provid = vProviderId;
            } else {
                provid = vNpi;
            }

            // Get a row from the EntityName table for this provider id
            // Method getEntityNameVo might return many, then will take the
            // first
            final String vWhereCls = "NM109ID_CD = '" + provid + "' " +
                "AND CLAIM_KEY IS NULL" + " AND ROWNUM = 1";
            
            // fails here.
  
            final EntityName rowEntityName = getEntityNameVo(vWhereCls, dbEM);
           

            // If one was found, use it, otherwise use "Unknown"
            if (rowEntityName == null) {
                crStg835Hdr.setProviderName("Unknown");
            } else {
                crStg835Hdr.setProviderName(rowEntityName.getNm103lastNm());
            }
        }

        System.out.println("Stg835Hdr fieldvalues {}" + crStg835Hdr.toString());
        
        try {
        		dbEM.persist(crStg835Hdr);
        }catch(Exception e) {
        	System.out.println("loadStg835Hdr() error: " + e.getMessage());
        	e.printStackTrace();
        }
        
        
        cntHdr++;
    }

    /**
     * This method is passed a sting of tokens that has been read from a staging data file.
     * The method expects the tokens to correspond to a billing provider line -
     * identified by the first token which should be "BPR".  It takes the tokens
     * and maps them into the Stg835HdrBpr object so that the billing provider data
     * may be stored in the database. The method uses the entity manager to store
     * the data in the database.  The entity manager in turn depends on the mappings
     * in META-INF/stg835LoadMappings.xml to store the objects in the database.
     *
     * @param tokens - the line that represents the billing provider data.
     * @param dbEM - the entity manager used to persist the data to the database.
     * @throws SQLException
     */

    // Load a line that starts with "BPR" - billing provider line
    private void loadStg835HdrBpr(final String[] tokens, EntityManager dbEM)
        throws SQLException {
        System.out.println("Processing Bpr");

        // Create a new row for table stg_835_hdr_bpr
        final Stg835HdrBpr crStg835HdrBpr = new Stg835HdrBpr();

        // Set some easy fields
        crStg835HdrBpr.setHdrBprKey((UUID.randomUUID().toString().toUpperCase()));
        crStg835HdrBpr.setHdrKey(vHdrKey);
        crStg835HdrBpr.setFileNumber(tokens[1]);
        crStg835HdrBpr.setSt02txnCtlNbr(tokens[2]);

        if (!((tokens[3].equals("")) || (tokens[3] == null))) {
            crStg835HdrBpr.setBpr02pymtAmt(new BigDecimal(tokens[3]));
        }

        crStg835HdrBpr.setBpr04pymtMthdCd(tokens[4]);

        if (!((tokens[5].equals("")) || (tokens[5] == null))) {
            crStg835HdrBpr.setBpr07pyrDfiId(tokens[5]);
        }

        if (!((tokens[6].equals("")) || (tokens[6] == null))) {
            crStg835HdrBpr.setBpr09pyrAcctNbr(tokens[6]);
        }

        if (!((tokens[7].equals("")) || (tokens[7] == null))) {
            crStg835HdrBpr.setBpr10pyrId(tokens[7]);
        }

        //TWP: Truncating to handle bug
        if (!((tokens[8].equals("")) || (tokens[8] == null))) {
            if (tokens[8].length() > 12) {
                crStg835HdrBpr.setBpr13rcvDfiId(tokens[8].substring(0, 12));
            } else {
                crStg835HdrBpr.setBpr13rcvDfiId(tokens[8]);
            }
        }

        if (!((tokens[9].equals("")) || (tokens[9] == null))) {
            crStg835HdrBpr.setBpr15rcvAcctNbr(tokens[9]);
        }

        if (!((tokens[10].equals("")) || (tokens[10] == null))) {
            crStg835HdrBpr.setBpr16ckeftDt(tokens[10]);
        }

        traceLOGGER.debug("Stg835HdrBpr fieldvalues {}", crStg835HdrBpr);
        dbEM.persist(crStg835HdrBpr);
    }

    /**
     * This method is passed a sting of tokens that has been read from a staging data file.
     * The method expects the tokens to correspond to the Reassociation Trace Number line -
     * identified by the first token which should be "TRN".  It takes the tokens
     * and maps them into the Stg835HdrTrn object so that the Reassociation Trace Number data
     * may be stored in the database. The method uses the entity manager to store
     * the data in the database.  The entity manager in turn depends on the mappings
     * in META-INF/stg835LoadMappings.xml to store the objects in the database.
     *
     * @param tokens - the line that represents the Reassociation Trace Number data.
     * @param dbEM - the entity manager used to persist the data to the database.
     * @throws SQLException
     */

    // Load a line that starts with "TRN" - Load the Reassociation Trace Number line, with a trace number for this
    // transaction
    private void loadStg835HdrTrn(final String[] tokens, EntityManager dbEM)
        throws SQLException {
        System.out.println("Processing Trn");

        // Create a new row for table stg_835_hdr_trn
        final Stg835HdrTrn crStg835HdrTrn = new Stg835HdrTrn();

        // Set some fields in it
        crStg835HdrTrn.setHdrTrnKey((UUID.randomUUID().toString().toUpperCase()));
        crStg835HdrTrn.setHdrKey(vHdrKey);
        crStg835HdrTrn.setFileNumber(tokens[1]);
        crStg835HdrTrn.setSt02txnCtlNbr(tokens[2]);

        if (!((tokens[3].equals("")) || (tokens[3] == null))) {
            crStg835HdrTrn.setTrn02ckeftTrcNbr(tokens[3]);
        }

        if (!((tokens[4].equals("")) || (tokens[4] == null))) {
            crStg835HdrTrn.setTrn03pyrEftNbr(tokens[4]);
        }

        if (!((tokens[5].equals("")) || (tokens[5] == null))) {
            crStg835HdrTrn.setProdDate(tokens[5]);
        }

        traceLOGGER.debug("Stg835HdrTrn fieldvalues {}", crStg835HdrTrn);
        // Insert the new row
        dbEM.persist(crStg835HdrTrn);
    }

    /**
     * This method is passed a sting of tokens that has been read from a staging data file.
     * The method expects the tokens to correspond to claim payment information -
     * identified by the first token which should be "CLP".  It takes the tokens
     * and maps them into the Stg835Clp object so that the claim payment information
     * may be stored in the database. The method uses the entity manager to store
     * the data in the database.  The entity manager in turn depends on the mappings
     * in META-INF/stg835LoadMappings.xml to store the objects in the database.
     *
     * @param tokens - the line that represents the  claim payment information.
     * @param dbEM - the entity manager used to persist the data to the database.
     * @throws SQLException
     */

    // Load a line that starts with "CLP" - a claim payment
    private void loadStg835Clp(final String[] tokens, EntityManager dbEM)
        throws SQLException {
        System.out.println("Processing Clp");

        // Create a new row for table stg_835_clp
        final Stg835Clp crStg835Clp = new Stg835Clp();

        // Set some fields
        vClpKey = (UUID.randomUUID().toString().toUpperCase());

        String vClaimKey = tokens[3];

        // If the claim to which this payment corresponds was old enough to have
        // been migrated, get the new, 36-character id out of a migration table
        if (vClaimKey.length() < 30) {
            vClaimKey = vClaimKey.trim();

            Query getHdrReposMigrationQuery = dbEM.createQuery(
                    "SELECT hdr FROM gov.va.med.domain.erepos.HdrReposMigration hdr WHERE hdr.hdrClaimId=?1");
            getHdrReposMigrationQuery.setParameter(1, vClaimKey);

            List hdrReposMigrationList = getHdrReposMigrationQuery.getResultList();
            Iterator hdrReposMigrationIterator = hdrReposMigrationList.iterator();

            if (hdrReposMigrationIterator.hasNext()) {
                final HdrReposMigration rowHdrReposMigration = (HdrReposMigration) hdrReposMigrationIterator.next();
                vClaimKey = rowHdrReposMigration.getClaimKey();
            }
        }

        // Set some more fields
        crStg835Clp.setClpKey(vClpKey);
        crStg835Clp.setHdrKey(vHdrKey);
        crStg835Clp.setFileNumber(tokens[1]);
        crStg835Clp.setSt02txnCtlNbr(tokens[2]);
        crStg835Clp.setClaimKey(vClaimKey);
        crStg835Clp.setClp01ptCtlNbr(tokens[8]);

        if (!((tokens[9].equals("")) || (tokens[9] == null))) {
            crStg835Clp.setClp03clmChrgAmt(new BigDecimal(tokens[9]));
        }

        if (!((tokens[10].equals("")) || (tokens[10] == null))) {
            crStg835Clp.setClp04pymtAmt(new BigDecimal(tokens[10]));
        }

        if (!((tokens[11].equals("")) || (tokens[11] == null))) {
            crStg835Clp.setClp05ptRespAmt(new BigDecimal(tokens[11]));
        }

        crStg835Clp.setClp07pyrClmNbr(tokens[12]);

        if (!((tokens[13].equals("")) || (tokens[13] == null))) {
            crStg835Clp.setClp11drgCd(tokens[13]);
        }

        if (!((tokens[14].equals("")) || (tokens[14] == null))) {
            crStg835Clp.setClp12drgWt(new BigDecimal(tokens[14]));
        }

        if (!((tokens[15].equals("")) || (tokens[15] == null))) {
            crStg835Clp.setClaimDate(tokens[15]);
        }

        crStg835Clp.setClpStatus(tokens[16]);
        crStg835Clp.setClp02clmStatusCd(tokens[17]);
        // crStg835Clp.setClp08fclyTypeCd("11");
        // crStg835Clp.setClp09clmFrqCd("2");
        crStg835Clp.setInterestAmount(zeroBD);

        final Stg835ClpName voStg835ClpName = new Stg835ClpName();
        String vFileNbr = "";
        String vSt02Nbr = "";
        String vSbsrId = "";

        Query getClaimQuery = dbEM.createQuery(
                "SELECT c FROM gov.va.med.domain.erepos.Claim c WHERE c.claimKey=?1");
        getClaimQuery.setParameter(1, vClaimKey);

        List claimList = getClaimQuery.getResultList();
        Iterator claimIterator = claimList.iterator();

        if (claimIterator.hasNext()) {
            final Claim rowClaim = (Claim) claimIterator.next();
            vFileNbr = rowClaim.getFileNumber();
            vSt02Nbr = rowClaim.getSt02txnCtlNbr();
            vSbsrId = rowClaim.getSubscriberId();
            crStg835Clp.setClp08fclyTypeCd(rowClaim.getClm05fclyTypeCd());

            // TWP: Test This we may want character 0 and not null in the db???
            if (rowClaim.getClm05clmFrqCd() != null) {
                crStg835Clp.setClp09clmFrqCd(rowClaim.getClm05clmFrqCd()
                                                     .toString());
            }
        }

        traceLOGGER.debug("Stg835Clp fieldvalues {}", crStg835Clp);
        // Insert the new claim payment record
        dbEM.persist(crStg835Clp);
        cntClp++;

        // Look for a corresponding Patient Record in table Entity Name
        // TODO: We have no index covering these four columns.
        String vWhereCls = "FILE_NUMBER = '" + vFileNbr + "' " +
            "AND ST02TXN_CTL_NBR = '" + vSt02Nbr + "' " +
            "AND LOOP_ID = '2010BA' " + "AND NM109ID_CD = '" + vSbsrId + "' ";
        EntityName rowEntityName = getEntityNameVo(vWhereCls, dbEM);

        // If one was found, create a new record in stg_835_clp_name for it
        if (rowEntityName != null) {
            final Stg835ClpName crStg835ClpName = new Stg835ClpName();
            crStg835ClpName.setClpNameKey((UUID.randomUUID().toString()
                                               .toUpperCase()));
            crStg835ClpName.setClpKey(vClpKey);
            crStg835ClpName.setNm101entyIdCd("QC");
            crStg835ClpName.setNm102entyQual(ConversionUtil.stringToSingleCharacter(
                    "1"));
            crStg835ClpName.setNm103lastNm(rowEntityName.getNm103lastNm());
            crStg835ClpName.setNm104fstNm(rowEntityName.getNm104fstNm());
            crStg835ClpName.setNm105midNm(rowEntityName.getNm105midNm());
            crStg835ClpName.setNm108idCdQual("34");
            crStg835ClpName.setNm109entyId(rowEntityName.getNm109idCd());
            traceLOGGER.debug("Stg835ClpName fieldvalues {}", crStg835ClpName);
            dbEM.persist(crStg835ClpName);
        }

        // Look for a corresponding Service Provider Record in table Entity Name
        // TODO: There is an index for this query, ENTITY_NAME_IDX9. Will it
        // still be used with these IN clauses?
        vWhereCls = "CLAIM_KEY = '" + vClaimKey + "' " +
            "AND LOOP_ID IN ('2310A', '2310B') " +
            "AND NM101ENTY_ID_CD IN ('71','82') ";
        rowEntityName = getEntityNameVo(vWhereCls, dbEM);

        // If one was found, create a new record in stg_835_clp_name for it
        if (rowEntityName != null) {
            final Stg835ClpName crStg835ClpName = new Stg835ClpName();
            crStg835ClpName.setClpNameKey((UUID.randomUUID().toString()
                                               .toUpperCase()));
            crStg835ClpName.setClpKey(vClpKey);
            crStg835ClpName.setNm101entyIdCd("82");
            crStg835ClpName.setNm102entyQual(ConversionUtil.stringToSingleCharacter(
                    "2"));
            crStg835ClpName.setNm103lastNm(rowEntityName.getNm103lastNm());

            final String vIdCd = rowEntityName.getNm109idCd();
            final String vIdCdQual = rowEntityName.getNm108idCdQual();

            if (vIdCdQual.equalsIgnoreCase("XX")) {
                crStg835ClpName.setNm108idCdQual("XX");
            } else {
                crStg835ClpName.setNm108idCdQual("FI");
            }

            crStg835ClpName.setNm109entyId(vIdCd);
            traceLOGGER.debug("Stg835ClpName fieldvalues {}", crStg835ClpName);
            dbEM.persist(crStg835ClpName);
        }
    }

    /**
     * This method is passed a sting of tokens that has been read from a staging data file.
     * The method expects the tokens to correspond to pharmacy claim payment information -
     * identified by the first token which should be "PHARM CLAIM".  It takes the tokens
     * and maps them into the Stg835Clp object so that the claim payment information
     * may be stored in the database. The method uses the entity manager to store
     * the data in the database.  The entity manager in turn depends on the mappings
     * in META-INF/stg835LoadMappings.xml to store the objects in the database.
     *
     * @param tokens - the line that represents the  pharmacy claim payment information.
     * @param dbEM - the entity manager used to persist the data to the database.
     * @throws SQLException
     */

    //Load a line that starts with "PHARM CLAIM" - a pharmacy claim payment
    private void loadStg835PharmClp(final String[] tokens, EntityManager dbEM)
        throws SQLException {
        System.out.println("Processing Pharm Clp");

        final Stg835Clp crStg835Clp = new Stg835Clp();
        vClpKey = (UUID.randomUUID().toString().toUpperCase());

        final String vClaimKey = tokens[3];

        crStg835Clp.setClpKey(vClpKey);
        crStg835Clp.setHdrKey(vHdrKey);
        crStg835Clp.setFileNumber(tokens[1]);
        crStg835Clp.setSt02txnCtlNbr(tokens[2]);
        crStg835Clp.setClaimKey(vClaimKey);
        crStg835Clp.setClp01ptCtlNbr(tokens[8]);

        if (!((tokens[9].equals("")) || (tokens[9] == null))) {
            crStg835Clp.setClp03clmChrgAmt(new BigDecimal(tokens[9]));
        }

        if (!((tokens[10].equals("")) || (tokens[10] == null))) {
            crStg835Clp.setClp04pymtAmt(new BigDecimal(tokens[10]));
        }

        if (!((tokens[11].equals("")) || (tokens[11] == null))) {
            crStg835Clp.setClp05ptRespAmt(new BigDecimal(tokens[11]));
        }

        crStg835Clp.setClp07pyrClmNbr(tokens[12]);

        if (!((tokens[13].equals("")) || (tokens[13] == null))) {
            crStg835Clp.setClp11drgCd(tokens[13]);
        }

        if (!((tokens[14].equals("")) || (tokens[14] == null))) {
            crStg835Clp.setClp12drgWt(new BigDecimal(tokens[14]));
        }

        if (!((tokens[15].equals("")) || (tokens[15] == null))) {
            crStg835Clp.setClaimDate(tokens[15]);
        }

        crStg835Clp.setClpStatus(tokens[16]);
        crStg835Clp.setClp02clmStatusCd(tokens[17]);
        // crStg835Clp.setClp08fclyTypeCd("11");
        // crStg835Clp.setClp09clmFrqCd("2");
        crStg835Clp.setInterestAmount(zeroBD);
        traceLOGGER.debug("Stg835PharmClp fieldvalues {}", crStg835Clp);
        dbEM.persist(crStg835Clp);
        cntClp++;
    }

    private void loadStg835PharmClpName(final String[] tokens,
        EntityManager dbEM) throws SQLException {
        final Stg835ClpName crStg835ClpName = new Stg835ClpName();
        crStg835ClpName.setClpNameKey((UUID.randomUUID().toString().toUpperCase()));
        crStg835ClpName.setClpKey(vClpKey);
        crStg835ClpName.setNm101entyIdCd("QC");
        crStg835ClpName.setNm102entyQual(ConversionUtil.stringToSingleCharacter(
                "1"));
        crStg835ClpName.setNm103lastNm(tokens[4]);
        crStg835ClpName.setNm104fstNm(tokens[5]);
        crStg835ClpName.setNm108idCdQual("34");
        crStg835ClpName.setNm109entyId(tokens[6]);
        traceLOGGER.debug("Stg835PharmClpName fieldvalues {}", crStg835ClpName);
        dbEM.persist(crStg835ClpName);
    }

    /**
     * This method is passed a sting of tokens that has been read from a staging data file.
     * The method expects the tokens to correspond to a Claim Adjustment related to a Claim Payment -
     * identified by the first token which should be "CLPCAS".  It takes the tokens
     * and maps them into the Stg835ClpCas object so that the Claim Adjustment data
     * may be stored in the database. The method uses the entity manager to store
     * the data in the database.  The entity manager in turn depends on the mappings
     * in META-INF/stg835LoadMappings.xml to store the objects in the database.
     *
     * @param tokens - the line that represents the Claim Adjustment.
     * @param dbEM - the entity manager used to persist the data to the database.
     * @throws SQLException
     */

    //Load a Claim Payment Claim Adjustment line
    private void loadStg835ClpCas(final String[] tokens, EntityManager dbEM)
        throws SQLException {
        System.out.println("Processing ClpCas");

        final Stg835ClpCas crStg835ClpCas = new Stg835ClpCas();

        crStg835ClpCas.setClpCasKey((UUID.randomUUID().toString().toUpperCase()));
        crStg835ClpCas.setClpKey(vClpKey);
        crStg835ClpCas.setClp07pyrClmNbr(tokens[4]);
        crStg835ClpCas.setCas01clmAdjGrp(tokens[5]);
        crStg835ClpCas.setCas02clmAdjRsn1(tokens[6]);

        
        // fails here............. expecting a number 
        try {
		        System.out.println("sec 7 is: " + tokens[7]);
		        if (!((tokens[7].equals("")) || (tokens[7] == null))) {
		            crStg835ClpCas.setCas03adjAmt1(new BigDecimal(tokens[7]));
		        }
		
		        System.out.println("sec 8 is: " + tokens[8]);
		        if (!((tokens[8].equals("")) || (tokens[8] == null))) {
		            crStg835ClpCas.setCas05clmAdjRsn2(tokens[8]);
		        }
		        // fails here with HDR type input
		        System.out.println("sec 9 is: " + tokens[9]);
		        if (!((tokens[9].equals("")) || (tokens[9] == null))) {
		            crStg835ClpCas.setCas06adjAmt2(new BigDecimal(tokens[9]));
		        }
		
		        System.out.println("sec 10 is: " + tokens[10]);
		        if (!((tokens[10].equals("")) || (tokens[10] == null))) {
		            crStg835ClpCas.setCas08clmAdjRsn3(tokens[10]);
		        }
		
		        System.out.println("sec 11 is: " + tokens[11]);
		        if (!((tokens[11].equals("")) || (tokens[11] == null))) {
		            crStg835ClpCas.setCas09adjAmt3(new BigDecimal(tokens[11]));
		        }
		
		        System.out.println("sec 12 is: " + tokens[12]);
		        if (!((tokens[12].equals("")) || (tokens[12] == null))) {
		            crStg835ClpCas.setCas11clmAdjRsn4(tokens[12]);
		        }
		
		        System.out.println("sec 13 is: " + tokens[13]);
		        if (!((tokens[13].equals("")) || (tokens[13] == null))) {
		            crStg835ClpCas.setCas12adjAmt4(new BigDecimal(tokens[13]));
		        }
		
		        System.out.println("sec 14 is: " + tokens[14]);
		        if (!((tokens[14].equals("")) || (tokens[14] == null))) {
		            crStg835ClpCas.setCas14clmAdjRsn5(tokens[14]);
		        }
		
		        System.out.println("sec 15 is: " + tokens[15]);
		        if (!((tokens[15].equals("")) || (tokens[15] == null))) {
		            crStg835ClpCas.setCas15adjAmt5(new BigDecimal(tokens[15]));
		        }
		
		        System.out.println("sec 16 is: " + tokens[16]);
		        if (!((tokens[16].equals("")) || (tokens[16] == null))) {
		            crStg835ClpCas.setCas17clmAdjRsn6(tokens[16]);
		        }
		
		        System.out.println("sec 17 is: " + tokens[17]);
		        if (!((tokens[17].equals("")) || (tokens[17] == null))) {
		            crStg835ClpCas.setCas18adjAmt6(new BigDecimal(tokens[17]));
		        }
		        
        } catch(Exception e) {
        	System.out.println("Stg835ClpCas fieldvalues {}" + crStg835ClpCas);
        	System.out.println("ERROR PARSING CLP_CAS: " + e.getMessage());
        	e.printStackTrace();
        }
        

        traceLOGGER.debug("Stg835ClpCas fieldvalues {}", crStg835ClpCas);
        
        dbEM.persist(crStg835ClpCas);
        
        traceLOGGER.debug("Stg835ClpCas fieldvalues {} WRITTEN");
        
     }	// END loadStg835ClpCas()
    
    
    
    
    

    /**
     * This method is passed a sting of tokens that has been read from a staging data file.
     * The method expects the tokens to correspond to Service Payment Information -
     * identified by the first token which should be "SVC".  It takes the tokens
     * and maps them into the Stg835Svc object so that the Service Payment data
     * may be stored in the database. The method uses the entity manager to store
     * the data in the database.  The entity manager in turn depends on the mappings
     * in META-INF/stg835LoadMappings.xml to store the objects in the database.
     *
     * @param tokens - the line that represents the Service Payment Information.
     * @param dbEM - the entity manager used to persist the data to the database.
     * @throws SQLException
     */

    //Load a Service Payment Information line
    private void loadStg835Svc(final String[] tokens, EntityManager dbEM)
        throws SQLException {
        System.out.println("Processing Svc");

        final Stg835Svc crStg835Svc = new Stg835Svc();
        vSvcKey = (UUID.randomUUID().toString().toUpperCase());
        crStg835Svc.setSvcKey(vSvcKey);
        crStg835Svc.setClpKey(vClpKey);

        if (!((tokens[6].equals("")) || (tokens[6] == null))) {
            crStg835Svc.setSvc01procCd(tokens[6]);
        }

        if (!((tokens[7].equals("")) || (tokens[7] == null))) {
            crStg835Svc.setSvc01svcIdQual(tokens[7]);
        }

        if (!((tokens[8].equals("")) || (tokens[8] == null))) {
            crStg835Svc.setSvc01procMod1(tokens[8]);
        }

        if (!((tokens[9].equals("")) || (tokens[9] == null))) {
            crStg835Svc.setSvc01procMod2(tokens[9]);
        }

        if (!((tokens[10].equals("")) || (tokens[10] == null))) {
            crStg835Svc.setSvc01procMod3(tokens[10]);
        }

        if (!((tokens[11].equals("")) || (tokens[11] == null))) {
            crStg835Svc.setSvc01procMod4(tokens[11]);
        }

        if (!((tokens[12].equals("")) || (tokens[12] == null))) {
            crStg835Svc.setSvc01procCdDesc(tokens[12]);
        }

        if (!((tokens[13].equals("")) || (tokens[13] == null))) {
            crStg835Svc.setSvc02chrgAmt(new BigDecimal(tokens[13]));
        }

        if (!((tokens[14].equals("")) || (tokens[14] == null))) {
            crStg835Svc.setSvc03prvdPmt(new BigDecimal(tokens[14]));
        }

        if (!((tokens[15].equals("")) || (tokens[15] == null))) {
            crStg835Svc.setSvc04nubcRevCd(tokens[15]);
        }

        if (!((tokens[16].equals("")) || (tokens[16] == null))) {
            crStg835Svc.setSvc05unitSvcPaid(new Integer(tokens[16]));
        }

        if (!((tokens[17].equals("")) || (tokens[17] == null))) {
            crStg835Svc.setSvc07orgnunitSvc(new Integer(tokens[17]));
        }

        if (!((tokens[18].equals("")) || (tokens[18] == null))) {
            crStg835Svc.setDtm01ddtmQual(tokens[18]);
        }

        if (!((tokens[19].equals("")) || (tokens[19] == null))) {
            crStg835Svc.setDtm01dtFld(tokens[19]);
        }

        if (!((tokens[20].equals("")) || (tokens[20] == null))) {
            crStg835Svc.setDtm02ddtmQual(tokens[20]);
        }

        if (!((tokens[21].equals("")) || (tokens[21] == null))) {
            crStg835Svc.setDtm02dtFld(tokens[21]);
        }

        if (!((tokens[22].equals("")) || (tokens[22] == null))) {
            crStg835Svc.setRef02adtlId(tokens[22]);
        }

        //New data elements added for SLA 835
        traceLOGGER.debug("Number of tokens: " + tokens.length);

        if (tokens.length > 23) {
            if (!((tokens[23].equals("")) || (tokens[23] == null))) {
                crStg835Svc.setSvcRefRenderingProvId(tokens[23]);
            }
        }

        if (tokens.length > 24) {
            if (!((tokens[24].equals("")) || (tokens[24] == null))) {
                crStg835Svc.setSvcAllowedAmt(new BigDecimal(tokens[24]));
            }
        }

        crStg835Svc.setSvcStatus("NEW");
        traceLOGGER.debug("Stg835Svc fieldvalues {}", crStg835Svc);
        dbEM.persist(crStg835Svc);
    }

    /**
     * This method is passed a sting of tokens that has been read from a staging data file.
     * The method expects the tokens to correspond to a Claim Adjustment related to Service Payment Information-
     * identified by the first token which should be "SVCCAS".  It takes the tokens
     * and maps them into the Stg835SvcCas object so that the Claim Adjustment data
     * may be stored in the database. The method uses the entity manager to store
     * the data in the database.  The entity manager in turn depends on the mappings
     * in META-INF/stg835LoadMappings.xml to store the objects in the database.
     *
     * @param tokens - the line that represents the Claim Adjustment.
     * @param dbEM - the entity manager used to persist the data to the database.
     * @throws SQLException
     */

    //Load a Service Payment Information Claim Adjustment line
    private void loadStg835SvcCas(final String[] tokens, EntityManager dbEM)
        throws SQLException {
        System.out.println("Processing SvcCas");

        final Stg835SvcCas crStg835SvcCas = new Stg835SvcCas();
        crStg835SvcCas.setSvcCasKey((UUID.randomUUID().toString().toUpperCase()));
        crStg835SvcCas.setSvcKey(vSvcKey);
        crStg835SvcCas.setClpKey(vClpKey);
        crStg835SvcCas.setCas01clmAdjGrp(tokens[6]);
        crStg835SvcCas.setCas02clmAdjRsn1(tokens[7]);

        if (!((tokens[8].equals("")) || (tokens[8] == null))) {
            crStg835SvcCas.setCas03adjAmt1(new BigDecimal(tokens[8]));
        }

        if (!((tokens[9].equals("")) || (tokens[9] == null))) {
            crStg835SvcCas.setCas05clmAdjRsn2(tokens[9]);
        }

        if (!((tokens[10].equals("")) || (tokens[10] == null))) {
            crStg835SvcCas.setCas06adjAmt2(new BigDecimal(tokens[10]));
        }

        if (!((tokens[11].equals("")) || (tokens[11] == null))) {
            crStg835SvcCas.setCas08clmAdjRsn3(tokens[11]);
        }

        if (!((tokens[12].equals("")) || (tokens[12] == null))) {
            crStg835SvcCas.setCas09adjAmt3(new BigDecimal(tokens[12]));
        }

        if (!((tokens[13].equals("")) || (tokens[13] == null))) {
            crStg835SvcCas.setCas11clmAdjRsn4(tokens[13]);
        }

        if (!((tokens[14].equals("")) || (tokens[14] == null))) {
            crStg835SvcCas.setCas12adjAmt4(new BigDecimal(tokens[14]));
        }

        if (!((tokens[15].equals("")) || (tokens[15] == null))) {
            crStg835SvcCas.setCas14clmAdjRsn5(tokens[15]);
        }

        if (!((tokens[16].equals("")) || (tokens[16] == null))) {
            crStg835SvcCas.setCas15adjAmt5(new BigDecimal(tokens[16]));
        }

        if (!((tokens[17].equals("")) || (tokens[17] == null))) {
            crStg835SvcCas.setCas17clmAdjRsn6(tokens[17]);
        }

        if (!((tokens[18].equals("")) || (tokens[18] == null))) {
            crStg835SvcCas.setCas18adjAmt6(new BigDecimal(tokens[18]));
        }

        traceLOGGER.debug("Stg835SvcCas fieldvalues {}", crStg835SvcCas);
        dbEM.persist(crStg835SvcCas);
    }

    /**
     * This method is passed a sting of tokens that has been read from a staging data file.
     * The method expects the tokens to correspond to a line of data representing Health Care Remark Codes-
     * identified by the first token which should be "SVCLQ".  It takes the tokens
     * and maps them into the Stg835SvcLq object so that the Health Care Remark Codes data
     * may be stored in the database. The method uses the entity manager to store
     * the data in the database.  The entity manager in turn depends on the mappings
     * in META-INF/stg835LoadMappings.xml to store the objects in the database.
     *
     * @param tokens - the line that represents the ealth Care Remark Codes.
     * @param dbEM - the entity manager used to persist the data to the database.
     * @throws SQLException
     */

    //Load a service payment Health Care Remark Codes related to a Service Payment
    private void loadStg835SvcLq(final String[] tokens, EntityManager dbEM)
        throws SQLException {
        System.out.println("Processing SvcLq");

        final Stg835SvcLq crStg835SvcLq = new Stg835SvcLq();
        crStg835SvcLq.setSvcLqKey((UUID.randomUUID().toString().toUpperCase()));
        crStg835SvcLq.setSvcKey(vSvcKey);
        crStg835SvcLq.setClpKey(vClpKey);
        crStg835SvcLq.setLq01cdQual(tokens[6]);
        crStg835SvcLq.setLq02rmkCd(tokens[7]);
        traceLOGGER.debug("Stg835SvcLq fieldvalues {}", crStg835SvcLq);
        dbEM.persist(crStg835SvcLq);
    }

    /**
     * This method is passed a sting of tokens that has been read from a staging data file.
     * The method expects the tokens to correspond to Provider Adjustment line -
     * identified by the first token which should be "PLB".  It takes the tokens
     * and maps them into the Stg835Plb object so that the Provider Adjustment data
     * may be stored in the database. The method uses the entity manager to store
     * the data in the database.  The entity manager in turn depends on the mappings
     * in META-INF/stg835LoadMappings.xml to store the objects in the database.
     *
     * @param tokens - the line that represents the Provider Adjustment line.
     * @param dbEM - the entity manager used to persist the data to the database.
     * @throws SQLException
     */

    //Load a Provider Adjustment line
    private void loadStg835Plb(final String[] tokens, EntityManager dbEM)
        throws SQLException {
        System.out.println("Processing Plb");

        final Stg835Plb crStg835Plb = new Stg835Plb();
        crStg835Plb.setPlbKey((UUID.randomUUID().toString().toUpperCase()));
        crStg835Plb.setFileNumber(tokens[1]);
        crStg835Plb.setSt02txnCtlNbr(tokens[2]);
        // crStg835Plb.setPlb01prvdId(tokens[3]);
        crStg835Plb.setPlb01prvdId(vProviderId);
        crStg835Plb.setPlb02fsclPerDt(tokens[4]);
        crStg835Plb.setPlb03adjRsnCd(tokens[5]);

        if (!((tokens[6].equals("")) || (tokens[6] == null))) {
            crStg835Plb.setPlb03prvdAdjId(tokens[6]);
        }

        if (!((tokens[7].equals("")) || (tokens[7] == null))) {
            crStg835Plb.setPlb04prvdAdjAmt(new BigDecimal(tokens[7]));
        }

        traceLOGGER.debug("Stg835Plb fieldvalues {}", crStg835Plb);
        dbEM.persist(crStg835Plb);
    }

    /**
     * Gets an entity name from the database using the where clause passed in.
     *
     * @param vWhereCls
     *            - selection criteria
     * @param dbEM
     *            - the entity manager that is connected to the database
     * @return - the entity selected from the database that meets the criteria
     */
    private EntityName getEntityNameVo(final String vWhereCls,
        EntityManager dbEM) {
        EntityName rowEntityName = null;
        
        
        try {
		        Query entityNameQuery = dbEM.createNativeQuery(
		                "SELECT * from ENTITY_NAME where " + vWhereCls, EntityName.class);
		        
		        
		        List entityNameList = entityNameQuery.getResultList();
		        
		        
		        Iterator entityNameIterator = entityNameList.iterator();
		
		        if (entityNameIterator.hasNext()) {
		            rowEntityName = (EntityName) entityNameIterator.next();
		        }
        }catch(Exception e) {
        	System.out.println("getEntityName() error: " + e.getMessage());
        	e.printStackTrace();
        }
		        
        return rowEntityName;
    }
}
