package gov.va.med.hac.edi.extract.b1.jpa;

import gov.va.med.domain.erepos.*;
import gov.va.med.hac.edi.utils.BadInputDataException;
import gov.va.med.hac.edi.utils.StringUtils;
import gov.va.med.hac.edi.utils.jpa.DbTransactionEnderException;

import java.util.*;

import org.hibernate.Session;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import javax.persistence.EntityManager;
import javax.persistence.Query;
import java.io.*;
import java.text.SimpleDateFormat;
import java.math.BigDecimal;
import gov.va.med.hac.edi.utils.logutils.jpa.*;

import static gov.va.med.hac.edi.utils.jpa.HibernateUtil.*;

/*
 * This class contains the logic responsible for extracting NCPDP B1 claims into a modified WPS format. 
 * This class contains the logic responsible for reading data from the database and creating the file.
 * processExtract is the entry point for this class.
 * 
 * Once processExtract is invoked the following series of events takes place:
 * 
 * 
 * processExtract selects all newly loaded batches for the sender by looking for any new claims that 
 * have been submitted related to a batch-
 * 	Select all BATCH_KEYs from table NCPDP_CLAIM where NCPDP_CLAIM.STATUS = 'NEW' and 
 * 	NCPDP_CLAIM.CLAIM_TYPE = 'B1' and NCPDP_BATCH.SENDER_ID = 'Sender Id from above'.
 * 
 * IF there are not any newly loaded claims/batches the system aborts.
 * 
 * If there are new claims processExtract loops thru each BATCH_KEY selected and creates 
 * extract file for each batch named: 'WPS_Extract' + 'Sender Id' + 'CurrentDate + '_' + CurrentTime. 
 * 
 * This file is created in directory path passed in.
 * 
 * To create the content in the extract the following methods are invoked:
 * The content of the file is created via the following methods 
 createSubmitterHeaderRec
 processHeaderRecord or createSxcProviderRec
 createSubmitterTrailerRec

 * The method createSubmitterHeaderRec creates a Header record for each SENDER_ID (submitter).
 * It gets the senders information from the Trading Partners Table (identified by sender id) and writes a submitter header 
 * record into the modified WPS format extract file.
 * 
 * The Method processHeaderRecord loops through all the the providers who submitted claims in this senders batch 
 * It writes the provider information in the extract file using the provider information 
 * obtained by searching for the provider based on TIN.  Then it writes all the claim related information
 * to the provider into the extract file.  It queries for claim information, patient information, insurance information,
 * pricing information, prescriber information, cob information, etc. and writes it to the file.
 * 
 * createSxcProviderRec performs similar functionality to processHeaderRecord but queries and writes it's provider
 * specific data (location, name, etc. in it's own format)
 * 
 * create submitted trailer record creates a trailer record for each sender id.
 * 
 * All in all the process creates a file that represents all the network pharmacy claim transactions in WPS format
 * for a specific sender's batch (SXC / Medical Matrix).
 *  
 */
public class ProcessRxExtractB1 {
	static final Logger LOGGER = LoggerFactory
			.getLogger(ProcessRxExtractB1.class);
	static final Logger traceLOGGER = LoggerFactory.getLogger("trace."
			+ ProcessRxExtractB1.class);

	// Constants
	private final String DELIMITER = "^";
	private final String transDateFormat = "yyyyMMdd";
	private final String transTimeFormat = "HHmmSSS";
	private final String FEDERAL_TAX_ID_QUALIFIER = "11";
	private final String NCPDP_PROVIDER_QUALIFIER = "07";
	private final String NATIONAL_PROVIDER_QUALIFIER = "01";

	private BigDecimal vDollarCnt = zeroBD;
	private int iPrvCnt = 0;
	private int iClmCnt = 0;
	private boolean isCitiFile = false;
	/**
	 * processExtract selects all newly loaded batches for the sender by looking
	 * for any new claims that have been submitted related to a batch- Select
	 * all BATCH_KEYs from table NCPDP_CLAIM where NCPDP_CLAIM.STATUS = 'NEW'
	 * and NCPDP_CLAIM.CLAIM_TYPE = 'B1' and NCPDP_BATCH.SENDER_ID = 'Sender Id
	 * from above'.
	 * 
	 * IF there are not any newly loaded claims/batches the system aborts.
	 * 
	 * If there are new claims processExtract loops thru each BATCH_KEY selected
	 * and creates extract file for each batch named: 'WPS_Extract' + 'Sender
	 * Id' + 'CurrentDate + '_' + CurrentTime.
	 * 
	 * This file is created in directory path passed in the parameter filePath.
	 * 
	 * To create the content in the extract the following methods are invoked:
	 * The content of the file is created via the following methods
	 * createSubmitterHeaderRec processHeaderRecord or createSxcProviderRec
	 * createSubmitterTrailerRec
	 * 
	 * @param filePath
	 *            - the path where the extract file will be written
	 * @param vPbm
	 *            - the sender id - an extract file will be created for each new
	 *            batch from this sender
	 * @param dbEM
	 *            - an entity manager connected to the NCPDP database schema
	 *            (used to query for new batches from sender)
	 * 
	 */

	protected void processExtract(String filePath, String vPbm,
			EntityManager dbEM) throws DbTransactionEnderException {
		String vFileNum = "";

		String vSenderId = "";
		String vBatchNumber = "";
	
		String vTransRefNum = "";

		List listOfClaimsToPersist = new ArrayList();
		try {
			// Get the batches that have B1 claims in status NEW for this sender
			Query batchKeysQuery = dbEM
					.createNativeQuery("SELECT DISTINCT BATCH_KEY, FILE_NUMBER "
							+ "  FROM NCPDP_CLAIM "
							+ " WHERE STATUS = 'NEW' "
							+ "AND CLAIM_TYPE = 'B1' "
							+ "   AND SENDER_ID = :senderId");
			batchKeysQuery.setParameter("senderId", vPbm);
			// query.setParameter(1, vPbm);
			List batchKeyList = batchKeysQuery.getResultList();

			int batchesCount = batchKeyList.size();
			LOGGER.info("Batches with NEW B1 claims for sender " + vPbm + ": "
					+ batchesCount);

			// Loop through the batches, processing them
			String nextBatchKey = "";
			String nextFileNumber = "";
			Iterator batchKeyIterator = batchKeyList.iterator();
			if (batchesCount > 0) {
				while (batchKeyIterator.hasNext()) {
					Object[] batchKeyFileNumberArray = (Object[]) batchKeyIterator
							.next();
					nextBatchKey = (String) batchKeyFileNumberArray[0];
					nextFileNumber = (String) batchKeyFileNumberArray[1];
					nextFileNumber = nextFileNumber.split("\\.")[0];
					
					if(nextFileNumber.startsWith("CITI"))
					{
						isCitiFile = true;
					} else {
                  isCitiFile = false;
               }


					traceLOGGER.trace("Next Batch Key:" + nextBatchKey);

					// Put together an output filename
					SimpleDateFormat transDate = new SimpleDateFormat(
							transDateFormat);
					SimpleDateFormat transTime = new SimpleDateFormat(
							transTimeFormat);
					java.util.Date date = new GregorianCalendar().getTime();
					String vCurrentDate = transDate.format(date);
					String vCurrentTime = transTime.format(date);
					String fileName = "WPS_" + vPbm + "_" + nextFileNumber
							+ "_" + vCurrentDate + "_" + vCurrentTime + ".DAT";

					String extractFileFullName = filePath + fileName;
					LOGGER.info("Output file: " + extractFileFullName);
					FileOutputStream extractFileOutputStream = new FileOutputStream(
							new File(extractFileFullName));
					PrintWriter vExtractFile = new PrintWriter(
							extractFileOutputStream);

					// Get the full batches from the list obtained earlier
					Query batchesQuery = dbEM.createQuery("SELECT b "
							+ "FROM gov.va.med.domain.erepos.NcpdpBatch b "
							+ "WHERE b.batchKey = :batchKey "
							+ "ORDER BY b.dateCreated");
					batchesQuery.setParameter("batchKey", nextBatchKey);
					List batches = batchesQuery.getResultList();

					// Loop through the batches, processing each one
					Iterator batchIterator = batches.iterator();
					if (batchIterator.hasNext()) {
						NcpdpBatch rowNcpdpBatch = (NcpdpBatch) batchIterator
								.next();
						traceLOGGER.trace("Batch as queried : {}",
								new NcpdpBatchLogger(rowNcpdpBatch));

						// Set some variables that are used in several places
						vSenderId = rowNcpdpBatch.getSenderId();
						vBatchNumber = rowNcpdpBatch.getBatchNumber()
								.substring(2);
						vFileNum = rowNcpdpBatch.getFileNumber();

						// Create a record in the output for the sender
						createSubmitterHeaderRec(rowNcpdpBatch, vExtractFile,
								dbEM, vSenderId, vBatchNumber);

						// Use different methods for SXC and anyone else
						if (vPbm.equals("752578509")) {

							ReturnVariables returnVariables = createSxcProviderRec(
									nextBatchKey, vExtractFile, dbEM,
									vSenderId, vBatchNumber, vFileNum, vTransRefNum);
							listOfClaimsToPersist.addAll(returnVariables
									.getListOfClaimsToReturn());
							vTransRefNum = returnVariables.getvTransRefNum();
						} else {
							throw new Exception(
									"Unknown partner - this process only works for SXC");
						}

						createSubmitterTrailerRec(vExtractFile, vSenderId,
								vBatchNumber);
						iPrvCnt = 0;
					} else {
						LOGGER.info("Found no batch with key: " + nextBatchKey);
					}

					vExtractFile.flush();
					vExtractFile.close();
					traceLOGGER.trace("Committing the database transaction");

					// Save the claims because their status is changed
					dbEM.getTransaction().begin();
					Iterator claimIterator = listOfClaimsToPersist.iterator();
					while (claimIterator.hasNext()) {
						dbEM.persist(claimIterator.next());
					}
					dbEM.getTransaction().commit();

					traceLOGGER.trace("Commit successful");
					initializeVariables();
					vSenderId = "";
					vBatchNumber = "";
					vTransRefNum = "";

					LOGGER.info("WPS Extract File for sender id " + vPbm
							+ " and file " + vFileNum
							+ " was Successfully Created.");
				}
			} else {
				LOGGER
						.info("There are no new NCPDP B1 Claims to be extracted for sender id "
								+ vPbm + ". ");
			}
		} catch (Exception ex) {
			// Validation for rollbacking the database
			if (dbEM != null && dbEM.getTransaction().isActive()) {
				LOGGER.warn("Rolling back the database transaction");
				dbEM.getTransaction().rollback();
				LOGGER.warn("Rollback successful");
			}
			throw new DbTransactionEnderException(
					"Rolled Back the Transaction", ex);
		}
	}

	/**
	 * Creates the submitter header that appears at the top of the extract file.
	 * Notice the line begins with a 0.
	 * 
	 * File Format Written
	 * 
	 * @param rowNcpdpBatch
	 * @param vExtFile
	 * @param dbEM
	 * @throws BadInputDataException
	 */
	public void createSubmitterHeaderRec(NcpdpBatch rowNcpdpBatch,
			PrintWriter vExtFile, EntityManager dbEM, String vSenderId,
			String vBatchNumber) throws BadInputDataException {
		StringBuffer sbHdrRec = new StringBuffer();
		sbHdrRec.append(delimit("0")); // Record Id
		// TWP: I guess processor number and sender id are the same thing
		sbHdrRec.append(delimit(vSenderId)); // Processor Number
		sbHdrRec.append(delimit(vBatchNumber)); // Batch Number

		Query tradingPartnersQuery = dbEM.createNativeQuery(
				"SELECT * FROM TRADING_PARTNERS WHERE TRADING_PARTNER_ID=?",
				TradingPartners.class);
		tradingPartnersQuery.setParameter(1, vSenderId);
		List tradingPartners = tradingPartnersQuery.getResultList();
		Iterator tradingPartnersIterator = tradingPartners.iterator();
		if (tradingPartnersIterator.hasNext()) {
			TradingPartners rowTradingPartners = (TradingPartners) tradingPartnersIterator
					.next();
			traceLOGGER.trace("Trading Partner as queried:"
					+ rowTradingPartners.getName());

			sbHdrRec.append(delimit(rowTradingPartners.getName())); // Processor
			// Name
			sbHdrRec.append(delimit(rowTradingPartners.getAddressLine1())); // Processor
			// Address
			sbHdrRec.append(delimit(rowTradingPartners.getCity()));
			sbHdrRec.append(delimit(rowTradingPartners.getState()));
			sbHdrRec.append(delimit(rowTradingPartners.getZipCode()));
			sbHdrRec.append(delimit(rowTradingPartners.getContactPhone()));
			String vDt = rowNcpdpBatch.getCreationDate().substring(2);
			String vRunDate = vDt.substring(2, 4) + vDt.substring(4, 6)
					+ vDt.substring(0, 2);

			sbHdrRec.append(delimit(vRunDate)); // Run Date

			sbHdrRec.append(delimit("H")); // Third Party Type always "H" =
			// CHAMPUS
			// TWP: Will we ever process M or P?
			sbHdrRec.append(delimit("39126WPS0000000")); // Tape Destination
			// Carrier
			sbHdrRec.append(delimit("")); // 000000000 - Tape Destination
			// Processor Number
			sbHdrRec.append(delimit("")); // 0000 - 9 Expanded Zip
			sbHdrRec.append(delimit(rowNcpdpBatch.getCreationTime())); // WPS
			// Extension
			// Transmission
			// Number
			// TWP:??

			traceLOGGER.trace("Submitter Header as written: {}", sbHdrRec);
			vExtFile.println(sbHdrRec);
			vExtFile.flush();
			sbHdrRec.delete(0, sbHdrRec.length());
			detach(rowTradingPartners, dbEM);
		} else {
			LOGGER.warn("Unknown Trading Partner: " + vSenderId);
			throw new BadInputDataException("Unknown Trading Partner : "
					+ vSenderId);
		}
	}

	/**
	 * Creates the submitter trailer that appears at the top of the extract
	 * file.
	 * 
	 * 1 Record ID R N "Right justified; Zero filled" 1 1 Must = "8" 2 Processor
	 * Number R N "Left justified; Space filled" 2 11 Must = Submitter Header --
	 * Field 2 3 Batch Number O N "Right justified; Zero filled" 12 16 4
	 * Pharmacy Count R N "Right justified; Zero filled" 17 20 5 Comments O A/N
	 * "Left justified;Space filled" 21 318
	 * 
	 * @param vExtFile
	 */
	public void createSubmitterTrailerRec(PrintWriter vExtFile,
			String vSenderId, String vBatchNumber) {
		StringBuffer sbTlrRec = new StringBuffer();
		sbTlrRec.append(delimit("80")); // Record Id
		sbTlrRec.append(delimit(vSenderId)); // Processor Number
		sbTlrRec.append(delimit(vBatchNumber)); // Batch Number
		sbTlrRec.append(delimit(Integer.toString(iPrvCnt))); // Pharmacy Count -
		// Always 1
		// since SXC is
		// only Pharmacy
		sbTlrRec.append(delimit("")); // Comments -TWP:mod: I think this is okay
		// and you should leave it alone
		traceLOGGER.trace("Submitter TrailerRec as written: {}", sbTlrRec);
		vExtFile.println(sbTlrRec);
		vExtFile.flush();
		sbTlrRec.delete(0, sbTlrRec.length());
	}

	// No need to group by the Provider ID as they will be replaced by SXC Tax
	// ID.
	public ReturnVariables createSxcProviderRec(String vBatchKey,
			PrintWriter vExtFile, EntityManager dbEM, String vSenderId,
			String vBatchNumber, String vFileNum, String vTransRefNum) {

		//Initialize the variable used in thei method
		String vHeaderKey = "";
		List listOfClaimsToPersist = new ArrayList();

		iClmCnt = 0;
		vDollarCnt = zeroBD;
		
		//The records contained in the file are structured differently for
		//normal (non-citi) files and citi files. The if blocks below control
		//which file structure will be created.  The record types are the same
		//in both files they are: 0=submitter header 10=provide header 
		//40=claim record 60=provider trailer 80=submitter trailer
		if(!isCitiFile)
			
		{ 
			/* The structure of records in the normal extract file is
			 * 0 20 40 .... 40 60 80 */
			vTransRefNum = writeExtractWithDefaultFileStructure(vBatchKey,
				vExtFile, dbEM, vSenderId, vBatchNumber, vFileNum,
				vTransRefNum, listOfClaimsToPersist);		 			
		}else
		{
			/* The structure of records in the CITI extract file is 
			 * 0 20 40 60 20 40 60 80 */			
			vTransRefNum = writeExtractFileWithCITIFileStructure(vBatchKey,
					vExtFile, dbEM, vSenderId, vBatchNumber, vFileNum,
					vTransRefNum, listOfClaimsToPersist);
		}
					

		vExtFile.flush();
		ReturnVariables returnValue = new ReturnVariables();
		returnValue.setListOfClaimsToReturn(listOfClaimsToPersist);
		returnValue.setvTransRefNum(vTransRefNum);
		
		iClmCnt = 0;
		vDollarCnt = zeroBD;
		
		return returnValue;
	}

	private String writeExtractFileWithCITIFileStructure(String vBatchKey,
			PrintWriter vExtFile, EntityManager dbEM, String vSenderId,
			String vBatchNumber, String vFileNum, 
			String vTransRefNum, List listOfClaimsToPersist) {
	  /* The structure of records in the CITI extract file is 
	   * 0 20 40 60 20 40 60 80 
	   */
		Query getHeaderQuery = dbEM.createQuery("SELECT h "
				+ "FROM gov.va.med.domain.erepos.NcpdpHeader h "
				+ "WHERE BATCH_KEY = ?");
		getHeaderQuery.setParameter(1, vBatchKey);
		List<NcpdpHeader> headers = getHeaderQuery.getResultList();
		Collections.sort(headers, new Comparator<NcpdpHeader>(){
			public int compare(NcpdpHeader a, NcpdpHeader b)
			{
				return a.getSvcPrv().compareTo(b.getSvcPrv());
			}
		});
						
		String vHeaderKey;
		String lastSvcPrv = "";
		for(int x = 0; x < headers.size(); x++) {
			
			NcpdpHeader rowNcpdpHeader = (NcpdpHeader) headers.get(x);
			traceLOGGER.trace("Header as queried: {}", new NcpdpHeaderLogger(
					rowNcpdpHeader));
			vHeaderKey = rowNcpdpHeader.getRealKey();
			vTransRefNum = rowNcpdpHeader.getTransactionRefNum();
			String vDateOfSvc = rowNcpdpHeader.getDateOfService();
			String vSvcPrv = rowNcpdpHeader.getSvcPrv();

			//Write 1 provider header/trailer record (20/60) 
			//for each group of claim record that share the same NPI - 
			//that is we only write a header or trailer/header combination
			//when we see a new NPI since the header list is sorted by NPI
			if(!lastSvcPrv.equalsIgnoreCase(vSvcPrv))
			{				
				//Every time we hit a new NPI other than the first time we need to write a trailer
				if(x != 0) 
				{
					//Create the 60 record to close the previous 20/60 pair
					createProviderTrailerRec(vExtFile, vSenderId, vBatchNumber);
					iClmCnt = 0;
					vDollarCnt = zeroBD;
				}
								
				//Create the 20 record - to indicate the next group of claims 
				//share the NPI associated with this first header
				createSxcProviderHeaderRec(vBatchKey, vExtFile, dbEM, vSenderId,
					vBatchNumber, vFileNum, vSvcPrv);
				iPrvCnt++;

				//Keep track of the last NPI seen so we know when it changes
				lastSvcPrv = vSvcPrv;
			}
												
			//Create the 40 records for this provider
			//This method writes all claims associated with header to the file
			//based on the header but it does not write the header
			listOfClaimsToPersist.addAll(processHeader(rowNcpdpHeader,
					vHeaderKey, vExtFile, dbEM, vSenderId, vBatchNumber,
					vSvcPrv, vTransRefNum, vDateOfSvc));
										
			
			detach(rowNcpdpHeader, dbEM);
		}
		
		//Create the 60 record for the last group of claims 
		createProviderTrailerRec(vExtFile, vSenderId, vBatchNumber);
		
		return vTransRefNum;
	}

	private String writeExtractWithDefaultFileStructure(String vBatchKey,
			PrintWriter vExtFile, EntityManager dbEM, String vSenderId,
			String vBatchNumber, String vFileNum,
			String vTransRefNum, List listOfClaimsToPersist) {
		String vHeaderKey;
		/* The structure of records in the normal extract file is
	   * 0 20 40 .... 40 60 80
	   */
		
		
		String vSvcPrv = vSenderId;  //For non-citi claims the service provide is always SXC (senderId)

		//Write 1 provider header/trailer record (20/60) 
		//that wraps all the claim records			
		createSxcProviderHeaderRec(vBatchKey, vExtFile, dbEM, vSenderId,
				vBatchNumber, vFileNum, vSvcPrv);
		iPrvCnt++;
		
		Query getHeaderQuery = dbEM.createQuery("SELECT h "
				+ "FROM gov.va.med.domain.erepos.NcpdpHeader h "
				+ "WHERE BATCH_KEY = ?");
		getHeaderQuery.setParameter(1, vBatchKey);
		List headers = getHeaderQuery.getResultList();
		Iterator headerIterator = headers.iterator();
		
		while (headerIterator.hasNext()) {
			NcpdpHeader rowNcpdpHeader = (NcpdpHeader) headerIterator.next();
			traceLOGGER.trace("Header as queried: {}", new NcpdpHeaderLogger(
					rowNcpdpHeader));
			vHeaderKey = rowNcpdpHeader.getRealKey();
			vTransRefNum = rowNcpdpHeader.getTransactionRefNum();
			String vDateOfSvc = rowNcpdpHeader.getDateOfService();
			listOfClaimsToPersist.addAll(processHeader(rowNcpdpHeader,
					vHeaderKey, vExtFile, dbEM, vSenderId, vBatchNumber,
					vSvcPrv, vTransRefNum, vDateOfSvc));
			detach(rowNcpdpHeader, dbEM);
		}
		
		createProviderTrailerRec(vExtFile, vSenderId, vBatchNumber);
		return vTransRefNum;
	}

	public void createSxcProviderHeaderRec(String vBatchKey,
			PrintWriter vExtFile, EntityManager dbEM, String vSenderId,
			String vBatchNumber, String vFileNum, String vSvcPrv) {
		StringBuffer prvHdrRec = new StringBuffer();
		prvHdrRec.append(delimit("20")); // Record ID
		prvHdrRec.append(delimit(vSenderId)); // Processor Number
		prvHdrRec.append(delimit(vBatchNumber)); // Batch Number
		prvHdrRec.append(delimit(vSvcPrv)); // Pharmacy Number (Service
		// Provider ID) - Will Always be SXC for non-citi claims

		// Provider TIN for Non-Network Claims only.
		boolean bNetwork = true;
		if (vFileNum.substring(0, 2).equalsIgnoreCase("NN")) {
			bNetwork = false;
		}

		if (bNetwork) {
			Query tradingPartnersQuery = dbEM
					.createNativeQuery(
							"SELECT * FROM TRADING_PARTNERS WHERE TRADING_PARTNER_ID=?",
							TradingPartners.class);
			tradingPartnersQuery.setParameter(1, vSenderId);
			List tradingPartners = tradingPartnersQuery.getResultList();
			Iterator tradingPartnersIterator = tradingPartners.iterator();

			if (tradingPartnersIterator.hasNext()) {
				TradingPartners rowTradingPartners = (TradingPartners) tradingPartnersIterator
						.next();
				if (rowTradingPartners != null) {
					prvHdrRec.append(delimit(getValue(rowTradingPartners
							.getName(), "", true, 20)));
					prvHdrRec.append(delimit(getValue(rowTradingPartners
							.getAddressLine1(), "", true, 20)));
					prvHdrRec.append(delimit(getValue(rowTradingPartners
							.getCity(), "", true, 18)));
					prvHdrRec.append(delimit(getValue(rowTradingPartners
							.getState(), "", false, -1)));
					prvHdrRec.append(delimit(getValue(rowTradingPartners
							.getZipCode(), "", true, 5)));
					prvHdrRec.append(delimit(getValue(rowTradingPartners
							.getContactPhone(), "", false, -1)));
					prvHdrRec.append(delimit(""));

				} else {
					createRepeatedDelimitedFields(prvHdrRec, "", 7);
				}
				detach(rowTradingPartners, dbEM);
			} else {
				createRepeatedDelimitedFields(prvHdrRec, "", 7);
			}

			prvHdrRec.append(delimit("Y"));// Network Indicator
			prvHdrRec.append(delimit(""));// TWP: Best Guess: WPS Extension:
			// Provider Sub ID-- Champus
			// Assigned Keep This Optional Field
			// though we do nothing with it
			prvHdrRec.append(delimit("Y"));// TWP: Best Guess: WPS Extension:
			// Provider Participation Indicator
			// Keep
			prvHdrRec.append(delimit("E"));// TWP: Best Guess: WPS Extension:
			// Provider ID Type Keep
			prvHdrRec.append(delimit("Y"));// TWP: Best Guess: WPS Extension:
			// Provider Release Indicator Keep

			traceLOGGER.trace("SXC ProviderHeaderRec: {}", prvHdrRec);
			vExtFile.println(prvHdrRec);
			vExtFile.flush();
			prvHdrRec.delete(0, prvHdrRec.length());

		} else {
			throw new RuntimeException(
					"Unexpected: Non-Network Claim found logic needs to be implemented to handle this scenario.");
		}

	}

	public void createProviderTrailerRec(PrintWriter vExtFile,
			String vSenderId, String vBatchNumber) {
		StringBuffer prvTlrRec = new StringBuffer();
		prvTlrRec.append(delimit("60")); // Record Id
		prvTlrRec.append(delimit(vSenderId)); // Processor Number
		prvTlrRec.append(delimit(vBatchNumber)); // Batch Number
		prvTlrRec.append(delimit(Integer.toString(iClmCnt))); // Claim Count
		prvTlrRec.append(delimit(vDollarCnt)); // Dollar Count - this is an
		// aggregate of all gross amount
		// due
		// -> TWP Question - is dollar count correct? are we paying someone? why
		// not gross - ohi payments maybe I don't understand?
		traceLOGGER.trace("ProviderTrailerRec as written: {}", prvTlrRec);
		vExtFile.println(prvTlrRec);
		vExtFile.flush();
	}

	public List processHeader(NcpdpHeader rowNcpdpHeader, String vRealKey,
			PrintWriter vExtFile, EntityManager dbEM, String vSenderId,
			String vBatchNumber, String vSvcPrv,
			String vTransRefNum, String vDateOfSvc) {

		List listOfClaimsToPersist = new ArrayList();
		StringBuffer claimOutput = new StringBuffer();

		// Query the claim, prescribers, pricings, and COBs for this header - we
		// are pre-fetching for performance
		Query getClaimQuery = dbEM.createQuery("SELECT DISTINCT c "
				+ "FROM gov.va.med.domain.erepos.NcpdpClaim c "
				+ "LEFT JOIN FETCH c.ncpdpPrescribers "
				+ "LEFT JOIN FETCH c.ncpdpPricings "
				+ "LEFT JOIN FETCH c.ncpdpCobs  " + "WHERE c.status =? "
				+ "AND c.ncpdpHeader=? " + "AND c.claimType =? "
				+ "ORDER BY c.dateCreated");
		getClaimQuery.setParameter(1, "NEW");
		getClaimQuery.setParameter(2, rowNcpdpHeader);
		getClaimQuery.setParameter(3, "B1");
		List claims = getClaimQuery.getResultList();

		createClaimRecords(rowNcpdpHeader, vExtFile, dbEM, vSenderId,
				vBatchNumber, vSvcPrv, vTransRefNum, vDateOfSvc,
				listOfClaimsToPersist, claimOutput, claims);

		return listOfClaimsToPersist;
	}

	protected void createClaimRecords(NcpdpHeader rowNcpdpHeader,
			PrintWriter vExtFile, EntityManager dbEM, String vSenderId,
			String vBatchNumber, String vSvcPrv,
			String vTransRefNum, String vDateOfSvc, List listOfClaimsToPersist,
			StringBuffer claimOutput, List claims) {
		// Get the patient for this header
		NcpdpPatient rowPatient = (NcpdpPatient) this
				.getFirstObjectFromJPASet(rowNcpdpHeader.getNcpdpPatients());
		traceLOGGER.trace("Patient as queried: {}", new NcpdpPatientLogger(
				rowPatient));

		// Get the insurance for this header
		NcpdpInsurance rowInsurance = (NcpdpInsurance) this
				.getFirstObjectFromJPASet(rowNcpdpHeader.getNcpdpInsurances());
		traceLOGGER.trace("Insurance as queried: {}", new NcpdpInsuranceLogger(
				rowInsurance));

		// Loop through the claims, processing each one
		Iterator claimsIterator = claims.iterator();
		while (claimsIterator.hasNext()) {
			iClmCnt++;

			NcpdpClaim rowClaim = (NcpdpClaim) claimsIterator.next();
			traceLOGGER.trace("Claim as queried: {}", new NcpdpClaimLogger(
					rowClaim));

			NcpdpPrescriber rowPrescriber = (NcpdpPrescriber) this
					.getFirstObjectFromJPASet(rowClaim.getNcpdpPrescribers());
			traceLOGGER.trace("Prescriber as queried: {}",
					new NcpdpPrescriberLogger(rowPrescriber));

			NcpdpPricing rowPricing = (NcpdpPricing) this
					.getFirstObjectFromJPASet(rowClaim.getNcpdpPricings());
			traceLOGGER.trace("Pricing as queried: {}", new NcpdpPricingLogger(
					rowPricing));

			// Put together the claim record
			claimOutput.append(delimit("40")); // Record Id
			claimOutput.append(delimit(vSenderId)); // Processor Number
			claimOutput.append(delimit(vBatchNumber));// Batch Number


			if (vSenderId.equals("752578509")) {
				if(isCitiFile)
				{
					claimOutput.append(delimit(rowNcpdpHeader.getSvcPrv()));
				}else
				{
					claimOutput.append(delimit(vSenderId)); // Pharmacy Number
				}
			} else {
				claimOutput.append(delimit(rowNcpdpHeader.getSvcPrv())); // Pharmacy Number
				LOGGER
						.warn("Claims came from pharmacy other than SXC - this code was designed for SXC - senderId:"
								+ vSenderId);
			}
		

			claimOutput.append(delimit(rowClaim.getRxSvcRefNum())); // Prescription
			// Number
			claimOutput.append(delimit(vDateOfSvc)); // Date of service - this
			// comes from the header
			// and is passed in
			// (Why?)
			claimOutput.append(delimit(rowClaim.getProductSvcId())); // This is
			// the
			// drug
			// code
			claimOutput.append(delimit(getValue(rowClaim.getCompoundCode(), "",
					false, -1))); // Compound Code
			claimOutput.append(delimit(rowClaim.getFillNumber())); // Fill
			// Number
			
			
			claimOutput.append(delimit(getDecimalFormattedQuantity(rowClaim.getQuantityDispensed()))); // Quantity
			// Dispensed
			claimOutput.append(delimit(rowClaim.getDaysSupply())); // Days
			// Supply

			String medicaidPdAmnt = "";
			String dispensingFeeSub = "";
			// If a rowPricing record was retrieved with the claim, use it now
			if (rowPricing != null) {

				if (rowPricing.getPatPaidAmt() != null) {
					claimOutput.append(delimit(rowPricing.getPatPaidAmt())); // Patient
					// Paid
					// Amount
					// Submitted
				} else {
					claimOutput.append(delimit("")); // Patient Paid Amount
					// Submitted if
					// PatPaidAmount id null
				}

				if (rowPricing.getGrossAmtDue() != null) {
					claimOutput.append(delimit(rowPricing.getGrossAmtDue())); // Gross
					// Amount
					// Due
					vDollarCnt = vDollarCnt.add(rowPricing.getGrossAmtDue());
				} else {
					claimOutput.append(delimit("")); // Gross Amount Due if
					// grossAmountDue is
					// null
				}

				// need to save off paid amount which is written to this record
				// furthur down
				medicaidPdAmnt = (rowPricing.getMedicaidPdAmnt() != null) ? rowPricing
						.getMedicaidPdAmnt().toString()
						: null;

				// also need to save off dispensing fee
				dispensingFeeSub = (rowPricing.getDispensingFeeSub() != null) ? rowPricing
						.getDispensingFeeSub().toString()
						: null;

				detach(rowPricing, dbEM);
			} else {
				claimOutput.append(delimit("")); // Patient Paid Amount
				// Submitted if rowPricing =
				// null
				claimOutput.append(delimit("")); // Gross Amount Due if
				// rowPricing = null
			}

			claimOutput.append(delimit(getValue(getPatientFullName(rowPatient),
					"", false, -1))); // Patient Name
			claimOutput.append(delimit(rowPatient.getDateOfBirth())); // Date of
			// Birth
			claimOutput.append(delimit(rowPatient.getPatGenderCode())); // Sex
			// of
			// Patient
			// - "1"
			// male
			// "2"
			// female
			claimOutput.append(delimit(getValue(rowPatient.getPatId(),
					rowInsurance.getCrdhldrId(), false, -1))); // Beneficiary
			// SSN (Patient
			// ID) (notice
			// default value
			// is different
			// not "")
			// if we don't have the patients information we use card holders
			claimOutput.append(delimit(StringUtils.padString(rowInsurance
					.getPatRelshpCd(), 2, '0', true))); // Patient Relationship

			// If a rowPrescriber record was retrieved with the claim, use it
			// now
			if (rowPrescriber != null) {
				claimOutput
						.append(delimit(rowPrescriber.getPrescriberIdQual())); // Prescriber
				// Id
				// Qualifier
				if (rowPrescriber.getPrescriberIdQual() != null) {
					if (rowPrescriber.getPrescriberIdQual().trim()
							.equalsIgnoreCase("12")) {
						claimOutput.append(delimit(getValue(rowPrescriber
								.getPrescriberId(), "", true, 10)));
					} else {
						claimOutput.append(delimit("")); // Physician ID
						// (Prescriber ID)
						// qualifier is not
						// 12
					}
				} else {
					claimOutput.append(delimit("")); // Physician ID (Prescriber
					// ID) if we don't have
					// a qualifier
				}
				detach(rowPrescriber, dbEM);
			} else {
				// Prescriber ID Qualifier if rowPrescriber = null
				// Physician ID (Prescriber ID) if rowPrescriber = null
				createRepeatedDelimitedFields(claimOutput, "", 2);
			}

			claimOutput.append(delimit(rowClaim.getDateRxWritten())); // Date
			// Written
			claimOutput.append(delimit(vTransRefNum)); // Matrix HIC Number -
			// comes from
			// NCPDP_HEADER
			// TRANSACTION_REF_NUM
			// SXC Transaction Reference Number

			claimOutput.append(delimit(getValue(rowPatient.getPatStreetAddr(),
					"", true, 15))); // Patient Street Address
			claimOutput.append(delimit(getValue(rowPatient.getPatCityAddr(),
					"", true, 10))); // Patient State (province)
			claimOutput.append(delimit(rowPatient.getPatStateProvAddr())); // Patient
			// State

			// Patient Zip Code -
			// we can't use get value because it can't handle checking for a
			// length greater than 9 and returning length of 8
			String vZipCode = rowPatient.getPatZipPstZone();
			if (vZipCode != null && vZipCode.length() > 9) {
				vZipCode = vZipCode.substring(0, 8);
			}
			claimOutput.append(delimit(getValue(vZipCode, "", true, 9)));

			// Network or Non_Network Claim indicator
			claimOutput.append(delimit(rowClaim.getNetworkIndicator()));

			Set providers = rowClaim.getNcpdpProviders();
			NcpdpProvider ncpdpProvider = (NcpdpProvider) getFirstObjectFromJPASet(providers);
			if (ncpdpProvider != null) {
				// TWP: Is this for sure the Provider Dispense ID Qualifier?
				claimOutput.append(delimit(getValue(ncpdpProvider
						.getProviderIdQual(), "", false, -1))); // Provider
				// Dispense ID
				// Qualifier
				claimOutput.append(delimit(getValue(ncpdpProvider
						.getProviderId(), "", false, -1))); // Provider Dispense
				// ID
			} else {
				// Provider Dispense ID Qualifier
				// Provider Dispense ID
				createRepeatedDelimitedFields(claimOutput, "", 2);
			}

			claimOutput.append(delimit(getValue(
					rowInsurance.getMedicaidIdNbr(), "", false, -1))); // Medicaid
			// Id
			// Number
			claimOutput.append(delimit(getValue(rowInsurance
					.getMedicaidAgencyNbr(), "", false, -1))); // Medicaid
			// Agency Number
			claimOutput
					.append(delimit(getValue(medicaidPdAmnt, "", false, -1))); // Medicaid
			// Payment
			claimOutput.append(delimit(getValue(rowClaim
					.getMedicaidSbrgtnIcntcn(), "", false, -1))); // Medicaid
			// Subrogation
			// Number
			claimOutput
					.append(delimit(getValue(dispensingFeeSub, "", false, -1))); // DISPENSING
			// FEE
			// SUBMITTED
			traceLOGGER.trace("Claim as written: {}", claimOutput);
			vExtFile.println(claimOutput.toString());

			// Get coordination of Benefits to get OHI data
			Set cobs = rowClaim.getNcpdpCobs();
			if (cobs != null) {
				Iterator cobIterator = cobs.iterator();

				if (cobIterator.hasNext()) {
					NcpdpCob rowCob = (NcpdpCob) cobIterator.next();
					traceLOGGER.trace("COB as queried: {}", new NcpdpCobLogger(
							rowCob));

					Set cobOtPayers = rowCob.getNcpdpCobOtPayers();
					createOHIRecords(dbEM, cobOtPayers, vExtFile);
					detach(rowCob, dbEM);
				}
			}

			// TWP:Adding support for multi ingredient - record types 50 and 51
			Set<NcpdpCompound> compounds = rowClaim.getNcpdpCompounds();
			if (compounds != null) {
				Iterator<NcpdpCompound> compoundIterator = compounds.iterator();

				// Realistically there is only one compound per claim although
				// DB
				// supports 1 to many
				if (compoundIterator.hasNext()) {
					// If the compound has multiple ingredients or modifiers to
					// those ingredients
					// we are writing them to the extract file

					NcpdpCompound currentCompound = compoundIterator.next();
					Set<NcpdpCompoundIngr> compoundIngrs = currentCompound
							.getNcpdpCompoundIngrs();
					if (compoundIngrs != null) {
						Iterator<NcpdpCompoundIngr> compoundIngrIterator = compoundIngrs
								.iterator();
						// We are about to write records for multiple
						// ingredients of
						// a
						// drug (compound) - Type 50
						while (compoundIngrIterator.hasNext()) {
							NcpdpCompoundIngr currentCompoundIngr = compoundIngrIterator
									.next();
							createMultiIngredientRecords(dbEM,
									currentCompoundIngr, vExtFile);

							Set<NcpdpCmpIngdtMod> compoundIngrMods = currentCompoundIngr
									.getNcpdpCmpIngdtMods();
							if (compoundIngrMods != null) {
								Iterator<NcpdpCmpIngdtMod> compoundIngrModIterator = compoundIngrMods
										.iterator();
								// We are about to write records for multiple
								// ingredient
								// Modifiers - Type 55
								while (compoundIngrModIterator.hasNext()) {
									NcpdpCmpIngdtMod currentCompoundIngrMod = compoundIngrModIterator
											.next();
									createMultiIngredientModRecords(dbEM,
											currentCompoundIngrMod, vExtFile);
									detach(currentCompoundIngrMod, dbEM);
								}
							}

							detach(currentCompoundIngr, dbEM);
						}
					}
				}

				detach(compounds, dbEM);

			}
			rowClaim.setStatus("PROCESSED");
			listOfClaimsToPersist.add(rowClaim);
			vExtFile.flush();
			claimOutput.delete(0, claimOutput.length());
		}

		detach(rowInsurance, dbEM);
		detach(rowPatient, dbEM);
	}

	/**
	 * This method creates a line in the output file that contains information
	 * about one of the ingredients in a muti-ingredient compound.
	 * 
	 * @param dbEM
	 * @param currentCompoundIngr
	 * @param vExtFile
	 */
	private void createMultiIngredientRecords(EntityManager dbEM,
			NcpdpCompoundIngr currentCompoundIngr, PrintWriter vExtFile) {
		StringBuffer compoundIngrOutput = new StringBuffer();
		compoundIngrOutput.append(delimit("50"));
		compoundIngrOutput.append(delimit(currentCompoundIngr
				.getCmpPrdtIdQual()));
		compoundIngrOutput.append(delimit(currentCompoundIngr.getCmpPrdtId()));
		compoundIngrOutput.append(delimit(getDecimalFormattedQuantity(currentCompoundIngr.getCmpIngdtQty())));
		compoundIngrOutput.append(delimit(currentCompoundIngr
				.getCmpIngdtDrgCost()));
		compoundIngrOutput.append(delimit(currentCompoundIngr
				.getCmpIngdtCostDet()));
		vExtFile.println(compoundIngrOutput.toString());
	}

	/**
	 * This method creates a line in the output file that contains information
	 * related to one single modification to a multi-ingredient compound.
	 * 
	 * @param dbEM
	 * @param currentCompoundIngrMod
	 * @param vExtFile
	 */
	private void createMultiIngredientModRecords(EntityManager dbEM,
			NcpdpCmpIngdtMod currentCompoundIngrMod, PrintWriter vExtFile) {
		StringBuffer compoundIngrModOutput = new StringBuffer();
		compoundIngrModOutput.append(delimit("51"));
		compoundIngrModOutput.append(delimit(currentCompoundIngrMod
				.getCmpIngdtModCd()));
		vExtFile.println(compoundIngrModOutput.toString());
	}

	/**
	 * 
	 * @param rowPatient
	 * @return
	 */
	private String getPatientFullName(NcpdpPatient rowPatient) {
		String vName = "";
		if (rowPatient.getPatLastName() != null) {
			vName = rowPatient.getPatLastName();
		}
		if (rowPatient.getPatFirstName() != null) {
			vName += "," + rowPatient.getPatFirstName();
		}
		return vName;
	}

	private Object getFirstObjectFromJPASet(Set someSet) {
		if (someSet != null && !someSet.isEmpty()) {
			if (someSet.iterator() != null) {
				Iterator someSetIterator = someSet.iterator();
				if (someSetIterator.hasNext()) {
					return someSetIterator.next();
				}
			}
		}
		return null;
	}

	private void createOHIRecords(EntityManager dbEM, Set cobOtPayers,
			PrintWriter vExtFile) {

		StringBuffer ohiOutput = new StringBuffer();
		if (cobOtPayers != null) {
			Iterator cobOtPayerIterator = cobOtPayers.iterator();
			while (cobOtPayerIterator.hasNext()) {
				ohiOutput.append(delimit("41"));

				NcpdpCobOtPayer rowCobOtPayer = (NcpdpCobOtPayer) cobOtPayerIterator
						.next();
				traceLOGGER.trace("COBOtPayer as queried: {}",
						new NcpdpCobOtPayerLogger(rowCobOtPayer));

				if (rowCobOtPayer != null) {

					ohiOutput.append(delimit(getValue(rowCobOtPayer
							.getOtPayerCvgType(), "", false, -1))); // OT Payer
					// Coverage Type
					ohiOutput.append(delimit(getValue(rowCobOtPayer
							.getOtherPayerId(), "", false, -1))); // Other Payer
					// Id
					ohiOutput.append(delimit(getValue(rowCobOtPayer
							.getOtherPayerDate(), "", false, -1))); // Other
					// Payer
					// Date
					vExtFile.println(ohiOutput.toString());
					createOHIPayments(dbEM, rowCobOtPayer, vExtFile);
					createRejectRecords(dbEM, rowCobOtPayer, vExtFile);
					createOHIPayerPatientResposibilityRecords(dbEM,
							rowCobOtPayer, vExtFile);
					createCOBBenifitStageRecords(dbEM, rowCobOtPayer, vExtFile);

					detach(rowCobOtPayer, dbEM);
				} else {
					// rowCobOtPayer is null
					ohiOutput.append(delimit("")); // Amount Paid
					vExtFile.println(ohiOutput.toString());
				}
				vExtFile.flush();
				ohiOutput.delete(0, ohiOutput.length());
			}
		}

	}

	private void createOHIPayments(EntityManager dbEM,
			NcpdpCobOtPayer rowCobOtPayer, PrintWriter vExtFile) {
		Set cobOtPayerAmounts = rowCobOtPayer.getNcpdpCobOtPayerAmounts();
		if (cobOtPayerAmounts != null) {
			Iterator cobOtPayerAmountIterator = cobOtPayerAmounts.iterator();
			while (cobOtPayerAmountIterator.hasNext()) {
				StringBuffer ohiPaymentOutput = new StringBuffer();
				NcpdpCobOtPayerAmount rowCobOtPayerAmount = (NcpdpCobOtPayerAmount) cobOtPayerAmountIterator
						.next();
				traceLOGGER.trace("COB Other Payer Amount: {}",
						new NcpdpCobOtPayerAmountLogger(rowCobOtPayerAmount));
				if (rowCobOtPayerAmount != null) {
					if (rowCobOtPayerAmount.getOtPyrAmtPdQual() != null) {
						ohiPaymentOutput.append(delimit("42"));
						BigDecimal otpyrAmtPaid = rowCobOtPayerAmount
								.getOtPyrAmtPaid();
						ohiPaymentOutput.append(delimit(rowCobOtPayerAmount
								.getOtPyrAmtPdQual()));
						ohiPaymentOutput.append(delimit(getValue(
								(otpyrAmtPaid != null) ? otpyrAmtPaid
										.toString() : null, "", false, -1)));
						vExtFile.println(ohiPaymentOutput.toString());
					}
					detach(rowCobOtPayerAmount, dbEM);
				}

			}
		}
	}

	private void createRejectRecords(EntityManager dbEM,
			NcpdpCobOtPayer rowCobOtPayer, PrintWriter vExtFile) {
		Set cobOtPayerRejects = rowCobOtPayer.getNcpdpCobOtPayerRejects();
		if (cobOtPayerRejects != null) {
			Iterator cobOtPayerRejectIterator = cobOtPayerRejects.iterator();
			while (cobOtPayerRejectIterator.hasNext()) {
				NcpdpCobOtPayerReject rowCobOtPayerReject = (NcpdpCobOtPayerReject) cobOtPayerRejectIterator
						.next();
				if (rowCobOtPayerReject != null) {
					StringBuffer rejectsOutput = new StringBuffer();
					rejectsOutput.append(delimit("43"));
					traceLOGGER.trace("COB Other Payer Reject: {}",
							rowCobOtPayerReject);
					rejectsOutput.append(delimit(getValue(rowCobOtPayerReject
							.getOtPyrRejCd(), "", false, -1))); // Ot Payer
					// Reject
					// Code
					detach(rowCobOtPayerReject, dbEM);
					vExtFile.println(rejectsOutput.toString());
				}
			}
		}
	}

	private void createOHIPayerPatientResposibilityRecords(EntityManager dbEM,
			NcpdpCobOtPayer rowCobOtPayer, PrintWriter vExtFile) {
		Set cobOtPayerPatientResposibilitityAmounts = rowCobOtPayer
				.getNcpdpCobOpPatRespAmounts();
		if (cobOtPayerPatientResposibilitityAmounts != null) {
			Iterator cobOtPayerPatientResposibilitityAmountIterator = cobOtPayerPatientResposibilitityAmounts
					.iterator();
			while (cobOtPayerPatientResposibilitityAmountIterator.hasNext()) {
				StringBuffer ohiPaymentOutput = new StringBuffer();
				NcpdpCobOpPatRespAmount rowCobOtPayerPatientResposibilitityAmount = (NcpdpCobOpPatRespAmount) cobOtPayerPatientResposibilitityAmountIterator
						.next();

				if (rowCobOtPayerPatientResposibilitityAmount != null) {
					if (rowCobOtPayerPatientResposibilitityAmount
							.getOpPatRespAmntPdQual() != null) {
						ohiPaymentOutput.append(delimit("44"));
						BigDecimal patRespAmount = rowCobOtPayerPatientResposibilitityAmount
								.getOpPatRespAmnt();
						ohiPaymentOutput
								.append(delimit(rowCobOtPayerPatientResposibilitityAmount
										.getOpPatRespAmntPdQual()));
						ohiPaymentOutput.append(delimit(getValue(
								(patRespAmount != null) ? patRespAmount
										.toString() : null, "", false, -1)));
						vExtFile.println(ohiPaymentOutput.toString());
					}
					detach(rowCobOtPayerPatientResposibilitityAmount, dbEM);
				}

			}
		}
	}

	private void createCOBBenifitStageRecords(EntityManager dbEM,
			NcpdpCobOtPayer rowCobOtPayer, PrintWriter vExtFile) {
		Set cobBeneStageAmounts = rowCobOtPayer.getNcpdpCobBeneStgAmounts();
		if (cobBeneStageAmounts != null) {
			Iterator cobBeneStageAmountIterator = cobBeneStageAmounts
					.iterator();
			while (cobBeneStageAmountIterator.hasNext()) {
				StringBuffer ohiPaymentOutput = new StringBuffer();
				NcpdpCobBeneStgAmount rowCobBeneStageAmount = (NcpdpCobBeneStgAmount) cobBeneStageAmountIterator
						.next();

				if (rowCobBeneStageAmount != null) {
					if (rowCobBeneStageAmount.getBeneStgQual() != null) {
						ohiPaymentOutput.append(delimit("45"));
						BigDecimal beneStageAmount = rowCobBeneStageAmount
								.getBeneStgAmnt();
						ohiPaymentOutput.append(delimit(rowCobBeneStageAmount
								.getBeneStgQual()));
						ohiPaymentOutput.append(delimit(getValue(
								(beneStageAmount != null) ? beneStageAmount
										.toString() : null, "", false, -1)));
						vExtFile.println(ohiPaymentOutput.toString());
					}
					detach(rowCobBeneStageAmount, dbEM);
				}

			}
		}
	}

	public void initializeVariables() {
		iPrvCnt = 0;
		iClmCnt = 0;
		vDollarCnt = zeroBD;
	}

	public void detach(Object entity, EntityManager dbEM) {
		if (entity == null)
			return;
		org.hibernate.Session session = (Session) dbEM.getDelegate();
		session.evict(entity);
	}

	public String delimit(String source) {
		return source + DELIMITER;
	}

	// convenience method that is often used
	public String delimit(final BigDecimal source) {
		if (source == null) {
			return DELIMITER;
		}
		return source.toString() + DELIMITER;
	}

	// convenience method that is often used
	public String delimit(final Long source) {
		if (source == null) {
			return DELIMITER;
		}
		return source.toString() + DELIMITER;
	}

	// convenience method that is seldom used

	// Classes used to get rid of Globals
	private class ReturnVariables {

		public List<Claim> listOfClaimsToReturn;
		String vTransRefNum = "";

		public List<Claim> getListOfClaimsToReturn() {
			return listOfClaimsToReturn;
		}

		public void setListOfClaimsToReturn(List<Claim> listOfClaimsToReturn) {
			this.listOfClaimsToReturn = listOfClaimsToReturn;
		}

		public String getvTransRefNum() {
			return vTransRefNum;
		}

		public void setvTransRefNum(String vTransRefNum) {
			this.vTransRefNum = vTransRefNum;
		}
	}

	/**
	 * This method takes a string and returns a default value if the string is
	 * null. If shouldTrimString it returns a value that is trimSize or shorter
	 * in length - Note: the default value is never trimmed. This is simply a
	 * refactoring of several blocks of code that existed in the past.
	 * 
	 * @param value
	 * @param defaultValue
	 * @param shouldTrimString
	 * @param trimSize
	 * @return
	 */
	private static String getValue(String value, String defaultValue,
			boolean shouldTrimString, int trimSize) {
		if (value == null)
			return defaultValue;

		if (shouldTrimString) {
			if (value.length() > trimSize) {
				return value.substring(0, trimSize);
			} else {
				return value;
			}
		} else {
			return value;
		}
	}

	private void createRepeatedDelimitedFields(StringBuffer buffer,
			String repeatedValue, int numDelimiters) {
		for (int i = 0; i < numDelimiters; i++) {
			buffer.append(delimit(repeatedValue));
		}
	}
	
	private String getDecimalFormattedQuantity(Long amount)
	{
		String returnString = "";
		if(amount != null)		
		{
			returnString = getDecimalFormattedQuantity(new BigDecimal(amount));
		}
		return returnString;		
	}
	
	private String getDecimalFormattedQuantity(BigDecimal amount)
	{
		String returnString = "";
		if(amount != null)		
		{
			BigDecimal quantityAsBD = amount;
			BigDecimal quantityDiv1000 = quantityAsBD.divide(new BigDecimal(1000));				
			returnString = String.format("%.3f", quantityDiv1000);
		}
		return returnString;
	}

}
