package gov.va.med.hac.edi.extract.b1.jpa;

import gov.va.med.hac.edi.utils.jpa.*;

import org.slf4j.*;
import org.apache.commons.cli.*;

import javax.persistence.EntityManager;
import javax.persistence.FlushModeType;

import java.io.File;

/**
 * There are two different proprietary file formats that are created during extract file processing from eTrans Repository:
 *	NCPDP B1 Claims are extracted into a modified WPS format.  
 *	NCPDP B2 Reversals are extracted into a reversal flat file format
 * This process extracts B1 Claims and creates the WPS format.
 * 
 * 
 * The general flow of RxExtractB1 is explained below.
 * 
 * The main method creates a new log file according to the SLF4J logging configuration specified in resources/logback.xml - 
 * the name of the log file may be specified by passing java a system varable e.g. java -Dapp_name=RxExtractB1 
 * gov.va.med.hac.edi.extract.b1.jpa. RxExtractB1.
 * 
 * Next, it reads the command line parameters and if any of the required command line parameters (-s or -p) are missing
 * the program aborts.
 * 
 * If the proper parameters have been passed to the application an entity manager is created that will be used
 * to control all interactions with the database.
 * 
 * Next, the processExtract method on ProcessRxExtractB1 is called to physically create the extract file. processExtract
 * contains the logic to extract data from the database and creates the WPS.
 * 
 * Last, processExtract handles any errors by logging them.
 * 
 * Please see the ProcessRxExtractB1 class comments to understand how the WPS file is created.
 *   
 *  
Main method calls the overloaded constructer RxExtractB1 passing the Directory Path and LogFile as parameters.

 * @author HIPAA-READY
 * 
 */
public class RxExtractB1 {
  // Use two loggers, one for a summary file and one for a trace file.
  static final Logger summaryLOGGER = LoggerFactory.getLogger(RxExtractB1.class);
  static final Logger traceLOGGER   = LoggerFactory.getLogger("trace." + RxExtractB1.class);

  // private constructor to prevent unneeded instance creation.
  private RxExtractB1() {
  }

  /**
   * Entry point for the RxExtractB1 process. The JPA connection configuration file name can be
   * passed into this method which will override the normal persistence.xml file found in the
   * META-INF directory of the jar file. The sender id and full path to the output file must also be set.
   * After running the extract an extract file named "WPS_<sender_id><CurrentDate>_<CurrentTime>.DAT
   * will be created in the directory specified by the -p command line parameter.
   * <br>
   * <code>usage: RxExtractB1 [-f &lt;file&gt;] [-h] -s sender -p output_path </code>
   * 
   * @param args
   *          set of options where &lt;file&gt; is the full path name of the JPA configuration file.
   */
  public static void main(String[] args) {
    // Parse the command line
    CommandLine line = null;
    Options opts = getCliOptions();
    try {
      CommandLineParser parser = new GnuParser();
      line = parser.parse(opts, args);
      // If the "help" option is specified, just show usage options and exit
      if (line.hasOption("h")) {
        printUsage(opts, args, false);
        System.exit(1);
      }
      // If any required command line parameter is missing, just show usage options and exit
    } catch (MissingOptionException moe) {
      printUsage(opts, args, true);
      System.exit(1);
    } catch (ParseException pe) {
      System.out.println("Problem parsing the command line");
      printUsage(opts, args, false);
      System.exit(1);
    }

    // Local variables
    boolean processSuccess = false;
    String destinationPath = "";
    String senderId = "";
    String configFilePath = "";
    EntityManager em = null;
    
    // Catch exceptions, log them, and swallow them to allow other things to run afterwards
    try {
      // Put an origination message into the log file
      LogUtils.stampStartingTime("Rx Extract B1 Start", summaryLOGGER);
      summaryLOGGER.info("Code Version: 5010 Project");

      // Get an EntityManager, based on the -f option if passed in, otherwise based on a default
      File configFile = null;
      configFilePath = line.getOptionValue("f");
      if (configFilePath != null && configFilePath.length() > 0) {
        configFile = JPAUtil.getJpaConfigFile(configFilePath);
      } else {
        configFile = JPAUtil.getJpaConfigFile("dbhibernate.properties");
      }
      em = JPAUtil.getEntityManager(configFile, "b1b2Extract");
      em.setFlushMode(FlushModeType.COMMIT);
      // If parameters are OK, run this process
      destinationPath = line.getOptionValue("p");
      senderId = line.getOptionValue("s");
      if (destinationPath != null && destinationPath.length() > 0 && senderId != null
          && senderId.length() > 0) {
        summaryLOGGER.info("Destination Path from parameters: " + destinationPath);
        summaryLOGGER.info("Sender Id from parameters: " + senderId);
        ProcessRxExtractB1 pExtractB1 = new ProcessRxExtractB1();
        pExtractB1.processExtract(destinationPath, senderId, em);

      } else {
        summaryLOGGER.error("RxExtractB1: Please Pass both Parameters");
      }

      // Put a completion message into the log file
      summaryLOGGER.info("Process completed successfully");
      LogUtils.stampEndingTime("RxExtractB1 process completed",
          summaryLOGGER);

      processSuccess = true;

    } catch (OutOfMemoryError e) {
      summaryLOGGER.error("!!!!! Out Of Memory Error !!!!!", e);
    } catch (Throwable t) {
      summaryLOGGER.error("Exception or Error Thrown During Processing", t);
    } finally {
      if (!processSuccess) {
        summaryLOGGER.error("!!!!! Process seems to have failed !!!!!");
      }
    }
  }

  /**
   * Helper method to populate command line options objects.
   * 
   * @return the <code>Options</code> object containing the set of command line option.
   */
  @SuppressWarnings("static-access")
  private static Options getCliOptions() {
    Options options = new Options();

    Option configfile = OptionBuilder.withArgName("jpaFile").hasArg().withLongOpt("jpaFile")
        .withDescription("The JPA Configuration file, such as hibernate.properties.").create("f");

    Option destinationPath = OptionBuilder.withArgName("destinationPath").hasArg().isRequired().
        withLongOpt("destinationPath").withDescription("The path to output files.").create("p");

    Option senderId = OptionBuilder.withArgName("senderId").hasArg().isRequired().
        withLongOpt("senderId").withDescription(
            "The identifier of the Pharmacy Benefits Manager "
                + "whose claims should be extracted.").create("s");

    Option help = OptionBuilder.withLongOpt("help").withDescription("Print this message.").create(
        "h");

    options.addOption(help);
    options.addOption(configfile);
    options.addOption(destinationPath);
    options.addOption(senderId);

    return options;
  }

  private static void printUsage(Options opts, String[] inArgs, boolean missingArg) {
    final int helpwidth = 80;
    HelpFormatter fmtr = new HelpFormatter();
    System.out.println();
    if (missingArg) {
      System.out.println("Some required commandline argument was missing");
    }
    fmtr.printHelp(helpwidth, "RxExtractB1", "", opts, "", true);
    if (missingArg) {
      System.out.println(inArgs.length + " command line arguments were received: ");
      for (int i = 0; i < inArgs.length; i++) {
        System.out.println(inArgs[i]);
      }
    }
  }

}
