/** $Id: EmdeonPendingClaimStatus.java,v 1.2 2012/02/23 17:56:21 vhahacpennet Exp $ */
package gov.va.med.hac.edi.extract.feeclaimstatus.pending;

import gov.va.med.hac.edi.utils.jpa.JPAUtil;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.MissingOptionException;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;

import java.text.SimpleDateFormat;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import javax.naming.ConfigurationException;

import javax.persistence.EntityManager;


/**
 * Performs the required Fee business edits.
 *
 */
public class EmdeonPendingClaimStatus {
    private static final String HELP_OPT = "h";
    private static final String E_REPOS_CONFIG_OPT = "r";
    private static final String FPPS_CONFIG_OPT = "f";
    private static final String PROGRAM_TRACKER_CONFIG_OPT = "p";
    private static final String OUTPUT_OPT = "o";
    private static final String SENDER_ID_OPT = "s";
    private static final String QUARTERLY_OPT = "q";
    private static final String BEGIN_DATE = "b";
    private static final String END_DATE = "e";

    /** the class logger. */
    static final Logger LOGGER = LoggerFactory.getLogger(EmdeonPendingClaimStatus.class);

    /** private constructor to prevent unneeded instance creation. */
    private EmdeonPendingClaimStatus() {
    }

    /**
     * Entry point for the EmdeonClaimStatus process. The JPA connection
     * configuration file name can be passed into this method which will
     * override the normal persistence.xml file found in the META-INF directory
     * of the jar file. <br>
     * <br>
     * <code>usage: EmdeonClaimStatus [-b &lt;beginDate&gt;] [-e
     * &lt;endDate&gt;] -f &lt;file&gt; [-h] -o
     * &lt;outputdir&gt; [-q] -s &lt;senderid&gt;</code>
     *
     * @param args
     *            set of options where &lt;file&gt; is the full path name of the
     *            JPA configuration file.
     */
    public static void main(String[] args) {
        LOGGER.info("\n\n");
        LOGGER.info(
            "<<<<<<<<<<<<< Beginning Claim Status Creation <<<<<<<<<<<<<<");

        final int quarterYear = 90;
        EntityManager eReposEM = null;
        EntityManager fppsEM = null;
        EntityManager programTrackingEM = null;
        Options opts = getCliOptions();
        SimpleDateFormat datefmtr = null;

        try {
            CommandLineParser parser = new GnuParser();
            CommandLine line = parser.parse(opts, args);

            if (line.hasOption(HELP_OPT)) {
                printUsage(opts);
                System.exit(1);
            }

            File eReposConfigFile = null;

            if (line.hasOption(E_REPOS_CONFIG_OPT)) {
                eReposConfigFile = JPAUtil.getJpaConfigFile(line.getOptionValue(
                            E_REPOS_CONFIG_OPT));
            }

            if (eReposConfigFile != null) {
                eReposEM = JPAUtil.getEntityManager(eReposConfigFile);

                if (eReposEM == null) {
                    throw new ConfigurationException(
                        "Unable to configure EREPOS Entity Manager.");
                }
            }

            File fppsConfigFile = null;

            if (line.hasOption(FPPS_CONFIG_OPT)) {
                fppsConfigFile = JPAUtil.getJpaConfigFile(line.getOptionValue(
                            FPPS_CONFIG_OPT));
            }

            if (fppsConfigFile != null) {
                fppsEM = JPAUtil.getEntityManager(fppsConfigFile);

                if (fppsEM == null) {
                    throw new ConfigurationException(
                        "Unable to configure FPPS Entity Manager.");
                }
            }

            File programTrackerConfigFile = null;

            if (line.hasOption(PROGRAM_TRACKER_CONFIG_OPT)) {
                programTrackerConfigFile = JPAUtil.getJpaConfigFile(line.getOptionValue(
                            PROGRAM_TRACKER_CONFIG_OPT));
            }

            if (programTrackerConfigFile != null) {
                programTrackingEM = JPAUtil.getEntityManager(programTrackerConfigFile);

                if (fppsEM == null) {
                    throw new ConfigurationException(
                        "Unable to configure PROGRAM_TRACKER Entity Manager.");
                }
            }

            String outputDir = null;

            if (line.hasOption(OUTPUT_OPT)) {
                outputDir = line.getOptionValue(OUTPUT_OPT);
            }

            String senderId = null;

            if (line.hasOption(SENDER_ID_OPT)) {
                senderId = line.getOptionValue(SENDER_ID_OPT);
            }

            Date beginDate = null;
            Date endDate = null;
            boolean quarterly = false;

            if (line.hasOption(QUARTERLY_OPT)) {
                quarterly = true;
                datefmtr = new SimpleDateFormat("dd-MMM-yyyy");

                if (line.hasOption(BEGIN_DATE)) {
                    String beginDateStr = line.getOptionValue(BEGIN_DATE);
                    beginDate = datefmtr.parse(beginDateStr);
                } else {
                    Calendar cal = new GregorianCalendar();
                    cal.add(Calendar.DATE, -quarterYear);
                    beginDate = cal.getTime();
                }

                if (line.hasOption(END_DATE)) {
                    String endDateStr = line.getOptionValue(END_DATE);
                    endDate = datefmtr.parse(endDateStr);
                } else {
                    Calendar cal = new GregorianCalendar();
                    endDate = cal.getTime();
                }
            }

            ProcessEmdeonPendingClaimStatus pecs = new ProcessEmdeonPendingClaimStatus();

            if (quarterly) {
                pecs.process(senderId, outputDir, beginDate, endDate, eReposEM,
                    fppsEM, programTrackingEM);
            } else {
                pecs.process(senderId, outputDir, eReposEM, fppsEM,
                    programTrackingEM);
            }
        } catch (MissingOptionException moe) {
            printUsage(opts);
        } catch (IllegalArgumentException iae) {
            System.out.println(iae.getMessage());
            printUsage(opts);
        } catch (Throwable t) {
            LOGGER.error("Unable to Create Fee Claim Status file.", t);
        } finally {
            if ((eReposEM != null) && eReposEM.isOpen()) {
                eReposEM.close();
            }

            if ((fppsEM != null) && fppsEM.isOpen()) {
                fppsEM.close();
            }

            if ((programTrackingEM != null) && programTrackingEM.isOpen()) {
                programTrackingEM.close();
            }

            LOGGER.info(
                "<<<<<<<<<<<<< Ending Claim Status Creation <<<<<<<<<<<<<<");
        }
    }

    static final void printUsage(Options opts) {
        final int helpwidth = 80;
        HelpFormatter fmtr = new HelpFormatter();
        System.out.println();
        fmtr.printHelp(helpwidth, "EmdeonClaimStatus", "", opts, "", true);
    }

    /**
     * Helper method to populate command line options objects.
     *
     * @return the <code>Options</code> object containing the set of command
     *         line option.
     */
    @SuppressWarnings("static-access")
    static Options getCliOptions() {
        Options options = new Options();

        Option eReposConfigfile = OptionBuilder.withArgName("eRepos")
                                               .isRequired().hasArg()
                                               .withLongOpt("ereposconfigfile")
                                               .withDescription("The eRepos EntityManager Configuration file.")
                                               .create(E_REPOS_CONFIG_OPT);

        Option fppsConfigFile = OptionBuilder.withArgName("fpps").isRequired()
                                             .hasArg()
                                             .withLongOpt("fppsconfigfile")
                                             .withDescription("The FPPS EntityManager Configuration file.")
                                             .create(FPPS_CONFIG_OPT);

        Option programTrackerConfigFile = OptionBuilder.withArgName(
                "programTracker").isRequired().hasArg()
                                                       .withLongOpt("programtrackerconfigfile")
                                                       .withDescription("The FPPS EntityManager Configuration file.")
                                                       .create(PROGRAM_TRACKER_CONFIG_OPT);

        Option outputdir = OptionBuilder.withArgName("outputdir").isRequired()
                                        .hasArg().withLongOpt("outputDir")
                                        .withDescription("The CSTAT Output directory.")
                                        .create(OUTPUT_OPT);

        Option senderId = OptionBuilder.withArgName("senderid").isRequired()
                                       .hasArg().withLongOpt("senderId")
                                       .withDescription("The Sender's Identification.")
                                       .create(SENDER_ID_OPT);

        Option quarterly = OptionBuilder.withLongOpt("quarterly")
                                        .withDescription("flag to set quarterly CSTAT file creation.")
                                        .create(QUARTERLY_OPT);

        Option beginDateStr = OptionBuilder.withArgName("beginDate")
                                           .withLongOpt("begindate").hasArg()
                                           .withDescription("Date to Begin the Quarterly report (dd-MMM-yyyy).")
                                           .create(BEGIN_DATE);

        Option endDateStr = OptionBuilder.withArgName("endDate")
                                         .withLongOpt("enddate").hasArg()
                                         .withDescription("Ending Date for the Quarterly report (dd-MMM-yyyy).")
                                         .create(END_DATE);

        Option help = OptionBuilder.withLongOpt("help")
                                   .withDescription("Print this message.")
                                   .create(HELP_OPT);

        options.addOption(help);
        options.addOption(eReposConfigfile);
        options.addOption(fppsConfigFile);
        options.addOption(programTrackerConfigFile);
        options.addOption(outputdir);
        options.addOption(senderId);
        options.addOption(beginDateStr);
        options.addOption(endDateStr);
        options.addOption(quarterly);

        return options;
    }
}
