package gov.va.med.hac.edi.extract.x12.jpa;
 
import gov.va.med.hac.edi.utils.jpa.DateUtil;
import gov.va.med.hac.edi.utils.jpa.DbTransactionEnderException;
import gov.va.med.hac.edi.utils.jpa.JPAUtil;

import java.io.*;
import java.util.*;

import javax.naming.NamingException;
import javax.persistence.EntityManager;
import javax.persistence.FlushModeType;
import javax.persistence.Query;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.lang.reflect.Method;
import java.math.BigDecimal;
import gov.va.med.domain.erepos.*;

/**
 * Contains methods that query the data for dental claim information for a clearing house and writes
 * a dental claim extract file for the clearing house named <cleainghouseid>_den_
 * 
 * The class extends process cacheRecords.
 * 
 * @author Hipaa-Ready
 * 
 */
public class ExtractDntlClaims extends ProcessCacheRecords {
  static final Logger summaryLOGGER = LoggerFactory.getLogger(ExtractDntlClaims.class);
  static final Logger traceLOGGER = LoggerFactory.getLogger("trace." + ExtractDntlClaims.class);
  static int PAGE_SIZE = 40;

  
  /**
   * This method drives the creation of the dental claim extract file.
   * 
   * It creates an extract file named <clearing house id> _den_
   * 
   * @param vChId
   *          - The id of the clearing house
   * @param vFilePath
   *          - the path where the extract file will be created
   * @param receiverIdStart
   *          -
   * @param vLogFile
   * @throws DbTransactionEnderException
   */
  protected void processDntlClaims(String vChId, String vFilePath,
      String receiverIdStart, boolean useClones, EntityManager dbEM) 
  throws DbTransactionEnderException {
    try {
    	
      final String fileName = createFileName(vChId, "_den_");

      File DntlFile = createTemporaryFileAndDir(vFilePath, fileName);
      if(DntlFile == null)
      {
    	  throw new IOException("Could not create a temporary file or directory to write the extract file to. " + vFilePath + "tmp : "+ fileName);
      }
      
      List claimKeysToUpdate = new ArrayList();
      int claimCount = 0;
      PrintWriter vDntlFile = null;
      
      // Process Dental claims - 97A1 - see - EDI Transaction Set Implementation Guide
        Query getClaimQuery = dbEM
            .createQuery("SELECT c FROM gov.va.med.domain.erepos.Claim c "+
            			 "WHERE (c.claimType = :claimType) AND c.statusCode=:statusCode AND c.senderId=:senderId "+
            			 "AND (c.receiverId LIKE :receiverId) ORDER BY c.dateCreated, c.fileNumber, c.st02txnCtlNbr");;
        getClaimQuery.setParameter("claimType", "005010X224");
        getClaimQuery.setParameter("statusCode", "NEW");
        getClaimQuery.setParameter("senderId", vChId);
        getClaimQuery.setParameter("receiverId", receiverIdStart + "%");

        List<Claim> claims = getClaimQuery.getResultList();
        Iterator<Claim> claimIterator;
		int numClaims = claims.size();
		summaryLOGGER.info("Number of Dental claims: " + claims.size());

		if(useClones)
		{
			List<Claim> clonedClaims = cloneAndClearClaims(claims);			
			claims = null;
			dbEM.clear();				
			System.gc();
			claimIterator = clonedClaims.iterator();
		}
		else
		{
			claimIterator = claims.iterator();
		}						
        
        if (numClaims > 0) {
            vDntlFile = new PrintWriter(new FileOutputStream( DntlFile));
        
          while (claimIterator.hasNext()) // -- loop for each claim
          {
            final Claim rowClaim = (Claim) claimIterator.next();
            final String vFileNumber = rowClaim.getFileNumber();
            final String vSt02txnCtlNbr = rowClaim.getSt02txnCtlNbr();
            writeARec(vFileNumber, vSt02txnCtlNbr, vDntlFile, dbEM);            
            writeBRec(rowClaim, vDntlFile, dbEM);
            writeCRec(rowClaim, vDntlFile, dbEM);
            writeERec(rowClaim, vDntlFile, dbEM);
            writeFRec(rowClaim, vDntlFile, dbEM);
            writeA999Rec(vDntlFile);
            claimKeysToUpdate.add(rowClaim.getClaimKey());  
			if(((50+claimCount) % 50) == 0)
			{							
				if(useClones) dbEM.clear();
				vDntlFile.flush();
				traceLOGGER.debug("cc= " + claimCount + " " + DateUtil.getCurrentTimeFormatted("hh:mm:ss"));
			}		
			claimCount++;
          }
          
          vDntlFile.flush();
          vDntlFile.close();
      }
        
      dbEM.clear();
      traceLOGGER.debug("Committing the database transaction");
      dbEM.getTransaction().begin();
      Iterator claimToPersistIterator = claimKeysToUpdate.iterator();
      Query deleteStatement;
      while (claimToPersistIterator.hasNext()) {
        deleteStatement = dbEM.createNativeQuery("update claim set status_code=:statusCode where claim_key=:claimKey");
        deleteStatement.setParameter("statusCode", "BUILT");
        deleteStatement.setParameter("claimKey", (String) claimToPersistIterator.next());
        deleteStatement.executeUpdate();
      }
      dbEM.getTransaction().commit();
      traceLOGGER.debug("Commit successful");
      
      File renamedFile = new File(vFilePath + fileName);
      DntlFile.renameTo(renamedFile);
      traceLOGGER.debug("Was able to rename the file to " + renamedFile.getAbsolutePath());
      
      traceLOGGER.debug("Dental Claims Processed : " + claimCount);
    } // -- end write Dntl clmains
    catch (Exception ex) {
      traceLOGGER
          .error("There was an Exception in writing Dental Claims ");
      traceLOGGER.debug("Rolling back the database transaction");
      if (dbEM != null) {
        if (dbEM.getTransaction().isActive()) {
          dbEM.getTransaction().rollback();
        }
      }
      traceLOGGER.debug("Rollback successful");
      throw new DbTransactionEnderException(
          "Rolled Back the Transaction", ex);
    }
  } // end processDntlClaims

  /**
   * Gets subscriber (insured) information and the destination payer data and writes it to the
   * extract file.
   * 
   * @param rowClaim
   *          - the claim related to the subscriber
   * @param vDntlFile
   *          - the extract file to write the data to
   * @param dbEM
   *          - the entity manager used to query the database
   */
  private void writeCRec(Claim rowClaim, PrintWriter vDntlFile,
      EntityManager dbEM) {
    String vSbrSecId = "";
    final String vDtOfDeath = "";
    final String vFileNumber = rowClaim.getFileNumber();
    final String vSt02txnCtlNbr = rowClaim.getSt02txnCtlNbr();
    final String vSbsrId = rowClaim.getSubscriberId();
    final String vSubscriberHrchlLevelKey = rowClaim.getSubscriberHrchlLevelKey();
    
    Query getSubscriberInfoQuery = dbEM
        .createQuery("SELECT si FROM gov.va.med.domain.erepos.SubscriberInfo si WHERE si.txnSet.id.fileNumber=:fileNumber AND si.txnSet.id.st02txnCtlNbr=:st02txnCtlNbr AND si.nm109sbsrId=:nm109sbsrId AND si.loopId=:loopId  AND si.hrchlLevelKey=:hrchlLevelKey");
    getSubscriberInfoQuery.setParameter("fileNumber", vFileNumber);
    getSubscriberInfoQuery.setParameter("st02txnCtlNbr", vSt02txnCtlNbr);
    getSubscriberInfoQuery.setParameter("nm109sbsrId", vSbsrId);
    getSubscriberInfoQuery.setParameter("loopId", "2000B");
    getSubscriberInfoQuery.setParameter("hrchlLevelKey", vSubscriberHrchlLevelKey);
    final SubscriberInfo rowSubscriberInfo = (SubscriberInfo) getSubscriberInfoQuery
        .getResultList().get(0);
    
    final String vWhereCls = "FILE_NUMBER = '" + vFileNumber + "' "
    + "AND ST02TXN_CTL_NBR = '" + vSt02txnCtlNbr + "' "
    + "AND NM109SBSR_ID = '" + vSbsrId + "' "
    + "AND HRCHL_LEVEL_KEY ='" + vSubscriberHrchlLevelKey +"'";

    //5010: ADDED NM101ENTY_ID_CD
    EntityName rowEntityName = getEntityName(vWhereCls
            + "AND LOOP_ID = '2010BA' AND NM101ENTY_ID_CD='IL'", dbEM);

    //5010: added HRCHL_LEVEL_KEY
    EntityAddress rowEntityAddress = null;
    rowEntityAddress = null;
    if(rowEntityName != null)
	{
    	rowEntityAddress = getEntityAddress(vWhereCls
            + "AND LOOP_ID = '2010BA' AND HRCHL_LEVEL_KEY='"+rowEntityName.getHrchlLevel().getHrchlLevelKey()+"'", dbEM);
	}

    //5010: Changed REF01ID_QLFR='IL' to REF01ID_QLFR='SY'
    vSbrSecId = getRefValue(vWhereCls + "AND LOOP_ID = '2010BA' AND REF01ID_QLFR='SY'", dbEM);
    
    Query getDemographicInfoQuery = dbEM
    .createQuery("SELECT di FROM gov.va.med.domain.erepos.DemographicInfo di WHERE di.txnSet.id.fileNumber=:fileNumber AND di.txnSet.id.st02txnCtlNbr=:st02txnCtlNbr AND di.nm109sbsrId=:nm109sbsrId AND di.loopId=:loopId  AND di.hrchlLevelKey=:hrchlLevelKey");
    getDemographicInfoQuery.setParameter("fileNumber", vFileNumber);
    getDemographicInfoQuery.setParameter("st02txnCtlNbr", vSt02txnCtlNbr);
    getDemographicInfoQuery.setParameter("nm109sbsrId", vSbsrId);
    getDemographicInfoQuery.setParameter("loopId", "2010BA");
    getDemographicInfoQuery.setParameter("hrchlLevelKey", vSubscriberHrchlLevelKey);
	DemographicInfo rowDemographicInfo = null;
	if(getDemographicInfoQuery.getResultList().size() > 0)
	{
		rowDemographicInfo = (DemographicInfo) getDemographicInfoQuery.getResultList().get(0);
	}

    writeC000Rec(vSbrSecId, vDtOfDeath, rowSubscriberInfo, rowEntityName,
        rowEntityAddress, rowDemographicInfo, vDntlFile);

    rowEntityName = getEntityName(vWhereCls + "AND LOOP_ID = '2010BB' ",
        dbEM);

    writeC005Rec(rowEntityName, vDntlFile);
    vDntlFile.flush();
  } // end writeCRec

  /**
   * Gets basic claim information and writes it to the extract file.
   * 
   * @param rowClaim
   *          - the claim
   * @param vOutputFile
   *          - the extract file
   * @param dbEM
   *          - entity manager attached to eRepository
   */
  private void writeERec(Claim rowClaim, PrintWriter vOutputFile,
      EntityManager dbEM) {
    final String vClaimKey = rowClaim.getClaimKey();
    final String vFileNumber = rowClaim.getFileNumber();
    final String vSt02txnCtlNbr = rowClaim.getSt02txnCtlNbr();

    String vTPClaimRecId = getRefValue(vClaimKey, "2300", "D9", dbEM);
    
    // E000 Record can already be written
    writeE000Rec(rowClaim, "C", vTPClaimRecId, vOutputFile);

    // E005 Record has a little logic
    processE005Rec(rowClaim, vOutputFile, dbEM);
    
    processE006Rec(rowClaim, vOutputFile, dbEM);

    processE007Rec(rowClaim, vOutputFile, dbEM);
    
    // E009 Record
    processE009Rec(vClaimKey, vOutputFile, dbEM);

    // Write the E010 record if the total billed is positive
    final BigDecimal vTotBilled = rowClaim.getClm02clmChrgAmt();
    final BigDecimal vPatAmtPaid = getClaimAmt(vClaimKey, "F5", dbEM);
    final BigDecimal vPatBalDue = getClaimAmt(vClaimKey, "F3", dbEM);
    
    /**
     *  HRL - TWP: Making a change similar to DEF009627 - 
     *  Prev: if (vTotBilled.signum() == 1) This was allowing only
     * 	non-zero, positive billed amounts. Zero-dollar claim amounts are valid
     */
    if (vTotBilled.signum() >= 0) {
      writeE010Rec(vTotBilled, vPatAmtPaid, vPatBalDue, vOutputFile);
    }

    // E015 Record is complicated
    processE015Rec(vClaimKey, vOutputFile, dbEM);
    
    //Write the health information codes to the output file E100, E105
    processHiCodesRec(rowClaim, vOutputFile, dbEM);

    // The following is the code to create E205 Record - Service Location
    EntityName rowEntityName = null;
    EntityAddress rowEntityAddress = null;
    String vEntWhereClause = "CLAIM_KEY = '" + vClaimKey + "' "
        + "AND NM101ENTY_ID_CD = '77' ";   //5010: Changed FA to 77
    rowEntityName = getEntityName(vEntWhereClause
        + "AND LOOP_ID = '2310C' ", dbEM);

    rowEntityAddress = null;
    if(rowEntityName != null)
	{
    	final String vAdrWhereClause = "CLAIM_KEY = '" + vClaimKey + "' AND HRCHL_LEVEL_KEY='"+rowEntityName.getHrchlLevel().getHrchlLevelKey()+"'"; //5010: added HRCHL_LEVEL_KEY
    	rowEntityAddress = getEntityAddress(vAdrWhereClause
        + "AND LOOP_ID = '2310C' ", dbEM);
	}

    if ((rowEntityName == null) || (rowEntityAddress == null)) {
      vEntWhereClause = "FILE_NUMBER = '" + vFileNumber + "' "
          + "AND ST02TXN_CTL_NBR = '" + vSt02txnCtlNbr + "' "
          + "AND LOOP_ID = '2010AA' ";
      rowEntityName = getEntityName(vEntWhereClause, dbEM);
      
      rowEntityAddress = null;
      if(rowEntityName != null)
  	  {
    	  vEntWhereClause = "FILE_NUMBER = '" + vFileNumber + "' "
    	  + "AND ST02TXN_CTL_NBR = '" + vSt02txnCtlNbr + "' "
    	  + "AND LOOP_ID = '2010AA' " + " AND HRCHL_LEVEL_KEY='"+rowEntityName.getHrchlLevel().getHrchlLevelKey()+"'"; //5010: added HRCHL_LEVEL_KEY
    	  rowEntityAddress = getEntityAddress(vEntWhereClause, dbEM);
  	  }
    }
    writeSvcLocRec(rowEntityName, rowEntityAddress, "CLAIM", vOutputFile);

    // The following is the code to create E208 Record - Referring Physician
    vEntWhereClause = "CLAIM_KEY = '" + vClaimKey + "' "
        + "AND NM101ENTY_ID_CD IN ('DN', 'P3') ";
    rowEntityName = getEntityName(vEntWhereClause
        + "AND LOOP_ID = '2310A' ", dbEM);

    String vTxwhereCls = "CLAIM_KEY = '" + vClaimKey + "' "
        + "AND PRV01PRVD_TYPE_CD = 'RF' " + "AND LOOP_ID = '2310A' ";
    String vAttnTaxonamyCd = getTaxonomyCd(vTxwhereCls, dbEM);
    if (rowEntityName != null) {
      writePhyRec(rowEntityName, vAttnTaxonamyCd, "E208", vOutputFile);
    }

    // The following is the code to create E209 Record - Rendering Physician
    vEntWhereClause = "CLAIM_KEY = '" + vClaimKey + "' "
        + "AND NM101ENTY_ID_CD = '82' ";
    rowEntityName = getEntityName(vEntWhereClause
        + "AND LOOP_ID = '2310B' ", dbEM);

    vTxwhereCls = "CLAIM_KEY = '" + vClaimKey + "' "
        + "AND PRV01PRVD_TYPE_CD = 'PE' " + "AND LOOP_ID = '2310B' ";
    vAttnTaxonamyCd = getTaxonomyCd(vTxwhereCls, dbEM);
    if (rowEntityName != null) {
      writePhyRec(rowEntityName, vAttnTaxonamyCd, "E209", vOutputFile);
    }

    // The following is the code to create E210 Record - Other Insurance
    // Info.
    processE210Rec(vClaimKey, vPatAmtPaid, vPatBalDue, vOutputFile, dbEM);

    
    //TWP added for 5010:
    vEntWhereClause = "CLAIM_KEY = '" + vClaimKey + "' "
    + "AND NM101ENTY_ID_CD = 'DD' ";
    rowEntityName = getEntityName(vEntWhereClause + "AND LOOP_ID = '2310D' ", dbEM);
    if (rowEntityName != null) {
        writePhyRec(rowEntityName, vAttnTaxonamyCd, "E211", vOutputFile);
    }
    
    // The following is the code to create E220 Record - OHI Claim
    // Adjustment
    processE220Rec(vClaimKey, vOutputFile, dbEM);

    // The following is the code to create E225 Record - Other Insurance
    // General Info.
    processE225Rec(vClaimKey, vOutputFile, dbEM);

    vOutputFile.flush();
  } // end writeERec

  /**
   * Gets claim dates and writes the dates to the extract file.
   * 
   * @param rowClaim
   *          - the claim used to get the dates from
   * @param vDntlFile
   *          - the extract file the dates are written to
   * @param dbEM
   */
  private void processE005Rec(Claim rowClaim, PrintWriter vDntlFile,
      EntityManager dbEM) {
    final String vClaimKey = rowClaim.getClaimKey();

    final String vStmtDt = getDentalStmtDate(vClaimKey, dbEM);
    final String vAdmissionDt = getDate(vClaimKey, "435", dbEM);
    final String vDischargeDt = getDate(vClaimKey, "096", dbEM);
    final String vInjuryDt = getDate(vClaimKey, "431", dbEM);
    final String vAccidentDt = getDate(vClaimKey, "439", dbEM);

    writeE005Rec(vStmtDt, vAdmissionDt, vDischargeDt, vInjuryDt, vAccidentDt, "","", vDntlFile);
  } // end processE005Rec

  /**
   * Gets orthodontic information. Added for 5010.
   * 
   * @param rowClaim
   *          - the claim used to get the dates from
   * @param vDntlFile
   *          - the extract file the dates are written to
   * @param dbEM
   */
  private void processE006Rec(Claim rowClaim, PrintWriter vDntlFile,
      EntityManager dbEM) {
	  
    // Richard 9/29
	  Query getOrthodonticInfoQuery = dbEM
      .createQuery("SELECT oi FROM gov.va.med.domain.erepos.OrthodonticInfo oi WHERE oi.claim.claimKey=:claimKey");
    getOrthodonticInfoQuery.setParameter("claimKey", rowClaim.getClaimKey());
	  // End Richard 9/29
	  List orthodonticInfos =  getOrthodonticInfoQuery.getResultList();
	  
	  OrthodonticInfo orthodonticInfo = null;

	  if(orthodonticInfos != null && !orthodonticInfos.isEmpty())
	  {
		  orthodonticInfo = (OrthodonticInfo) orthodonticInfos.get(0);
		  writeE006Rec(orthodonticInfo, vDntlFile);		  
	  }

	 

  } // end processE006Rec
  

  private void processE007Rec(Claim rowClaim, PrintWriter vDntlFile,
      EntityManager dbEM) {
	  
    // Richard 9/29
	  Query getToothStatusQuery = dbEM
      .createQuery("SELECT ts FROM gov.va.med.domain.erepos.ToothStatus ts WHERE ts.claim.claimKey =:claimKey");
    getToothStatusQuery.setParameter("claimKey", rowClaim.getClaimKey());
	  // End Richard 9/29
	  List toothStatuses =  getToothStatusQuery.getResultList();
	  
	  if(toothStatuses != null)
	  {
		Iterator toothStatusIterator =  toothStatuses.iterator();
		while(toothStatusIterator.hasNext())
		{			
			ToothStatus toothStatus = (ToothStatus) toothStatusIterator.next();
			writeE007Rec(toothStatus, vDntlFile);
		}
	  }

  } // end processE006Rec

  
  
  /**
   * writes line level information to the extract file
   * 
   * @param rowClaim
   *          - the claim
   * @param vOutputFile
   * @param dbEM
   */
  private void writeFRec(Claim rowClaim, PrintWriter vOutputFile,
      EntityManager dbEM) {
    final String vClaimKey = rowClaim.getClaimKey();
    processF006Rec(vClaimKey, vOutputFile, dbEM);
        
    vOutputFile.flush();
  } // end writeFRec

  /**
   * Gets dental service line information and writes it to the extract file.
   * 
   * @param vClaimKey
   *          - the claim key
   * @param vDntlFile
   *          - the dental extract file to write the data to
   * @param dbEM
   */
  private void processF006Rec(String vClaimKey, PrintWriter vDntlFile,
      EntityManager dbEM) {
    // The following is the code to create F006 Record - Dental Service Line
    DentalSvcLine rowDentalSvcLine = null;

    Query getClaimQuery = dbEM
        .createQuery("SELECT dsl FROM gov.va.med.domain.erepos.DentalSvcLine dsl WHERE dsl.claim.claimKey=:claimKey ");
    getClaimQuery.setParameter("claimKey", vClaimKey);
    List claims = getClaimQuery.getResultList();
    Iterator claimIterator = claims.iterator();

    final ArrayList elementsF006 = new ArrayList(6);
    while (claimIterator.hasNext()) {
      // Start the formatted line over from scratch each iteration
      elementsF006.clear();

      rowDentalSvcLine = (DentalSvcLine) claimIterator.next();
      final String vDenKey = rowDentalSvcLine.getServiceKey();
      final String vSvcKey = "AND DENTAL_SERVICE_KEY = '" + vDenKey
          + "' ";
      final String vServiceDt = getDate(vClaimKey, vSvcKey, "472", dbEM);
      final String vLineCtrlNo = getRefValue(vClaimKey, "2400", vSvcKey,
          "6R", dbEM);
      final BigDecimal vOhiAprovAmt = getClaimAmt(vClaimKey, vSvcKey,
          "AAE", dbEM);
      final BigDecimal vSalesTaxAmt = getClaimAmt(vClaimKey, vSvcKey,
          "T", dbEM);
      final String vRefWhereCls = "CLAIM_KEY = '" + vClaimKey + "' "
          + "AND LOOP_ID = '2400' " + vSvcKey
          + "AND REF01ID_QLFR = 'G1' ";  //5010: Changed IN ('G1', '9F') to ='G1' 
      //the results will be ordered by seq and the first one will be used
      final String vPreAuthNo = getRefValue(vRefWhereCls, dbEM);

      if (vServiceDt == null) {
        elementsF006.add(" ");
        elementsF006.add(" ");
      } else {
        elementsF006.add(vServiceDt);
        elementsF006.add(" ");
      }

      if (vLineCtrlNo == null) {
        elementsF006.add(" ");
      } else {
        elementsF006.add(vLineCtrlNo);
      }

      elementsF006.add(vOhiAprovAmt);
      elementsF006.add(vSalesTaxAmt);

      if (vPreAuthNo == null) {
        elementsF006.add(" ");
      } else {
        elementsF006.add(vPreAuthNo);
      }

      writeF006Rec(rowDentalSvcLine, elementsF006, vDntlFile);

      // The following is the code to create F007 Record - Tooth
      // Information
      ToothInformation rowToothInfo = null;

      Query getToothInfosQuery = dbEM
          .createNativeQuery(
              "SELECT * FROM TOOTH_INFORMATION WHERE CLAIM_KEY=:claimKey and DENTAL_SERVICE_KEY=:dentalServiceKey",
              ToothInformation.class);
      getToothInfosQuery.setParameter("claimKey", vClaimKey);
      getToothInfosQuery.setParameter("dentalServiceKey", vDenKey);
      List toothInfos = getToothInfosQuery.getResultList();
      Iterator toothInfoIterator = toothInfos.iterator();

      while (toothInfoIterator.hasNext()) {
        rowToothInfo = (ToothInformation) toothInfoIterator.next();
        writeF007Rec(rowToothInfo, vDntlFile);
      }

      processF020Rec(vDenKey, vDntlFile, dbEM);
      
      // The following is the code to create F034 Record - Rendering
      // Physician Line level
      EntityName rowEntityName = null;
      final String vEntWhereClause = "DENTAL_SERVICE_KEY = '" + vDenKey
          + "' " + "AND NM101ENTY_ID_CD = '82' ";
      rowEntityName = getEntityName(vEntWhereClause
          + "AND LOOP_ID = '2420A' ", dbEM);

      final String vTxwhereCls = "DENTAL_SERVICE_KEY = '" + vDenKey
          + "' " + "AND PRV01PRVD_TYPE_CD = 'PE' "
          + "AND LOOP_ID = '2420A' ";
      final String vAttnTaxonamyCd = getTaxonomyCd(vTxwhereCls, dbEM);
      if (rowEntityName != null) {
        writePhyRec(rowEntityName, vAttnTaxonamyCd, "F034", vDntlFile);
      }
      
      String vRenderingEntWhereClause = "DENTAL_SERVICE_KEY = '" + vDenKey
      + "' " + "AND NM101ENTY_ID_CD = 'DD' ";
      EntityName rowRenderingEntityName = getEntityName(vRenderingEntWhereClause
              + "AND LOOP_ID = '2420B' ", dbEM);        
      if (rowRenderingEntityName != null) {
            writePhyRec(rowRenderingEntityName, "", "F035", vDntlFile);
      }
     

      // The following is the code to create F040 Record - OHI Service Line level Info
      //final String vDenDtKey = "DENTAL_SERVICE_KEY = '" + vDenKey + "' ";
      processF040Rec(vClaimKey, vDenKey, vDntlFile, dbEM);
    }
  } // end processF006Rec

  
  /**
	 * gets and writes Health Care Pricing information to the extract file
	 * 
	 * @param vSvcKey
	 *            - the service line key
	 * @param vPrfsnFile
	 *            - the extract file to write the data to
	 * @param dbEM
	 */
	private void processF020Rec(String vSvcKey, PrintWriter vDntlFile, EntityManager dbEM) {
		// The following is the code to create F020 Record - Health Care Pricing/Repricing
		ClaimPrcReprc rowClaimPrcReprc = null;
		Query getClaimPrcReprcQuery = dbEM.createNativeQuery(
			"SELECT * FROM CLAIM_PRC_REPRC WHERE DENTAL_SERVICE_KEY=:dentalServiceKey",
			ClaimPrcReprc.class);
		getClaimPrcReprcQuery.setParameter("dentalServiceKey", vSvcKey);
		List claimPrcReprcs = getClaimPrcReprcQuery.getResultList();
		Iterator claimPrcReprcIterator = claimPrcReprcs.iterator();
		if (claimPrcReprcIterator.hasNext()) {
			rowClaimPrcReprc = (ClaimPrcReprc) claimPrcReprcIterator.next();
			writeF020Rec(rowClaimPrcReprc, vDntlFile);
			detach(rowClaimPrcReprc, dbEM);
		}
	}
		
		/**
		 * Drives the processing of F040(OHI Svc Line Info) records both the data
		 * Acquisition and writing the data to the extract file.
		 * 
		 * @param vClaimKey
		 * @param vSvcKey
		 * @param vOutputFile
		 * @param dbEM
		 */
		protected void processF040Rec(String vClaimKey, String vDentalKey,
				PrintWriter vOutputFile, EntityManager dbEM) {
			// The following is the code to create F040 Record - OHI Service Line level Info
			final String vSvcKeyClause = "DENTAL_SERVICE_KEY = '" + vDentalKey	+ "' ";
			SvcLineAdjdct rowSvcLineAdjdct = null;

			String serviceLineAdjudicationsQueryString = 
					"SELECT * FROM SVC_LINE_ADJDCT WHERE " + vSvcKeyClause + " order by SEQ_NBR";
			//summaryLOGGER.info("Service Line Adjudications Query: " + serviceLineAdjudicationsQueryString);
      Query getSvcLineAdjdctQuery = dbEM.createNativeQuery(serviceLineAdjudicationsQueryString, SvcLineAdjdct.class);
			List svcLineAdjdcts = getSvcLineAdjdctQuery.getResultList();
			Iterator svcLineAdjdctIterator = svcLineAdjdcts.iterator();
			String vDate = null;
	//TWP: TODO: Richard: Please take a look at this should the service dates/pat remaining amount be specific to an OHI?
	//They are not in this code		
			final String vSvcAdjDt = getDate(vClaimKey, "AND " + vSvcKeyClause, "573",dbEM);			
			final BigDecimal vRemainingPatientAmount = getClaimAmt( vClaimKey, "AND " + vSvcKeyClause,"EAF", dbEM);
		
			while (svcLineAdjdctIterator.hasNext()) {
				rowSvcLineAdjdct = (SvcLineAdjdct) svcLineAdjdctIterator.next();
				writeF040Rec(rowSvcLineAdjdct, vSvcAdjDt, vRemainingPatientAmount, vOutputFile);
				Cases rowCases = null;
				Query getCasesQuery = dbEM
						.createNativeQuery("SELECT * FROM CASES WHERE DENTAL_SERVICE_KEY=:prfnServiceKey and LOOP_SEQ_NBR=:loopSeqNbr order by SEQ_NBR",Cases.class);
				getCasesQuery.setParameter("prfnServiceKey", vDentalKey);
				getCasesQuery.setParameter("loopSeqNbr", rowSvcLineAdjdct.getSeqNbr()+"");

				List cases = getCasesQuery.getResultList();
				Iterator caseIterator = cases.iterator();
				while (caseIterator.hasNext()) {
					rowCases = (Cases) caseIterator.next();
					writeCasesRec(rowCases, "F050", vOutputFile);				
				}
				detach(cases, dbEM);
			}
			detach(svcLineAdjdcts, dbEM);
		} // end
				
  
  /**
   * Returns the dental statement date associated with the claim.
   * 
   * @param vClaimKey
   * @param dbEM
   * @return
   */
  private String getDentalStmtDate(String vClaimKey, EntityManager dbEM) {
    String vDentalStmtDate = null;

    Query getDentalStmtDateQuery = dbEM
        .createNativeQuery(
            "SELECT * FROM CLAIM_DT WHERE CLAIM_KEY=:claimKey AND DTP01DTTM_QUAL=:dtm01dtmQual AND DENTAL_SERVICE_KEY IS NULL",
            ClaimDt.class);
    getDentalStmtDateQuery.setParameter("claimKey", vClaimKey);
    getDentalStmtDateQuery.setParameter("dtm01dtmQual", "472");
    List dentalStmtDates = getDentalStmtDateQuery.getResultList();
    Iterator dentalStmtDateIterator = dentalStmtDates.iterator();

    if (dentalStmtDates.size() > 0) {
      final ClaimDt rowClaimDtB = (ClaimDt) dentalStmtDateIterator.next();
      vDentalStmtDate = rowClaimDtB.getDtp03dtFld();
    } else {
      vDentalStmtDate = getStmtDate(vClaimKey,
          "AND DENTAL_SERVICE_KEY IS NOT NULL ", dbEM);
    }
    return vDentalStmtDate;
  }

  /**
   * Gets Reference Identification Qualifiers and writes them to the extract
   * file. writes the code, date, amount etc to the file
   * 
   * @param rowClaim
   *            - the claim that has reference identifiers associated with it
   * @param vInstFile
   *            - the file to write the codes to
   * @param dbEM
   */
  private void processHiCodesRec(Claim rowClaim, PrintWriter vPrfsnFile,
		  EntityManager dbEM) {
	  final String vClaimKey = rowClaim.getClaimKey();
	  String vHiRec = "";
	  String attrName = "";
	  String ic = "";
	  String vHiCdListQual = "";
	  String vHiCd = "";
	  String vHiDt = "";
	  String strTemp = "";
	  BigDecimal vHiAmt = new BigDecimal(0);
	  String vHiCndtnResponseCd = "";		
	  BigDecimal vHiQnty = new BigDecimal(0);

	  HiCodes rowHiCodes = null;
	  Query getHiCodesQuery = dbEM.createNativeQuery(
			  "SELECT * FROM HI_CODES WHERE CLAIM_KEY=:claimKey",
			  HiCodes.class);
	  getHiCodesQuery.setParameter("claimKey", vClaimKey);
	  List hiCodes = getHiCodesQuery.getResultList();
	  Iterator hiCodeIterator = hiCodes.iterator();

	  StringBuffer e100Rec = new StringBuffer("E100");
	  StringBuffer e105Rec = new StringBuffer("E105");
	  StringBuffer e110Rec = new StringBuffer("E110");
	  // HRL - MSB Added new if condition to check for existence of ICD Codes.
	  // If no codes exist write blank record with record type for E100 record
	  // (like "E100") Cache load process expects E100 record to be mandatory and as per EDI
	  // guide it situational, so writing this record with type avoids crashing the
	  // process on Cache side.
	  if (hiCodeIterator.hasNext()) {
		  while (hiCodeIterator.hasNext()) {
			  e100Rec = new StringBuffer("E100");
			  e105Rec = new StringBuffer("E105");
			  e110Rec = new StringBuffer("E110");

			  rowHiCodes = (HiCodes) hiCodeIterator.next();			 
			  for (int i = 1; i <= 4; i++) {

			    // Empty out any fields obtained from the previous record.
          // This fixes a series of bugs turned up in 5010 testing.
          vHiCndtnResponseCd = "";
          vHiAmt = new BigDecimal(0);
          vHiQnty = new BigDecimal(0);
          vHiDt = "";
          
				  // These two were already being emptied out
          vHiCdListQual = "";
				  vHiCd = "";

				  ic = "0" + i;
				  ic = ic.substring(ic.length() - 2);

				  attrName = "Hi" + ic + "cdListQual";
				  if (getAttribute(attrName, rowHiCodes) != null) {
					  vHiCdListQual = (String) getAttribute(attrName, rowHiCodes);
				  }

				  attrName = "Hi" + ic + "procCd";
				  if (getAttribute(attrName, rowHiCodes) != null) {
					  vHiCd = (String) getAttribute(attrName, rowHiCodes);
				  }

				  // Get the poa indicator
				  attrName = "Hi" + ic + "cndtnResponseCd";
				  if (getAttribute(attrName, rowHiCodes) != null) {
					  vHiCndtnResponseCd = (String) ""+getAttribute(attrName, rowHiCodes);
				  }

				  attrName = "Hi" + ic + "procDt";
				  if (getAttribute(attrName, rowHiCodes) != null) {
					  vHiDt = (String) getAttribute(attrName, rowHiCodes);
				  }

				  attrName = "Hi" + ic + "amount";
				  if (getAttribute(attrName, rowHiCodes) != null) {
					  vHiAmt = (BigDecimal) getAttribute(attrName, rowHiCodes);
				  }

				  // Get the quantity field for this segment
				  attrName = "Hi" + ic + "qnty";
				  if (getAttribute(attrName, rowHiCodes) != null) {
					  vHiQnty = (BigDecimal) getAttribute(attrName, rowHiCodes);
				  }					

				  strTemp = vHiCdListQual + vHiCd;
				  if (!("".equals(strTemp))) {
					  vHiRec =  writeHiRec(vHiCdListQual, vHiCd, vHiCndtnResponseCd, vHiDt, vHiAmt,vHiQnty);
					  if (i <= 4) {
						  e100Rec.append(vHiRec);
					  } else if ((i > 4) && (i <= 8)) {
						  e105Rec.append(vHiRec);
					  } else if (i > 8) {
						  e110Rec.append(vHiRec);
					  }
				  }
			  }
			  writeHiRecs(e100Rec.toString(), e105Rec.toString(), e110Rec
					  .toString(), vPrfsnFile);
		  }
		  detach(hiCodes, dbEM);
		  //HRL - MSB Writing the E100 record with just record type value   
	  } else
	  {
		  writeHiRecs("E100", "", "", vPrfsnFile);	      	
	  }		
  }

  /**
   * Utility method that allows a caller to get an attribute on an object via java reflection.
   * 
   * @param attrName
   * @param target
   * @return
   */
  private Object getAttribute(String attrName, Object target) {
	  Object returnObject = null;
	  try {
		  Class targetClass = target.getClass();
		  String methodName = "get" + attrName;
		  Method targetGetMethod = targetClass.getMethod(methodName);
		  returnObject = targetGetMethod.invoke(target, (Object[]) null);
	  } catch(NoSuchMethodException nsme) {
		  //do nothing this is expected.
	  } catch (Exception e) {
		  throw new RuntimeException(
				  "Could not access necessary data via reflection in ExtractInstClaims", e);
	  }
	  return returnObject;
  }
}
