package gov.va.med.hac.edi.extract.x12.jpa;
 
import gov.va.med.hac.edi.utils.jpa.DateUtil;
import gov.va.med.hac.edi.utils.jpa.DbTransactionEnderException;
import java.io.*;
import java.util.*;

import javax.persistence.EntityManager;
import javax.persistence.Query;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import gov.va.med.domain.erepos.*;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This class drives the creation of the institutional claim extract file for a clearing house. It
 * produces a file named <clearinghouseid>_ipt
 * 
 * The class extends process cacheRecords.
 * 
 * @author vhahacpennet
 * 
 */
public class ExtractInstClaims extends ProcessCacheRecords {

  static final Logger summaryLOGGER = LoggerFactory.getLogger(ExtractInstClaims.class);
  static final Logger traceLOGGER = LoggerFactory.getLogger("trace." + ExtractInstClaims.class);
  final private static BigDecimal zeroBD = new BigDecimal("0");

  boolean logIt = false;

  /**
   * This method drives the creation of the institutional claim extract file.
   * 
   * It creates an extract file named <clearing house id> _ipt_
   * 
   * @param vChId
   *          - The id of the clearing house
   * @param vFilePath
   *          - the path where the extract file will be created
   * @param receiverIdStart
   *          -
   * @param vLogFile
   * @throws DbTransactionEnderException
   */
  protected void processInstClaims(String vChId, String vFilePath,
      String receiverIdStart, boolean useClones, EntityManager dbEM)
      throws DbTransactionEnderException {
    try {

	  
      final String fileName = createFileName(vChId, "_ipt_");
   /** Start 1 TWP Inserted code to create the file in a temporary location **/   
      File InstFile = createTemporaryFileAndDir(vFilePath, fileName);
      if(InstFile == null)
      {
    	  throw new IOException("Could not create a temporary file or directory to write the extract file to. " + vFilePath + "tmp : "+ fileName);
      }
   /** End 1 TWP Inserted code to create the file in a temporary location **/
      
      List claimKeysToUpdate = new ArrayList();
      int claimCount = 0;
      PrintWriter vInstFile = null;

      // -- Process Inst claims
      Query getClaimQuery = dbEM
          .createQuery("SELECT c FROM gov.va.med.domain.erepos.Claim c WHERE (c.claimType=:claimType) AND c.statusCode=:statusCode AND c.senderId=:senderId AND (c.receiverId LIKE :receiverId) ORDER BY c.dateCreated, c.fileNumber, c.st02txnCtlNbr");
      getClaimQuery.setParameter("claimType", "005010X223");
      getClaimQuery.setParameter("statusCode", "NEW");
      getClaimQuery.setParameter("senderId", vChId);
      getClaimQuery.setParameter("receiverId", receiverIdStart + "%");

      List<Claim> claims = getClaimQuery.getResultList();
      Iterator<Claim> claimIterator;
      int numClaims = claims.size();
      summaryLOGGER.info("Number of Institutional claims: " + claims.size());

      if (useClones) {
        List<Claim> clonedClaims = cloneAndClearClaims(claims);
        claims = null;
        dbEM.clear();
        System.gc();
        claimIterator = clonedClaims.iterator();
      } else {
        claimIterator = claims.iterator();
      }

      if (numClaims > 0) {
        vInstFile = new PrintWriter(new FileOutputStream(InstFile));

        while (claimIterator.hasNext()) // -- loop for each claim
        {
          final Claim rowClaim = (Claim) claimIterator.next();
          final String vFileNumber = rowClaim.getFileNumber();
          final String vSt02txnCtlNbr = rowClaim.getSt02txnCtlNbr();
          writeARec(vFileNumber, vSt02txnCtlNbr, vInstFile, dbEM);
          writeBRec(rowClaim, vInstFile, dbEM);
          writeCRec(rowClaim, vInstFile, dbEM);
          writeERec(rowClaim, vInstFile, dbEM);
          writeFRec(rowClaim, vInstFile, dbEM);
          writeA999Rec(vInstFile);
          claimKeysToUpdate.add(rowClaim.getClaimKey());
          if (((50 + claimCount) % 50) == 0) {
            if (useClones)
              dbEM.clear();
            vInstFile.flush();
            traceLOGGER.debug("cc= " + claimCount + " "
                + DateUtil.getCurrentTimeFormatted("hh:mm:ss"));
          }
          claimCount++;
        }

        vInstFile.flush();
        vInstFile.close();
      }

      dbEM.clear();
      traceLOGGER.debug("Committing the database transaction");
      dbEM.getTransaction().begin();
      Iterator claimToPersistIterator = claimKeysToUpdate.iterator();
      Query deleteStatement;
      while (claimToPersistIterator.hasNext()) {
        deleteStatement = dbEM
            .createNativeQuery("update claim set status_code=:statusCode where claim_key=:claimKey");
        deleteStatement.setParameter("statusCode", "BUILT");
        deleteStatement.setParameter("claimKey", (String) claimToPersistIterator.next());
        deleteStatement.executeUpdate();
      }
      dbEM.getTransaction().commit();
      traceLOGGER.debug("Commit successful");
      
      /** Start 2 TWP Inserted code to create the file in a temporary location & move it to final location**/
      File renamedFile = new File(vFilePath + fileName);
      InstFile.renameTo(renamedFile);
      traceLOGGER.debug("Was able to rename the file to " + renamedFile.getAbsolutePath());
      /** Start 2 TWP Inserted code to create the file in a temporary location & move it to final location**/
      traceLOGGER.debug("Inst   Claims Processed : " + claimCount);
    } // -- end write Inst clmains
    catch (Exception ex) {
      traceLOGGER
          .debug("There was an Exception in writing Institutinal Claims ", ex);
      traceLOGGER.debug("Rolling back the database transaction");
      if (dbEM != null) {
        if (dbEM.getTransaction().isActive()) {
          dbEM.getTransaction().rollback();
        }
      }
      traceLOGGER.debug("Rollback successful");
      throw new DbTransactionEnderException(
          "Rolled Back the Transaction", ex);
    }
  }

  /**
   * Gets subscriber (insured) information and the destination payer data and writes it to the
   * extract file.
   * 
   * @param rowClaim
   *          - the claim related to the subscriber
   * @param vInstFile
   *          - the extract file to write the data to
   * @param dbEM
   *          - the entity manager used to query the database
   */
  private void writeCRec(Claim rowClaim, PrintWriter vInstFile,
      EntityManager dbEM) {
    String vSbrSecId = "";
    final String vDtOfDeath = "";
    final String vFileNumber = rowClaim.getFileNumber();
    final String vSt02txnCtlNbr = rowClaim.getSt02txnCtlNbr();
    final String vSbsrId = rowClaim.getSubscriberId();
    final String vSubscriberHrchlLevelKey = rowClaim.getSubscriberHrchlLevelKey();
    
    Query getSubscriberInfoQuery = dbEM
        .createQuery("SELECT si FROM gov.va.med.domain.erepos.SubscriberInfo si WHERE si.txnSet.id.fileNumber=:fileNumber AND si.txnSet.id.st02txnCtlNbr=:st02txnCtlNbr AND si.nm109sbsrId=:nm109sbsrId AND si.loopId=:loopId AND si.hrchlLevelKey=:hrchlLevelKey");
    getSubscriberInfoQuery.setParameter("fileNumber", vFileNumber);
    getSubscriberInfoQuery.setParameter("st02txnCtlNbr", vSt02txnCtlNbr);
    getSubscriberInfoQuery.setParameter("nm109sbsrId", vSbsrId);
    getSubscriberInfoQuery.setParameter("loopId", "2000B");
    getSubscriberInfoQuery.setParameter("hrchlLevelKey", vSubscriberHrchlLevelKey);
    final SubscriberInfo rowSubscriberInfo = (SubscriberInfo) getSubscriberInfoQuery
        .getResultList().get(0);
    
    final String vWhereCls = "FILE_NUMBER = '" + vFileNumber + "' "
    + "AND ST02TXN_CTL_NBR = '" + vSt02txnCtlNbr + "' "
    + "AND NM109SBSR_ID = '" + vSbsrId + "' "
    + "AND HRCHL_LEVEL_KEY ='" + vSubscriberHrchlLevelKey +"'";
    
    //5010: ADDED NM101ENTY_ID_CD
    EntityName rowEntityName = getEntityName(vWhereCls + "AND LOOP_ID = '2010BA' AND NM101ENTY_ID_CD='IL'", dbEM);


	
    EntityAddress rowEntityAddress = null;
	if(rowEntityName != null)
	{
		//5010: added HRCHL_LEVEL_KEY
		rowEntityAddress = getEntityAddress(vWhereCls + "AND LOOP_ID = '2010BA' AND HRCHL_LEVEL_KEY='"+rowEntityName.getHrchlLevel().getHrchlLevelKey()+"'", dbEM);
	}

    //5010: Changed REF01ID_QLFR='IL' to REF01ID_QLFR='SY'
    vSbrSecId = getRefValue(vWhereCls + "AND LOOP_ID = '2010BA' AND REF01ID_QLFR='SY' ", dbEM);

    Query getDemographicInfoQuery = dbEM
        .createQuery("SELECT di FROM gov.va.med.domain.erepos.DemographicInfo di WHERE di.txnSet.id.fileNumber=:fileNumber AND di.txnSet.id.st02txnCtlNbr=:st02txnCtlNbr AND di.nm109sbsrId=:nm109sbsrId AND di.loopId=:loopId AND di.hrchlLevelKey=:hrchlLevelKey");
    getDemographicInfoQuery.setParameter("fileNumber", vFileNumber);
    getDemographicInfoQuery.setParameter("st02txnCtlNbr", vSt02txnCtlNbr);
    getDemographicInfoQuery.setParameter("nm109sbsrId", vSbsrId);
    getDemographicInfoQuery.setParameter("loopId", "2010BA");
    getDemographicInfoQuery.setParameter("hrchlLevelKey", vSubscriberHrchlLevelKey);
	DemographicInfo rowDemographicInfo = null;
	if(getDemographicInfoQuery.getResultList().size() > 0)
	{
		rowDemographicInfo = (DemographicInfo) getDemographicInfoQuery.getResultList().get(0);
	}
	
    writeC000Rec(vSbrSecId, vDtOfDeath, rowSubscriberInfo, rowEntityName,
        rowEntityAddress, rowDemographicInfo, vInstFile);

    rowEntityName = getEntityName(vWhereCls + "AND LOOP_ID = '2010BB' ", dbEM);

    writeC005Rec(rowEntityName, vInstFile);
    vInstFile.flush();
    detach(rowSubscriberInfo, dbEM);
    detach(rowDemographicInfo, dbEM);
    detach(rowEntityName, dbEM);
    detach(rowEntityAddress, dbEM);
  }

  /**
   * Gets basic claim information and writes it to the extract file.
   * 
   * @param rowClaim
   *          - the claim
   * @param vOutputFile
   *          - the extract file
   * @param dbEM
   *          - entity manager attached to eRepository
   */
  private void writeERec(Claim rowClaim, PrintWriter vOutputFile, EntityManager dbEM) {
    final String vClaimKey = rowClaim.getClaimKey();
    final String vFileNumber = rowClaim.getFileNumber();
    final String vSt02txnCtlNbr = rowClaim.getSt02txnCtlNbr();
    String vTPClaimRecId = getRefValue(vClaimKey, "2300", "D9", dbEM);
    // E000 Record can already be written
    writeE000Rec(rowClaim, "A", vTPClaimRecId, vOutputFile);

    // E005 Record has a little logic
    processE005Rec(rowClaim, vOutputFile, dbEM);

    // E008 Record
    InstClaimCode rowInstClaimCode = null;
    Query getClaimCodeQuery = dbEM
        .createQuery("SELECT icc FROM gov.va.med.domain.erepos.InstClaimCode icc WHERE icc.claimKey=:claimKey ");
    getClaimCodeQuery.setParameter("claimKey", vClaimKey);
    List claimCodes = getClaimCodeQuery.getResultList();
    Iterator claimCodeIterator = claimCodes.iterator();

    if (claimCodeIterator.hasNext()) {
      rowInstClaimCode = (InstClaimCode) claimCodeIterator.next();
        writeE008Rec(rowInstClaimCode, vOutputFile);
    }
    
    // E009 Record
    processE009Rec(vClaimKey, vOutputFile, dbEM);

    // Write the E010 record if the total billed is positive
    final BigDecimal vTotBilled = rowClaim.getClm02clmChrgAmt();    
    final BigDecimal vPatAmtPaid = getPatAmtPaid(vClaimKey, dbEM);
    final BigDecimal vPatBalDue = getClaimAmt(vClaimKey, "F3", dbEM);
       
    /**
     * DEF009627 - EAT;20100507; Prev: if (vTotBilled.signum() == 1) This was allowing only
     * non-zero, positive billed amounts. Zero-dollar claim amounts are valid
     */
    if (vTotBilled.signum() >= 0) {
      writeE010Rec(vTotBilled, vPatAmtPaid, vPatBalDue, vOutputFile);
    }

    // E015 Record is complicated
    processE015Rec(vClaimKey, vOutputFile, dbEM);

    // E100, E105 and E110 records

    processHiCodesRec(rowClaim, vOutputFile, dbEM);

    ClaimPrcReprc rowClaimPrcReprc = null;
    Query getClaimPrcReprcQuery = dbEM
        .createNativeQuery(
            "SELECT * FROM CLAIM_PRC_REPRC WHERE CLAIM_KEY=:claimKey AND INST_SERVICE_KEY IS NULL AND PRFSN_SERVICE_KEY IS NULL",
            ClaimPrcReprc.class);
    getClaimPrcReprcQuery.setParameter("claimKey", vClaimKey);
    List claimPrcReprcs = getClaimPrcReprcQuery.getResultList();
    Iterator claimPrcReprcIterator = claimPrcReprcs.iterator();

    if (claimPrcReprcIterator.hasNext()) {
      rowClaimPrcReprc = (ClaimPrcReprc) claimPrcReprcIterator.next();
      writeE125Rec(rowClaimPrcReprc, vOutputFile);
    }
    detach(claimPrcReprcs, dbEM);

    // The following is the code to create E205 Record - Service Location
    EntityName rowEntityName = null;
    EntityAddress rowEntityAddress = null;
    
    String vEntWhereClause = "CLAIM_KEY = '" + vClaimKey + "' "
        + "AND NM101ENTY_ID_CD = '77' "; //Changed to 77 from FA
    rowEntityName = getEntityName(vEntWhereClause
        + "AND LOOP_ID = '2310E' ", dbEM);

    rowEntityAddress = null;
    if(rowEntityName != null)
	{
    	final String vAdrWhereClause = "CLAIM_KEY = '" + vClaimKey + "' AND HRCHL_LEVEL_KEY='"+rowEntityName.getHrchlLevel().getHrchlLevelKey()+"'"; //5010: added HRCHL_LEVEL_KEY
    	rowEntityAddress = getEntityAddress(vAdrWhereClause
        + "AND LOOP_ID = '2310E' ", dbEM);
	}

    if ((rowEntityName == null) || (rowEntityAddress == null)) {
      vEntWhereClause = "FILE_NUMBER = '" + vFileNumber + "' "
          + "AND ST02TXN_CTL_NBR = '" + vSt02txnCtlNbr + "' "
          + "AND LOOP_ID = '2010AA' ";
      detach(rowEntityName, dbEM);
      detach(rowEntityAddress, dbEM);
      rowEntityName = getEntityName(vEntWhereClause, dbEM);

      rowEntityAddress = null;
      if(rowEntityName != null)
  	  {
    	  vEntWhereClause = "FILE_NUMBER = '" + vFileNumber + "' "
    	  + "AND ST02TXN_CTL_NBR = '" + vSt02txnCtlNbr + "' "
    	  + "AND LOOP_ID = '2010AA' " + " AND HRCHL_LEVEL_KEY='"+rowEntityName.getHrchlLevel().getHrchlLevelKey()+"'"; //5010: added HRCHL_LEVEL_KEY
    	  rowEntityAddress = getEntityAddress(vEntWhereClause, dbEM);
  	  }
    }
    writeSvcLocRec(rowEntityName, rowEntityAddress, "CLAIM", vOutputFile);

    // The following is the code to create E206 Record - Attending Physician
    vEntWhereClause = "CLAIM_KEY = '" + vClaimKey + "' "
        + "AND NM101ENTY_ID_CD = '71' ";
    detach(rowEntityName, dbEM);
    rowEntityName = getEntityName(vEntWhereClause + "AND LOOP_ID = '2310A'", dbEM);  

    final String vTxwhereCls = "CLAIM_KEY = '" + vClaimKey + "' "
        + "AND PRV01PRVD_TYPE_CD = 'AT' " + "AND LOOP_ID = '2310A' ";
    final String vAttnTaxonamyCd = getTaxonomyCd(vTxwhereCls, dbEM);
    if ((rowEntityName != null)) {
      writePhyRec(rowEntityName, vAttnTaxonamyCd, "E206", vOutputFile);
    }

    String vRenderingEntWhereClause = "CLAIM_KEY = '" + vClaimKey + "' "
    + "AND NM101ENTY_ID_CD = '82' ";
    EntityName rowRenderingEntityName = getEntityName(vRenderingEntWhereClause
            + "AND LOOP_ID = '2310D' ", dbEM);        
    if (rowRenderingEntityName != null) {
          writePhyRec(rowRenderingEntityName, "", "E209", vOutputFile);
    }
    
    // The following is the code to create E210 Record - Other Insurance Info.
    processE210Rec(vClaimKey, vPatAmtPaid, vPatBalDue, vOutputFile, dbEM);

    // The following is the code to create E214 Record - Medicare Inpatient
    // Adjudication Info.

    MediInptAdjdct rowMediInptAdjdct = null;
    Query getMediInptAdjdctQuery = dbEM.createNativeQuery(
        "SELECT * FROM MEDI_INPT_ADJDCT WHERE CLAIM_KEY=:claimKey",
        MediInptAdjdct.class);
    getMediInptAdjdctQuery.setParameter("claimKey", vClaimKey);
    List mediInptAdjdcts = getMediInptAdjdctQuery.getResultList();
    Iterator mediInptAdjdctIterator = mediInptAdjdcts.iterator();

    if (mediInptAdjdctIterator.hasNext()) {
      rowMediInptAdjdct = (MediInptAdjdct) mediInptAdjdctIterator.next();
      writeE214Rec(rowMediInptAdjdct, vOutputFile);
    }

    // The following is the code to create E216 Record - Medicare Outpatient
    // Adjudication Info.
    MediOutptAdjdct rowMediOutptAdjdct = null;
    Query getMediOutptAdjdctQuery = dbEM.createNativeQuery(
        "SELECT * FROM MEDI_OUTPT_ADJDCT WHERE CLAIM_KEY=:claimKey",
        MediOutptAdjdct.class);
    getMediOutptAdjdctQuery.setParameter("claimKey", vClaimKey);
    List mediOutptAdjdcts = getMediOutptAdjdctQuery.getResultList();
    Iterator mediOutptAdjdctIterator = mediOutptAdjdcts.iterator();

    if (mediOutptAdjdctIterator.hasNext()) {
      rowMediOutptAdjdct = (MediOutptAdjdct) mediOutptAdjdctIterator
          .next();
      writeE216Rec(rowMediOutptAdjdct, vOutputFile);
    }

    // The following is the code to create E220 Record - OHI Claim
    // Adjustment
    processE220Rec(vClaimKey, vOutputFile, dbEM);

    // The following is the code to create E225 Record - Other Insurance
    // General Info.
    processE225Rec(vClaimKey, vOutputFile, dbEM);

    detach(claimCodes, dbEM);
    detach(rowEntityName, dbEM);
    detach(rowEntityAddress, dbEM);
    detach(mediInptAdjdcts, dbEM);
    detach(mediOutptAdjdcts, dbEM);
    detach(rowClaimPrcReprc, dbEM);

    vOutputFile.flush();
  } // end writeERec

  /**
   * Gets claim dates and writes the dates to the extract file.
   * 
   * @param rowClaim
   *          - the claim used to get the dates from
   * @param vInstFile
   *          - the extract file the dates are written to
   * @param dbEM
   */
  private void processE005Rec(Claim rowClaim, PrintWriter vInstFile,
      EntityManager dbEM) {
    final String vClaimKey = rowClaim.getClaimKey();

    final String vStmtDt = getDate(vClaimKey, "434", dbEM);
    final String vAdmissionDt = getDate(vClaimKey, "435", dbEM);
    final String vDischargeDt = getDate(vClaimKey, "096", dbEM);
    final String vInjuryDt = getDate(vClaimKey, "431", dbEM);
    final String vAccidentDt = getDate(vClaimKey, "439", dbEM);

    writeE005Rec(vStmtDt, vAdmissionDt, vDischargeDt, vInjuryDt,
        vAccidentDt, "","", vInstFile);
  } // end processE005Rec

  /**
   * Gets Reference Identification Qualifiers and writes them to the extract file. writes the code,
   * date, amount etc to the file
   * 
   * @param rowClaim
   *          - the claim that has reference identifiers associated with it
   * @param vInstFile
   *          - the file to write the codes to
   * @param dbEM
   */
	private void processHiCodesRec(Claim rowClaim, PrintWriter vInstFile,EntityManager dbEM) {
		//summaryLOGGER.info ("In processHiCodesRec !!!!!!!!!!!!!!!!!!");
	  final String vClaimKey = rowClaim.getClaimKey();
		String vHiCodeSegment = "";
		String attrName = "";
		String segmentNumber = "";
		String vHiQual = "";
		
		String qualAndCode = "";
		int numberOfSegments = 0;

    // Fetched with getAttribute(), which uses reflection
		String vHiCdListQual = "";
    String vHiCd = "";
    String vHiCndtnResponseCd = "";
    BigDecimal vHiAmt = new BigDecimal(0);
    BigDecimal vHiQnty = new BigDecimal(0);
    String vHiDt = "";

		// Table HI_CODES has 60 columns, 5 each for segments 1-12, with names
		// like
		// HIxxCD_LIST_QUAL
		// HIxxPROC_CD
		// HIxxDTTM_QUAL - this date/time qualifier seems to be ignored, leaving
		// 48 possible columns
		// HIxxPROC_DT
		// HIxxAMOUNT
		// The code below decides how many of these 12 segments to consider,
		// based on the value
		// found in the CD_LIST_QUAL for segment 
		HiCodes rowHiCodes = null;
		Query getHiCodesQuery = dbEM
				.createNativeQuery(
						"SELECT * FROM HI_CODES WHERE CLAIM_KEY=:claimKey ORDER BY SEQ_NBR",
						HiCodes.class);
		getHiCodesQuery.setParameter("claimKey", vClaimKey);
		List hiCodes = getHiCodesQuery.getResultList();
		Iterator hiCodeIterator = hiCodes.iterator();

		StringBuffer e100Rec;
		StringBuffer e105Rec;
		StringBuffer e110Rec;

		// HRL - MSB Added new if condition to check for existance of ICD Codes.
		// If no codes exist write blank record with record type for E100 record
		// (like "E100")
		// Cache load process expects E100 record to be mandatory and as per EDI
		// guide it
		// situational, so writing this record with type avoids crashing the
		// process on Cache side.
		if (hiCodeIterator.hasNext()) {
			// Loop through the rows from HI_CODES, for which there might be
			// multiple on one claim
			while (hiCodeIterator.hasNext()) {
			  // Debugging
			  logIt = false;
				
			  rowHiCodes = (HiCodes) hiCodeIterator.next();

				e100Rec = new StringBuffer("E100");
				e105Rec = new StringBuffer("E105");
				e110Rec = new StringBuffer("E110");

				// The cd list qualifier determines the number of segments that
				// will be considered
				vHiQual = rowHiCodes.getHi01cdListQual().trim();
				//summaryLOGGER.info ("This next HI Code: " + vHiQual);

				if (vHiQual.equalsIgnoreCase("BK") || vHiQual.equalsIgnoreCase("ABK")) {
					numberOfSegments = 1;
				} else if (vHiQual.equalsIgnoreCase("BJ") || vHiQual.equalsIgnoreCase("ABJ")) {
				  numberOfSegments = 1;
				} else if (vHiQual.equalsIgnoreCase("PR") || vHiQual.equalsIgnoreCase("APR")) {
					numberOfSegments = 3;
				} else if (vHiQual.equalsIgnoreCase("BN") || vHiQual.equalsIgnoreCase("ABN")) {
          //summaryLOGGER.info("Found a BN hicode");
          //logIt = true;
					numberOfSegments = 12;
				} else if (vHiQual.equalsIgnoreCase("DR")) {
					numberOfSegments = 1;
				} else if (vHiQual.equalsIgnoreCase("BF") || vHiQual.equalsIgnoreCase("ABF")) {
					numberOfSegments = 12;
				} else if (((vHiQual.equalsIgnoreCase("BR"))
						|| (vHiQual.equalsIgnoreCase("BBR")))) {
					numberOfSegments = 1;
				} else if ((vHiQual.equalsIgnoreCase("BQ"))
						|| (vHiQual.equalsIgnoreCase("ABQ"))) {
					numberOfSegments = 12;
				} else if (vHiQual.equalsIgnoreCase("BI")) {
					numberOfSegments = 12;
				} else if (vHiQual.equalsIgnoreCase("BH")) {
					numberOfSegments = 12;
				}  else if (vHiQual.equalsIgnoreCase("BE")) {
					numberOfSegments = 12;
				} else if (vHiQual.equalsIgnoreCase("BG")) {
					numberOfSegments = 12;
				} else if (vHiQual.equalsIgnoreCase("TC")) {
					numberOfSegments = 12;
				}

				// Loop through the number of segments that was just defined
				// Really, we are looping through the many columns in the table
				for (int i = 1; i <= numberOfSegments; i++) {
	        // Empty out any fields obtained from the previous record.
	        // This fixes a series of bugs turned up in 5010 testing.
	        vHiCndtnResponseCd = "";
	        vHiAmt = new BigDecimal(0);
	        vHiQnty = new BigDecimal(0);
	        vHiDt = "";
	        
	        // These two were already being emptied out
					vHiCdListQual = "";
					vHiCd = "";

					// If the segment number is single-digit, put a zero in
					// front of it
					segmentNumber = "0" + i;
					segmentNumber = segmentNumber.substring(segmentNumber.length() - 2);

					// Get the code list qualifier field for this segment
					attrName = "Hi" + segmentNumber + "cdListQual";
					if (getAttribute(attrName, rowHiCodes) != null) {
						vHiCdListQual = (String) getAttribute(attrName,rowHiCodes);
					}
					
					// Get the code field for this segment
					attrName = "Hi" + segmentNumber + "procCd";
					if (getAttribute(attrName, rowHiCodes) != null) {
						vHiCd = (String) getAttribute(attrName, rowHiCodes);
					}

					// Get the poa indicator
					attrName = "Hi" + segmentNumber + "cndtnResponseCd";
//					if (logIt) {
//					  summaryLOGGER.info ("Going to getAttribute for vHiCndtnResponseCd, currently: " + vHiCndtnResponseCd);
//					}
					if (getAttribute(attrName, rowHiCodes) != null) {
						vHiCndtnResponseCd = (String) ""+getAttribute(attrName, rowHiCodes);
//					} else {
//					  if (logIt) {
//					    summaryLOGGER.info ("Not setting vHiCndtnResponseCd because it returned null");
//					  }
					}

//          if (logIt) {
//					  summaryLOGGER.info("vHiCndtnResponseCd: " + vHiCndtnResponseCd);
//            String directResult = rowHiCodes.getHi01cndtnResponseCd();
//            summaryLOGGER.info("Direct result: " + directResult);
//          }
          
					// Get the date field for this segment
					attrName = "Hi" + segmentNumber + "procDt";
					if (getAttribute(attrName, rowHiCodes) != null) {
						vHiDt = (String) getAttribute(attrName, rowHiCodes);
					}

					// Get the amount field for this segment
					attrName = "Hi" + segmentNumber + "amount";
					if (getAttribute(attrName, rowHiCodes) != null) {
						vHiAmt = (BigDecimal) getAttribute(attrName, rowHiCodes);
					}
					
					// Get the quantity field for this segment
					attrName = "Hi" + segmentNumber + "qnty";
					if (getAttribute(attrName, rowHiCodes) != null) {
						vHiQnty = (BigDecimal) getAttribute(attrName, rowHiCodes);
					}

					// We start out with the code qualifier and code
					qualAndCode = vHiCdListQual + vHiCd;

					// This code was used for pre-POA processing
					if (!("".equals(qualAndCode))) {
						// If we have a qualifier and code, create the Hi
						// record
						vHiCodeSegment = writeHiRec(vHiCdListQual, vHiCd, vHiCndtnResponseCd, vHiDt, vHiAmt,vHiQnty);

						// Segments 1-4 go onto the e100 record
						if (i <= 4) {
							e100Rec.append(vHiCodeSegment);
							// Segments 5-8 go onto the e105 record
						} else if ((i > 4) && (i <= 8)) {
							e105Rec.append(vHiCodeSegment);
							// Segments 9-12 go onto the e110 record
						} else if (i > 8) {
							e110Rec.append(vHiCodeSegment);
						}

						// Reset variables
						vHiCdListQual = "";
						vHiCd = "";
						vHiDt = "";
						vHiAmt = zeroBD;

					} 
				} // End of looping through the segments

				writeHiRecs(e100Rec.toString(), e105Rec.toString(), e110Rec
						.toString(), vInstFile);
				detach(hiCodes, dbEM);
			} // End of looping through the rows of HI_CODES
			// HRL - MSB Writing the E100 record with just record type value
		} else {
			writeHiRecs("E100", "", "", vInstFile);
		}
	}

  // Return the characters found in the FILE_INFORMATION table for the given claim
  // Only the characters between the "POA" and the "Z" or "X" at the end are returned.
  private char[] getPoa(String poaStr) {
    if (poaStr == null) {
      return new char[] {};
    }

    String rtnVal = null;

    // This pattern excludes the "POA" and the Z or X.
    Pattern p = Pattern.compile(".*POA([Y|N|U|W|1]+)[Z|X].*",
          Pattern.CASE_INSENSITIVE);
    Matcher m = p.matcher(poaStr);

    if (m.matches()) {
      rtnVal = m.group(1);
    }

    return (rtnVal != null) ? rtnVal.toCharArray() : new char[] {};
  }

  

  /**
   * Writes service line level information to the extract file.
   * 
   * @param rowClaim
   * @param vOutputFile
   * @param dbEM
   */
  private void writeFRec(Claim rowClaim, PrintWriter vOutputFile,
      EntityManager dbEM) {
    final String vClaimKey = rowClaim.getClaimKey();
    processF005Rec(vClaimKey, vOutputFile, dbEM);
    vOutputFile.flush();
  } // end writeFRec

  /**
   * Gets institutional service line information and writes it to the extract file.
   * 
   * @param vClaimKey
   *          - the claim key
   * @param vInstFile
   *          - the extract file to write the data to
   * @param dbEM
   */
  private void processF005Rec(String vClaimKey, PrintWriter vInstFile,
      EntityManager dbEM) // , PrintWriter vLogFile)
  {

    InstSvcLine rowInstSvcLine = null;
    Query getInstSvcLineQuery = dbEM.createNativeQuery(
        "SELECT * FROM INST_SVC_LINE WHERE CLAIM_KEY=:claimKey",
        InstSvcLine.class);
    getInstSvcLineQuery.setParameter("claimKey", vClaimKey);
    List instSvcLines = getInstSvcLineQuery.getResultList();
    Iterator instSvcLineIterator = instSvcLines.iterator();

    final ArrayList elementsF005 = new ArrayList(4);
    while (instSvcLineIterator.hasNext()) {
      elementsF005.clear();
      rowInstSvcLine = (InstSvcLine) instSvcLineIterator.next();
      final String vInstKey = rowInstSvcLine.getServiceKey();
      final String vSvcKey = "AND INST_SERVICE_KEY = '" + vInstKey + "' ";
      final String vServiceDt = getDate(vClaimKey, vSvcKey, "472", dbEM);
      final BigDecimal vSvcTaxAmt = getClaimAmt(vClaimKey, vSvcKey, "GT", dbEM);
      final BigDecimal vFacTaxAmt = getClaimAmt(vClaimKey, vSvcKey, "N8", dbEM);
      final String vLineCtrlNo = getRefValue(vClaimKey, "2400", vSvcKey, "6R", dbEM);
      
      if (vServiceDt == null) {
        elementsF005.add(" ");  //from service date index = 0
        elementsF005.add(" ");  //to service date index = 1
      } else {
        elementsF005.add(vServiceDt.substring(0, 8));
        if (vServiceDt.length() > 16) {
          elementsF005.add(vServiceDt.substring(9, 17));
        } else {
          elementsF005.add(" ");
        }
      }

      elementsF005.add(vSvcTaxAmt);  //Service Tax Amount index = 2
      elementsF005.add(vFacTaxAmt);  //Facility Tax Amount index = 3
      elementsF005.add(vLineCtrlNo);  //Line Item Control Number index = 4
      
      
      writeF005Rec(rowInstSvcLine, elementsF005, vInstFile);
      processF020Rec(vInstKey, vInstFile, dbEM);
      processF021Rec(vInstKey, vInstFile, dbEM);

      // The following is the code to create F025 Record - Attending
      // Physician Service Line level
      EntityName rowEntityName = null;
      final String vEntWhereClause = "INST_SERVICE_KEY = '" + vInstKey
          + "' " + "AND NM101ENTY_ID_CD = '71' ";
      rowEntityName = getEntityName(vEntWhereClause
          + "AND LOOP_ID = '2420A' ", dbEM);

      if (rowEntityName != null) {
        writePhyRec(rowEntityName, " ", "F025", vInstFile);
      }

      // The following is the code to create F040 Record - OHI Service
      // Line level Info
      //final String vInstDtKey = "INST_SERVICE_KEY = '" + vInstKey + "' ";
      processF040Rec(vClaimKey, vInstKey, vInstFile, dbEM);

     
      detach(rowEntityName, dbEM);
    }

    detach(instSvcLines, dbEM);
  }

  /**
   * gets and writes Health Care Pricing information to the extract file
   * 
   * @param vSvcKey
   *          - the service line key
   * @param vInstFile
   *          - the extract file to write the data to
   * @param dbEM
   */
  private void processF020Rec(String vSvcKey, PrintWriter vInstFile,
      EntityManager dbEM) {
    // The following is the code to create F020 Record - Health Care
    // Pricing/Repricing
    ClaimPrcReprc rowClaimPrcReprc = null;
    Query getClaimPrcReprcQuery = dbEM
        .createNativeQuery(
            "SELECT * FROM CLAIM_PRC_REPRC WHERE INST_SERVICE_KEY=:instServiceKey",
            ClaimPrcReprc.class);
    getClaimPrcReprcQuery.setParameter("instServiceKey", vSvcKey);
    List claimPrcReprcs = getClaimPrcReprcQuery.getResultList();
    Iterator claimPrcReprcIterator = claimPrcReprcs.iterator();

    if (claimPrcReprcIterator.hasNext()) {
      rowClaimPrcReprc = (ClaimPrcReprc) claimPrcReprcIterator.next();
      writeF020Rec(rowClaimPrcReprc, vInstFile);
    }
    detach(claimPrcReprcs, dbEM);
  }

  /**
   * Gets and writes information about dispensed drugs to the extract file.
   * 
   * @param vSvcKey
   *          - the service line key
   * @param vOutputFile
   *          - the extract file to write the data to
   * @param dbEM
   */
  private void processF021Rec(String vSvcKey, PrintWriter vOutputFile,
      EntityManager dbEM) {
    // The following is the code to create F021 Record - Dispensing Drug
    String vNdcCd = "";
    final String vRefWhereCls = "INST_SERVICE_KEY = '" + vSvcKey + "' "
        + "AND LOOP_ID = '2410' " + "AND REF01ID_QLFR = 'XZ' ";
    final String vRxNumber = getRefValue(vRefWhereCls, dbEM);

    DrugId rowDrugId = null;
    Query getDrugIdQuery = dbEM.createNativeQuery(
        "SELECT * FROM DRUG_ID WHERE INST_SERVICE_KEY=:instServiceKey",
        DrugId.class);
    getDrugIdQuery.setParameter("instServiceKey", vSvcKey);
    List drugIds = getDrugIdQuery.getResultList();
    Iterator drugIdIterator = drugIds.iterator();

    if (drugIdIterator.hasNext()) {
      rowDrugId = (DrugId) drugIdIterator.next();
      vNdcCd = rowDrugId.getLin03natlDrugCd();
    }

    DrugPricing rowDrugPricing = null;
    Query getDrugPricingQuery = dbEM
        .createNativeQuery(
            "SELECT * FROM DRUG_PRICING WHERE INST_SERVICE_KEY=:instServiceKey",
            DrugPricing.class);
    getDrugPricingQuery.setParameter("instServiceKey", vSvcKey);
    List drugPricings = getDrugPricingQuery.getResultList();
    Iterator drugPricingIterator = drugPricings.iterator();
    if (drugPricingIterator.hasNext()) {
      rowDrugPricing = (DrugPricing) drugPricingIterator.next();
      writeF021Rec(rowDrugPricing, vNdcCd, vRxNumber, vOutputFile);
    }
    detach(drugIds, dbEM);
    detach(drugPricings, dbEM);
  }
  
	
	/**
	 * Drives the processing of F040(OHI Svc Line Info) records both the data
	 * aquisition and writing the data to the extract file.
	 * 
	 * @param vClaimKey
	 * @param vSvcKey
	 * @param vOutputFile
	 * @param dbEM
	 */
	protected void processF040Rec(String vClaimKey, String vPrfsnKey,
			PrintWriter vOutputFile, EntityManager dbEM) {
		// The following is the code to create F040 Record - OHI Service Line
		// level Info
		final String vSvcKey = "INST_SERVICE_KEY = '" + vPrfsnKey
		+ "' ";
		SvcLineAdjdct rowSvcLineAdjdct = null;

		//Query getSvcLineAdjdctQuery = dbEM.createNativeQuery(
		//		"SELECT * FROM SVC_LINE_ADJDCT WHERE " + vSvcKey+ " order by SEQ_NBR", SvcLineAdjdct.class);
		String svcLindAdjudicationQueryString = "SELECT * FROM SVC_LINE_ADJDCT WHERE " + vSvcKey+ " order by SEQ_NBR";
    //summaryLOGGER.info("Query String: " + svcLindAdjudicationQueryString);
		Query getSvcLineAdjdctQuery = dbEM.createNativeQuery(svcLindAdjudicationQueryString, SvcLineAdjdct.class);
		List svcLineAdjdcts = getSvcLineAdjdctQuery.getResultList();
		Iterator svcLineAdjdctIterator = svcLineAdjdcts.iterator();
		String vDate = null;
//TWP: TODO: Richard: Please take a look at this should the service dates/pat remaining amount be specific to an OHI?
//They are not in this code		
		final String vSvcAdjDt = getDate(vClaimKey, "AND " + vSvcKey, "573",dbEM);			
		final BigDecimal vRemainingPatientAmount = getClaimAmt( vClaimKey, "AND " + vSvcKey,"EAF", dbEM);
	
		while (svcLineAdjdctIterator.hasNext()) {
			rowSvcLineAdjdct = (SvcLineAdjdct) svcLineAdjdctIterator.next();
			writeF040Rec(rowSvcLineAdjdct, vSvcAdjDt, vRemainingPatientAmount, vOutputFile);
			Cases rowCases = null;
			Query getCasesQuery = dbEM
					.createNativeQuery("SELECT * FROM CASES WHERE INST_SERVICE_KEY=:prfnServiceKey and LOOP_SEQ_NBR=:loopSeqNbr order by SEQ_NBR",Cases.class);
			getCasesQuery.setParameter("prfnServiceKey", vPrfsnKey);
			getCasesQuery.setParameter("loopSeqNbr", rowSvcLineAdjdct.getSeqNbr()+"");

			List cases = getCasesQuery.getResultList();
			Iterator caseIterator = cases.iterator();
			while (caseIterator.hasNext()) {
				rowCases = (Cases) caseIterator.next();
				writeCasesRec(rowCases, "F050", vOutputFile);				
			}
			detach(cases, dbEM);
		}
		detach(svcLineAdjdcts, dbEM);
	} // end
		

  /**
   * Utility method that allows a caller to get an attribute on an object via java reflection.
   * 
   * @param attrName
   * @param target
   * @return
   */
  private Object getAttribute(String attrName, Object target) {
//    if (logIt) {
//      summaryLOGGER.info ("In getAttribute with attrName:" + attrName + ", target: " + target);
//    }
    Object returnObject = null;
    try {
      Class targetClass = target.getClass();
      String methodName = "get" + attrName;
//      if (logIt) {
//        summaryLOGGER.info ("methodName:" + methodName);
//      }
      Method targetGetMethod = targetClass.getMethod(methodName);
      returnObject = targetGetMethod.invoke(target, (Object[]) null);
      } catch(NoSuchMethodException nsme) {
        //do nothing this is expected.
    } catch (Exception e) {
      throw new RuntimeException(
          "Could not access necessary data via reflection in ExtractInstClaims", e);
    }
    return returnObject;
  }
  
  private BigDecimal getPatAmtPaid(String claimKey, EntityManager dbEM)
  {
	BigDecimal returnValue = null;
  
		
			// Get the batches that have B1 claims in status NEW for this sender
	Query patPaidAmounts = dbEM
			.createNativeQuery("SELECT * FROM HI_CODES WHERE CLAIM_KEY=:claimKey AND " +
					 "((HI01CD_LIST_QUAL = 'BE ' AND HI01PROC_CD = 'FC') OR "+
					 "(HI02CD_LIST_QUAL = 'BE ' AND HI02PROC_CD = 'FC') OR "+
					 "(HI03CD_LIST_QUAL = 'BE ' AND HI03PROC_CD = 'FC') OR "+
					 "(HI04CD_LIST_QUAL = 'BE ' AND HI04PROC_CD = 'FC') OR "+
					 "(HI05CD_LIST_QUAL = 'BE ' AND HI05PROC_CD = 'FC') OR "+
					 "(HI06CD_LIST_QUAL = 'BE ' AND HI06PROC_CD = 'FC') OR "+
					 "(HI07CD_LIST_QUAL = 'BE ' AND HI07PROC_CD = 'FC') OR "+
					 "(HI08CD_LIST_QUAL = 'BE ' AND HI08PROC_CD = 'FC') OR "+ 
					 "(HI09CD_LIST_QUAL = 'BE ' AND HI09PROC_CD = 'FC') OR "+
					 "(HI10CD_LIST_QUAL = 'BE ' AND HI10PROC_CD = 'FC') OR "+
					 "(HI11CD_LIST_QUAL = 'BE ' AND HI11PROC_CD = 'FC') OR "+
					 "(HI12CD_LIST_QUAL = 'BE ' AND HI12PROC_CD = 'FC')) ORDER BY SEQ_NBR", HiCodes.class) ;
					 
					 
	patPaidAmounts.setParameter("claimKey", claimKey);
	// query.setParameter(1, vPbm);
	List patPaidAmountsList = patPaidAmounts.getResultList();
	Iterator patPaidAmountsListIterator = patPaidAmountsList.iterator();
	
	
	if ( patPaidAmountsListIterator.hasNext() ) 
	{
		
		HiCodes hiCodes = (HiCodes) patPaidAmountsListIterator.next();
		if("BE ".equals(hiCodes.getHi01cdListQual()) &&
		   "FC".equals(hiCodes.getHi01procCd()) )
		{
			returnValue = hiCodes.getHi01amount();
		}if("BE ".equals(hiCodes.getHi02cdListQual()) &&
		   "FC".equals(hiCodes.getHi02procCd()) )
		{
			returnValue = hiCodes.getHi02amount();
		}else
		if("BE ".equals(hiCodes.getHi03cdListQual()) &&
		   "FC".equals(hiCodes.getHi03procCd()) )
		{
			returnValue = hiCodes.getHi03amount();
		}else
		if("BE ".equals(hiCodes.getHi04cdListQual()) &&
		   "FC".equals(hiCodes.getHi04procCd()) )
		{
			returnValue = hiCodes.getHi04amount();
		}else
		if("BE ".equals(hiCodes.getHi05cdListQual()) &&
		   "FC".equals(hiCodes.getHi05procCd()) )
		{
			returnValue = hiCodes.getHi05amount();
		}else
		if("BE ".equals(hiCodes.getHi06cdListQual()) &&
		   "FC".equals(hiCodes.getHi06procCd()) )
		{
			returnValue = hiCodes.getHi06amount();
		}else
		if("BE ".equals(hiCodes.getHi07cdListQual()) &&
		   "FC".equals(hiCodes.getHi07procCd()) )
		{
			returnValue = hiCodes.getHi07amount();
		}else
		if("BE ".equals(hiCodes.getHi08cdListQual()) &&
		   "FC".equals(hiCodes.getHi08procCd()) )
		{
			returnValue = hiCodes.getHi08amount();
		}else
		if("BE ".equals(hiCodes.getHi09cdListQual()) &&
		   "FC".equals(hiCodes.getHi09procCd()) )
		{
			returnValue = hiCodes.getHi09amount();
		}else
		if("BE ".equals(hiCodes.getHi10cdListQual()) &&
		   "FC".equals(hiCodes.getHi10procCd()) )
		{
			returnValue = hiCodes.getHi10amount();
		}else
		if("BE ".equals(hiCodes.getHi11cdListQual()) &&
		   "FC".equals(hiCodes.getHi11procCd()) )
		{
			returnValue = hiCodes.getHi11amount();
		}else
		if("BE ".equals(hiCodes.getHi12cdListQual()) &&
		   "FC".equals(hiCodes.getHi12procCd()) )
		{
			returnValue = hiCodes.getHi12amount();
		}							   				   				  			
	}
	return returnValue;
  }
}