package gov.va.med.hac.edi.reconciliation;

import static gov.va.med.hac.edi.reconciliation.Helper.DATA_ERROR;
import static gov.va.med.hac.edi.reconciliation.Helper.RESOLUTION_STATUS_CODE_ERROR;
import gov.va.med.domain.fppsowner.FeeClaimPayments;
import gov.va.med.domain.fppsowner.FeeErrorCode;
import gov.va.med.domain.fppsowner.FeePaymentErrors;

import java.util.List;

import javax.persistence.EntityManager;
import javax.persistence.EntityTransaction;
import javax.persistence.Query;

import org.apache.commons.lang.exception.ExceptionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * 
 * Reconciliation process for Fee Payments that are not 
 * of type 'C' or of type 'L'.  These type of payments 
 * are considered errors.
 *
 */
@SuppressWarnings("unchecked")
public class OtherTypeReconciliation {
   static final Logger LOGGER = LoggerFactory.getLogger(OtherTypeReconciliation.class);
   EntityManager em;
   String username;

   /**
    * The constructor for <code>OtherTypeReconciliation</code>. 
    * 
    * @param em the JPA entity manager.
    * @param username the user named used to connect to the database.
    * 
    */
   public OtherTypeReconciliation(EntityManager em, String username) {
      this.em = em;
      this.username = username;
   }


   final void markNewFeeClaimPaymentsOtherType() {
      java.sql.Date today = new java.sql.Date(new java.util.Date().getTime());

      Query qfec = em.createQuery("select fec from foFeeErrorCode fec where fec.feeErrorCd = :code ");
      qfec.setParameter("code", DATA_ERROR);
      FeeErrorCode fec = (FeeErrorCode) qfec.getSingleResult();

      Query q = em.createNativeQuery("SELECT FEE_CLAIM_PAYMENT_ID FROM FEE_CLAIM_PAYMENTS "
                                   + "WHERE RESOLUTION_STATUS_CODE =  'NEW' "
                                   + "AND TRANSACTION_TYPE_CODE != 'L' "
                                   + "AND TRANSACTION_TYPE_CODE != 'C' ");
      List<String> fcpIds = q.getResultList();

      for (String id : fcpIds) {
         EntityTransaction tx = null;
         FeePaymentErrors fpe = null;
         try {
            tx = em.getTransaction();
            tx.begin();
            // mark non-claim and non-line type payments as errors from VITRIA.
            Query q1 = em.createQuery("select f from foFeeClaimPayments f where f.feeClaimPaymentId = :id");
            q1.setParameter("id", id);
            FeeClaimPayments fcp = (FeeClaimPayments) q1.getSingleResult();
            fpe = new FeePaymentErrors();
            fpe.setFeeErrorCode(fec);
            fpe.setCreatedBy(username);
            fpe.setDateCreated(today);
            fpe.setErrorDate(today);
            fpe.setFeeClaimPayments(fcp);
            fcp.getFeePaymentErrors().add(fpe);

            // update fee claim payment status
            fcp.setResolutionStatusCode(RESOLUTION_STATUS_CODE_ERROR);
            tx.commit();
            LOGGER.debug("Payment Error Recorded for Claim: {} UNKNOWN TRANSACTION CODE: {} "
                  , id, fcp.getTransactionTypeCode());
         } catch (Exception e) {
            Throwable t =  ExceptionUtils.getRootCause(e);
            LOGGER.warn("Unable to update payment: {} due to {} "
                  , id, t == null ? e.getMessage() : t.getMessage());
            if (tx != null) {
               tx.rollback();
            }
            markAsFailure(id);
         }
      }

      LOGGER.info("Number of Other Type Payments marked as Errors: {}", fcpIds.size());
   }

   final void markAsFailure(String id) {
      EntityTransaction tx = null;
      try {
         tx = em.getTransaction();
         tx.begin();
         Query q = em.createQuery("select f from foFeeClaimPayments f where f.feeClaimPaymentId = :id ");
         q.setParameter("id", id);
         FeeClaimPayments fcp = (FeeClaimPayments) q.getSingleResult();
         fcp.setResolutionStatusCode("UPDATE_FAILURE");
         tx.commit();
      } catch (Exception e) {
         if (tx != null) {
            tx.rollback();
         }
         Throwable t =  ExceptionUtils.getRootCause(e);
         LOGGER.warn("Unable to mark payment {} failure due to {} "
                     , id, t == null ? e.getMessage() : t.getMessage());
      }
   }
}

