package gov.va.med.hac.edi.reconciliation;

import gov.va.med.hac.edi.utils.jpa.JPAUtil;

import gov.va.med.hac.edi.reconciliation.LineTypeReconciliation;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Properties;

import javax.naming.ConfigurationException;
import javax.persistence.EntityManager;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.MissingOptionException;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * 
 * 
 *
 */
public class Reconcile {

   static final Logger LOGGER = LoggerFactory.getLogger(Reconcile.class);

   /** private constructor to prevent unneeded instance creation. */
   private Reconcile() { }


   /**
    * This is the entry point for the FEE payment reconciliation 
    * process. 
    * 
    * @param args -h for usage help. -f &lt;configfile&gt; for JPA configuration.
    * 
    */
   public static void main(String[] args) {
     System.out.println("\n\n");
     System.out.println("<<<<<<<<<<<<< Beginning FEE Reconciliation <<<<<<<<<<<<<");
      CommandLineParser parser = new GnuParser();
      Options opts = getCliOptions();

      EntityManager em = null;
      try {
         String helpOpt =   "h";
         String configOpt = "f";

         CommandLine line = parser.parse(opts, args);

         if (line.hasOption(helpOpt)) {
            printUsage(opts);
            System.exit(1);
         }

         File configfile = null;
         if (line.hasOption(configOpt)) {
            configfile = JPAUtil.getJpaConfigFile(line.getOptionValue(configOpt));
         }

         if (configfile != null) {
            em = JPAUtil.getEntityManager(configfile, "reconcile");
         }

         if (em == null) {
            throw new ConfigurationException("Unable to configure Entity Manager.");
         }

         String username = getUserName(configfile);

        System.out.println("Reconciling 'C' type claim payments");
         ClaimTypeReconciliation claimRecon = new ClaimTypeReconciliation(em, username);
         claimRecon.reconcileClaimTypePayments();

        System.out.println("Reconciling 'L' type claim payments");
         LineTypeReconciliation lineRecon = new LineTypeReconciliation(em, username);
         lineRecon.reconcileLineTypePayments();

        System.out.println("Marking other claim payment types as errors");
        OtherTypeReconciliation otherTypeClaims = new OtherTypeReconciliation(em, username);
        otherTypeClaims.markNewFeeClaimPaymentsOtherType();

      } catch (MissingOptionException moe) {
         printUsage(opts);
      } catch (Throwable t) {
         LOGGER.error(t.getMessage(), t);
      } finally {
         if (em != null && em.isOpen()) {
            em.close();
         }
        System.out.println("<<<<<<<<<<<<<<< Ending FEE Reconciliation <<<<<<<<<<<<<<");
      }
   }

   private static void printUsage(Options opts) {
      final int helpwidth = 80;
      HelpFormatter fmtr = new HelpFormatter();
      System.out.println();
      fmtr.printHelp(helpwidth, "Reconcile", "", opts, "", true);
   }


   @SuppressWarnings("static-access")
   static Options getCliOptions() {
      Options options = new Options();

      Option configfile = OptionBuilder.withArgName("file")
                          .isRequired()
                          .hasArg()
                          .withLongOpt("configfile")
                          .withDescription("The EntityManager Configuration file.")
                          .create("f");

      Option help = OptionBuilder.withLongOpt("help")
                    .withDescription("Print this message.")
                    .create("h");

      options.addOption(help);
      options.addOption(configfile);

      return options;
   }

   static String getUserName(File propFile) {
      Properties properties = new Properties();
      String username = "UNKNOWN";
      try {
         properties.load(new FileReader(propFile));
         String uname = properties.getProperty("hibernate.connection.username");
         if (uname != null || "".equals(uname)) {
            username = uname.toUpperCase();
         }
      } catch (FileNotFoundException e) {
         LOGGER.warn("Unable to identify User cannot find config file", e);
      } catch (IOException e) {
         LOGGER.warn("Unable to identify User cannot read config file", e);
      }
      return username;
   }

}

