package gov.va.med.hac.edi.reposload.x12.jpa;

import gov.va.med.hac.edi.utils.jpa.*;

import java.sql.SQLException;
import java.util.*;

import javax.persistence.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import java.io.*;
import gov.va.med.domain.erepos.*;

public class ProcessX12Load extends eReposLoad {
  static final Logger LOGGER = LoggerFactory.getLogger(ProcessX12Load.class);
  static final Logger traceLOGGER = LoggerFactory.getLogger("trace." + ProcessX12Load.class);
  private static final String PROFESSIONAL_CLAIM_TYPE = "005010X222";
  private static final String INSTITUTIONAL_CLAIM_TYPE = "005010X223";
  private static final String DENTAL_CLAIM_TYPE = "005010X224";
  /**
   * processLoad is the top-level method used to drive the loading of data into the x12 tables. The
   * data is loaded from .DAT files in the directory identified by the vFilePath parameter. Each
   * .DAT is a flat file representation of a single database table. The method calls several helper
   * methods to load data one for each .DAT file available. Since this method is used to load
   * Dental/Institutional and Professional claims the claim type must be passed in as an argument to
   * distinguish the type of claims the process should load (e.g. 004010X096A1 = institutional,
   * 004010X098A1= professional, 004010X097A1 = dental). The process uses JPA/Hibernate to perform
   * queries and inserts and it is configured via dbHibernate.properties file in the resources
   * directory.
   * 
   * There are a few things that are important to understand when reading this code. They are listed
   * in the notes below.
   * 
   * Note: Referential integrity is enforced by the db, not by JPA/hibernate. Therefore, missing
   * data might not produce an error until a flush or commit command is issued.
   * 
   * Note: Although this code uses JPA we are mapping foreign keys directly. That is, we are not
   * associating objects with one another. We accomplish this via the DAO subclasses and the
   * mappings found in eReposLoadMappings.xml.
   * 
   * For Example, in the line crNcpdpHeader.setBatchKey(tokens[1]); we are setting the foreign key
   * of the batch object Compare to the usual usage of Hibernate, in which we would set the batch
   * object, as in crNcpdpHeader.setBatch(ncpdpBatch); We have departed from the typical hibernate
   * semantics in keeping with the legacy code.
   * 
   * Please see the technical design document for more information.
   * 
   * @param filePath
   *          - path to the .DAT files
   * @throws LoggingException
   *           - this exception is thrown if there is a problem outside of a database transaction
   * @throws DbTransactionEnderException
   *           - this exception is throw to explicitly signal to the caller - that something went
   *           wrong and a database transaction was rolled back.
   */

  protected void processLoad(String claimType, String vFilePath, EntityManager dbEM)
      throws DbTransactionEnderException {
    boolean exists = false;

    try {
      dbEM.setFlushMode(FlushModeType.COMMIT);

      // Put everything into the jpa session and commit it later once the file has been read
      dbEM.getTransaction().begin();

      // Open the control file to find out what the file number (before the files were split into
      // .DAT files)
      // was - then check to see if the data from this file has already been loaded into
      // the database if it has - warn the user and skip processing the data
      // otherwise process the data
      final IntchgControl voIntchgControl = new IntchgControl();
      final BufferedReader inFile = FileUtils.readFile(vFilePath + "INTCHG_CONTROL.DAT");
      String vLoadRec = "";
      String vX12FileName = "";

      if ((vLoadRec = inFile.readLine()) != null) {
        inFile.close();
        vX12FileName = vLoadRec.substring(0, Math.min(15, vLoadRec.length()));
        if (isDataAlreadyLoaded(dbEM, vX12FileName)) {
          LOGGER.warn("Found a Duplicate X12 File with File Name : " + vX12FileName);
        } else {

          // There should always be a control file, a functional group and transaction sets
          // that describe the data to be loaded - load those structures/ that data.
          // Load interchange/exchange control related data into the INTCHG_CONTROL table
          loadIntchgControl(FileUtils.readFile(vFilePath + "INTCHG_CONTROL.DAT"), dbEM);
          LOGGER.debug("IntchgControl loaded");

          // Load the functional group related into the FUNCTIONAL_GROUP
          loadFunctionalGroup(FileUtils.readFile(vFilePath + "FUNCTIONAL_GROUP.DAT"), dbEM);
          LOGGER.debug("FunctionalGroup loaded");

          // Load the transaction set related into the TXN_SET
          loadTxnSet(FileUtils.readFile(vFilePath + "TXN_SET.DAT"), dbEM);
          LOGGER.debug("TxnSet loaded");
          clearCachedEntities(dbEM);

          // If there is a FOREIGN_CURRENCY.DAT file located in the directory
          // indicated by vFilePath that contains foreign currancy data load it into the
          // FOREIGN_CURRENCY table
          if (new File(vFilePath + "FOREIGN_CURRENCY.DAT").exists()) {
            loadForeignCurrency(FileUtils.readFile(vFilePath + "FOREIGN_CURRENCY.DAT"), dbEM);
            LOGGER.debug("ForeignCurrency loaded");
          }

          // Load data describing the data hierarchy into the HRCHL_LEVEL table
          loadHrchlLevel(FileUtils.readFile(vFilePath + "HRCHL_LEVEL.DAT"), dbEM);
          LOGGER.debug("HrchlLevel loaded");

          // Load information about subscribers into the SUBSCRIBER_INFO table
          loadSubscriberInfo(FileUtils.readFile(vFilePath + "SUBSCRIBER_INFO.DAT"), dbEM);
          LOGGER.debug("SubscriberInfo loaded");
          clearCachedEntities(dbEM);

          // If there is a PATIENT_INFO.DAT file located in the directory
          // indicated by vFilePath that contains patient information load it into the
          // PATIENT_INFO table
          if (new File(vFilePath + "PATIENT_INFO.DAT").exists()) {
            loadPatientInfo(FileUtils.readFile(vFilePath + "PATIENT_INFO.DAT"), dbEM);
            LOGGER.debug("PatientInfo loaded");
          }

          clearCachedEntities(dbEM);
          
          // Load claim data into the CLAIM table.
          loadClaim(FileUtils.readFile(vFilePath + "CLAIM.DAT"), dbEM);
          LOGGER.debug("Claim loaded");

          clearCachedEntities(dbEM);
          
          // If there is a HI_CODES.DAT file located in the directory
          // indicated by vFilePath load the date into the HI_CODES table
          if (new File(vFilePath + "HI_CODES.DAT").exists()) {
            loadHiCodes(FileUtils.readFile(vFilePath + "HI_CODES.DAT"), dbEM);
            LOGGER.debug("HiCodes loaded");
          }

          clearCachedEntities(dbEM);
          
          // If there is a MEDI_INPT_ADJDCT.DAT file located in the directory
          // indicated by vFilePath load the date into the MEDI_INPT_ADJDCT table
          if (new File(vFilePath + "MEDI_INPT_ADJDCT.DAT").exists()) {
            loadMediInptAdjdct(FileUtils.readFile(vFilePath + "MEDI_INPT_ADJDCT.DAT"), dbEM);
            LOGGER.debug("MediInptAdjdct loaded");
          }

          // If there is a MEDI_OUTPT_ADJDCT.DAT file located in the directory
          // indicated by vFilePath load the date into the MEDI_OUTPT_ADJDCT table
          if (new File(vFilePath + "MEDI_OUTPT_ADJDCT.DAT").exists()) {
            loadMediOutptAdjdct(FileUtils.readFile(vFilePath + "MEDI_OUTPT_ADJDCT.DAT"), dbEM);
            LOGGER.debug("MediOutptAdjdct loaded");
          }

          // If there is a OT_INSURANCE_COV.DAT file located in the directory
          // indicated by vFilePath load the date into the OT_INSURANCE_COV table
          if (new File(vFilePath + "OT_INSURANCE_COV.DAT").exists()) {
            loadOtInsuranceCov(FileUtils.readFile(vFilePath + "OT_INSURANCE_COV.DAT"), dbEM);
            LOGGER.debug("OtInsuranceCov loaded");
          }

          clearCachedEntities(dbEM);
          
          // If there is a HOME_HCARE_PLAN_IN.DAT file located in the directory
          // indicated by vFilePath load the date into the HOME_HCARE_PLAN_IN table
          if (new File(vFilePath + "HOME_HCARE_PLAN_IN.DAT").exists()) {
            loadHomeHcarePlanIn(FileUtils.readFile(vFilePath + "HOME_HCARE_PLAN_IN.DAT"), dbEM);
            LOGGER.debug("HomeHcarePlanIn loaded");
          }

          // If there is a CONTRACT_INFO.DAT file located in the directory
          // indicated by vFilePath load the date into the CONTRACT_INFO table
          if (new File(vFilePath + "CONTRACT_INFO.DAT").exists()) {
            loadContractInfo(FileUtils.readFile(vFilePath + "CONTRACT_INFO.DAT"), dbEM);
            LOGGER.debug("HContractInfo loaded");
          }

          // If there is a ORTHODONTIC_INFO.DAT file located in the directory
          // indicated by vFilePath load the date into the ORTHODONTIC_INFO table
          if (new File(vFilePath + "ORTHODONTIC_INFO.DAT").exists()) {
            loadOrthodonticInfo(FileUtils.readFile(vFilePath + "ORTHODONTIC_INFO.DAT"), dbEM);
            LOGGER.debug("OrthodonticInfo loaded");
          }

          clearCachedEntities(dbEM);
          
          // If there is a TOOTH_STATUS.DAT file located in the directory
          // indicated by vFilePath load the date into the TOOTH_STATUS table
          if (new File(vFilePath + "TOOTH_STATUS.DAT").exists()) {
            loadToothStatus(FileUtils.readFile(vFilePath + "TOOTH_STATUS.DAT"), dbEM);
            LOGGER.debug("ToothStatus loaded");
          }

          // If there is a INST_CLAIM_CODE.DAT file located in the directory
          // indicated by vFilePath load the date into the INST_CLAIM_CODE table
          if (new File(vFilePath + "INST_CLAIM_CODE.DAT").exists()) {
            loadInstClaimCode(FileUtils.readFile(vFilePath + "INST_CLAIM_CODE.DAT"), dbEM);
            LOGGER.debug("InstClaimCode loaded");
          }

          // If there is a INST_SVC_LINE.DAT file located in the directory
          // indicated by vFilePath load the date into the INST_SVC_LINE table
          if (new File(vFilePath + "INST_SVC_LINE.DAT").exists()) {
            loadInstSvcLine(FileUtils.readFile(vFilePath + "INST_SVC_LINE.DAT"), dbEM);
            LOGGER.debug("InstSvcLine loaded");
          }

          // If there is a PRFSN_SVC_LINE.DAT file located in the directory
          // indicated by vFilePath load the date into the PRFSN_SVC_LINE table
          if (new File(vFilePath + "PRFSN_SVC_LINE.DAT").exists()) {
            loadPrfsnSvcLine(FileUtils.readFile(vFilePath + "PRFSN_SVC_LINE.DAT"), dbEM);
            LOGGER.debug("PrfsnSvcLine loaded");
          }
          
          clearCachedEntities(dbEM);

          // If there is a DENTAL_SVC_LINE.DAT file located in the directory
          // indicated by vFilePath load the date into the DENTAL_SVC_LINE table
          if (new File(vFilePath + "DENTAL_SVC_LINE.DAT").exists()) {
            loadDentalSvcLine(FileUtils.readFile(vFilePath + "DENTAL_SVC_LINE.DAT"), dbEM);
            LOGGER.debug("DentalSvcLine loaded");
          }

          // If there is a PRV_SPCL_INFO.DAT file located in the directory
          // indicated by vFilePath load the date into the PRV_SPCL_INFO table
          if (new File(vFilePath + "PRV_SPCL_INFO.DAT").exists()) {
            loadPrvSpclInfo(FileUtils.readFile(vFilePath + "PRV_SPCL_INFO.DAT"), dbEM);
            LOGGER.debug("PrvSpclInfo loaded");
          }

          clearCachedEntities(dbEM);
          
          // If there is a ENTITY_NAME.DAT file located in the directory
          // indicated by vFilePath load the date into the ENTITY_NAME table
          if (new File(vFilePath + "ENTITY_NAME.DAT").exists()) {
            loadEntityName(FileUtils.readFile(vFilePath + "ENTITY_NAME.DAT"), dbEM);
            LOGGER.debug("EntityName loaded");
          }

          clearCachedEntities(dbEM);
          
          // If there is a ENTITY_ADTL_ID.DAT file located in the directory
          // indicated by vFilePath load the date into the ENTITY_ADTL_ID table
          if (new File(vFilePath + "ENTITY_ADTL_ID.DAT").exists()) {

            loadEntityAdtlId(FileUtils.readFile(vFilePath + "ENTITY_ADTL_ID.DAT"), dbEM);
            LOGGER.debug("EntityAdtlId loaded");
          }
          clearCachedEntities(dbEM);
          // If there is a ENTITY_ADDRESS.DAT file located in the directory
          // indicated by vFilePath load the date into the ENTITY_ADDRESS table
          if (new File(vFilePath + "ENTITY_ADDRESS.DAT").exists()) {

            loadEntityAddress(FileUtils.readFile(vFilePath + "ENTITY_ADDRESS.DAT"), dbEM);
            LOGGER.debug("EntityAddress loaded");
          }
          clearCachedEntities(dbEM);
          // If there is a CNTC_INFO.DAT file located in the directory
          // indicated by vFilePath load the date into the CNTC_INFO table
          if (new File(vFilePath + "CNTC_INFO.DAT").exists()) {

            loadCntcInfo(FileUtils.readFile(vFilePath + "CNTC_INFO.DAT"), dbEM);
            LOGGER.debug("CntcInfo loaded");
          }
          clearCachedEntities(dbEM);

          // If there is a DEMOGRAPHIC_INFO.DAT file located in the directory
          // indicated by vFilePath load the date into the DEMOGRAPHIC_INFO table
          if (new File(vFilePath + "DEMOGRAPHIC_INFO.DAT").exists()) {

            loadDemographicInfo(FileUtils.readFile(vFilePath + "DEMOGRAPHIC_INFO.DAT"), dbEM);
            LOGGER.debug("DemographicInfo loaded");
          }

          // If there is a SPNL_MPLT_SVC_INFO.DAT file located in the directory
          // indicated by vFilePath load the date into the SPNL_MPLT_SVC_INFO table
          exists = (new File(vFilePath + "SPNL_MPLT_SVC_INFO.DAT")).exists();
          if (exists) {

            loadSpnlMpltSvcInfo(FileUtils.readFile(vFilePath + "SPNL_MPLT_SVC_INFO.DAT"), dbEM);
            LOGGER.debug("SpnlMpltSvcInfo loaded");
          }
          clearCachedEntities(dbEM);
          // If there is a CLAIM_DT.DAT file located in the directory
          // indicated by vFilePath load the date into the CLAIM_DT table
          exists = (new File(vFilePath + "CLAIM_DT.DAT")).exists();
          if (exists) {
            loadClaimDt(FileUtils.readFile(vFilePath + "CLAIM_DT.DAT"), dbEM);
            LOGGER.debug("ClaimDt loaded");
          }
          clearCachedEntities(dbEM);

          // If there is a CLAIM_AMOUNT.DAT file located in the directory
          // indicated by vFilePath load the date into the CLAIM_AMOUNT table
          exists = (new File(vFilePath + "CLAIM_AMOUNT.DAT")).exists();
          if (exists) {

            loadClaimAmount(FileUtils.readFile(vFilePath + "CLAIM_AMOUNT.DAT"), dbEM);
            LOGGER.debug("ClaimAmount loaded");
          }

          // If there is a CLAIM_QUANTITY.DAT file located in the directory
          // indicated by vFilePath load the date into the CLAIM_QUANTITY table
          exists = (new File(vFilePath + "CLAIM_QUANTITY.DAT")).exists();
          if (exists) {

            loadClaimQuantity(FileUtils.readFile(vFilePath + "CLAIM_QUANTITY.DAT"), dbEM);
            LOGGER.debug("ClaimQuantity loaded");
          }

          // If there is a CLAIM_NOTE.DAT file located in the directory
          // indicated by vFilePath load the date into the CLAIM_NOTE table
          exists = (new File(vFilePath + "CLAIM_NOTE.DAT")).exists();
          if (exists) {

            loadClaimNote(FileUtils.readFile(vFilePath + "CLAIM_NOTE.DAT"), dbEM);
            LOGGER.debug("ClaimNote loaded");
          }
          clearCachedEntities(dbEM);

          // If there is a CASES.DAT file located in the directory
          // indicated by vFilePath load the date into the CASES table
          exists = (new File(vFilePath + "CASES.DAT")).exists();
          if (exists) {

            loadCases(FileUtils.readFile(vFilePath + "CASES.DAT"), dbEM);
            LOGGER.debug("Cases loaded");
          }

          // If there is a CLAIM_SUPL_INFO.DAT file located in the directory
          // indicated by vFilePath load the date into the CLAIM_SUPL_INFO table
          exists = (new File(vFilePath + "CLAIM_SUPL_INFO.DAT")).exists();
          if (exists) {

            loadClaimSuplInfo(FileUtils.readFile(vFilePath + "CLAIM_SUPL_INFO.DAT"), dbEM);
            LOGGER.debug("ClaimSuplInfo loaded");
          }

          // If there is a FILE_INFORMATION.DAT file located in the directory
          // indicated by vFilePath load the date into the FILE_INFORMATION table
          exists = (new File(vFilePath + "FILE_INFORMATION.DAT")).exists();
          if (exists) {

            loadFileInformation(FileUtils.readFile(vFilePath + "FILE_INFORMATION.DAT"), dbEM);
            LOGGER.debug("FileInformation loaded");
          }
          clearCachedEntities(dbEM);
          // If there is a HOME_HCARE_INFO.DAT file located in the directory
          // indicated by vFilePath load the date into the HOME_HCARE_INFO table
          exists = (new File(vFilePath + "HOME_HCARE_INFO.DAT")).exists();
          if (exists) {

            loadHomeHcareInfo(FileUtils.readFile(vFilePath + "HOME_HCARE_INFO.DAT"), dbEM);
            LOGGER.debug("HomeHcareInfo loaded");
          }

          // If there is a CRC_SEGMENT.DAT file located in the directory
          // indicated by vFilePath load the date into the CRC_SEGMENT table
          exists = (new File(vFilePath + "CRC_SEGMENT.DAT")).exists();
          if (exists) {

            loadCrcSegment(FileUtils.readFile(vFilePath + "CRC_SEGMENT.DAT"), dbEM);
            LOGGER.debug("CrcSegment loaded");
          }

          // If there is a HCARE_SERVICE_DLVR.DAT file located in the directory
          // indicated by vFilePath load the date into the HCARE_SERVICE_DLVR table
          exists = (new File(vFilePath + "HCARE_SERVICE_DLVR.DAT")).exists();
          if (exists) {

            loadHcareServiceDlvr(FileUtils.readFile(vFilePath + "HCARE_SERVICE_DLVR.DAT"), dbEM);
            LOGGER.debug("HcareServiceDlvr loaded");
          }
          clearCachedEntities(dbEM);

          // If there is a AMBLN_TRNPT_INFO.DAT file located in the directory
          // indicated by vFilePath load the date into the AMBLN_TRNPT_INFO table
          exists = (new File(vFilePath + "AMBLN_TRNPT_INFO.DAT")).exists();
          if (exists) {

            loadAmblnTrnptInfo(FileUtils.readFile(vFilePath + "AMBLN_TRNPT_INFO.DAT"), dbEM);
            LOGGER.debug("AmblnTrnptInfo loaded");
          }

          // If there is a CLAIM_PRC_REPRC.DAT file located in the directory
          // indicated by vFilePath load the date into the CLAIM_PRC_REPRC table
          exists = (new File(vFilePath + "CLAIM_PRC_REPRC.DAT")).exists();
          if (exists) {

            loadClaimPrcReprc(FileUtils.readFile(vFilePath + "CLAIM_PRC_REPRC.DAT"), dbEM);
            LOGGER.debug("ClaimPrcReprc loaded");
          }

          // If there is a DBL_MED_EQP_SVC.DAT file located in the directory
          // indicated by vFilePath load the date into the DBL_MED_EQP_SVC table
          exists = (new File(vFilePath + "DBL_MED_EQP_SVC.DAT")).exists();
          if (exists) {

            loadDblMedEqpSvc(FileUtils.readFile(vFilePath + "DBL_MED_EQP_SVC.DAT"), dbEM);
            LOGGER.debug("DblMedEqpSvc loaded");
          }
          clearCachedEntities(dbEM);

          // If there is a HOME_OXN_THPY_CERT.DAT file located in the directory
          // indicated by vFilePath load the date into the HOME_OXN_THPY_CERT table
          exists = (new File(vFilePath + "HOME_OXN_THPY_CERT.DAT")).exists();
          if (exists) {

            loadHomeOxnThpyCert(FileUtils.readFile(vFilePath + "HOME_OXN_THPY_CERT.DAT"), dbEM);
            LOGGER.debug("HomeOxnThpyCert loaded");
          }

          // If there is a DRUG_ID.DAT file located in the directory
          // indicated by vFilePath load the date into the DRUG_ID table
          exists = (new File(vFilePath + "DRUG_ID.DAT")).exists();
          if (exists) {

            loadDrugId(FileUtils.readFile(vFilePath + "DRUG_ID.DAT"), dbEM);
            LOGGER.debug("DrugId loaded");
          }

          // If there is a DRUG_PRICING.DAT file located in the directory
          // indicated by vFilePath load the date into the DRUG_PRICING table
          exists = (new File(vFilePath + "DRUG_PRICING.DAT")).exists();
          if (exists) {

            loadDrugPricing(FileUtils.readFile(vFilePath + "DRUG_PRICING.DAT"), dbEM);
            LOGGER.debug("DrugPricing loaded");
          }
          clearCachedEntities(dbEM);

          // If there is a TEST_RESULT_MEA.DAT file located in the directory
          // indicated by vFilePath load the date into the TEST_RESULT_MEA table
          exists = (new File(vFilePath + "TEST_RESULT_MEA.DAT")).exists();
          if (exists) {

            loadTestResultMea(FileUtils.readFile(vFilePath + "TEST_RESULT_MEA.DAT"), dbEM);
            LOGGER.debug("TestResultMea loaded");
          }

          // If there is a PURCH_SERVICE_INFO.DAT file located in the directory
          // indicated by vFilePath load the date into the PURCH_SERVICE_INFO table
          exists = (new File(vFilePath + "PURCH_SERVICE_INFO.DAT")).exists();
          if (exists) {

            loadPurchServiceInfo(FileUtils.readFile(vFilePath + "PURCH_SERVICE_INFO.DAT"), dbEM);
            LOGGER.debug("PurchServiceInfo loaded");
          }

          // If there is a DBL_MDCL_EQP_CERT.DAT file located in the directory
          // indicated by vFilePath load the date into the DBL_MDCL_EQP_CERT table
          exists = (new File(vFilePath + "DBL_MDCL_EQP_CERT.DAT")).exists();
          if (exists) {

            loadDblMdclEqpCert(FileUtils.readFile(vFilePath + "DBL_MDCL_EQP_CERT.DAT"), dbEM);
            LOGGER.debug("DblMdclEqpCert loaded");
          }
          
          clearCachedEntities(dbEM);

          // If there is a SUPTNG_DOC.DAT file located in the directory
          // indicated by vFilePath load the date into the SUPTNG_DOC table
          exists = (new File(vFilePath + "SUPTNG_DOC.DAT")).exists();
          if (exists) {

            loadSuptngDoc(FileUtils.readFile(vFilePath + "SUPTNG_DOC.DAT"), dbEM);
            LOGGER.debug("SuptngDoc loaded");
          }

          // If there is a SVC_LINE_ADJDCT.DAT file located in the directory
          // indicated by vFilePath load the date into the SVC_LINE_ADJDCT table
          exists = (new File(vFilePath + "SVC_LINE_ADJDCT.DAT")).exists();
          if (exists) {

            loadSvcLineAdjdct(FileUtils.readFile(vFilePath + "SVC_LINE_ADJDCT.DAT"), dbEM);
            LOGGER.debug("SvcLineAdjdct loaded");
          }

          // If there is a TOOTH_INFORMATION.DAT file located in the directory
          // indicated by vFilePath load the date into the TOOTH_INFORMATION table
          exists = (new File(vFilePath + "TOOTH_INFORMATION.DAT")).exists();
          if (exists) {

            loadToothInformation(FileUtils.readFile(vFilePath + "TOOTH_INFORMATION.DAT"), dbEM);
            LOGGER.debug("ToothInformation loaded");
          }
          
          // If there is a "RMTTNC_DLVRY_MTHD.DAT file located in the directory
          // indicated by vFilePath load the date into the RMTTNC_DLVRY_MTHD.DAT table
          exists = (new File(vFilePath + "RMTTNC_DLVRY_MTHD.DAT")).exists();
          if (exists) {

        	loadRmttncDlvryMthd(FileUtils.readFile(vFilePath + "RMTTNC_DLVRY_MTHD.DAT"), dbEM);
            LOGGER.debug("RmttncDlvryMthd loaded");
          }


          
          clearCachedEntities(dbEM);
          LOGGER.info("Committing the database transaction");

          // Commit the entire load at once
          dbEM.getTransaction().commit();
          LOGGER.info("Commit successful");

          String vFileNbr = "";
          String vClaimCount = "";
          int vFileCountTotal = 0;

          String vTransDate = "";

          // x12 loads are run for a particular type of claim - i.e. load all institutional claims,
          // load all dental claims or load all professional claims -
          // At the end of the load, statistics are written to the process log such as how many
          // claims were loaded
          // in order to query properly - i.e. to figure out how many institutional or dental claims
          // were loaded - we need to know the type of claims that were loaded.

          // This block of logic is run for institutional claims
          if (claimType.equalsIgnoreCase(INSTITUTIONAL_CLAIM_TYPE)) {

            // The file number and count of the # of new claims related to that file should be
            // returned
            Query instClaimQuery = dbEM.createNativeQuery(
                "SELECT DISTINCT FILE_NUMBER FileANbr, COUNT(*) "
                    + "FROM CLAIM "
                    + "WHERE FILE_NUMBER = ? "
                    + "AND CLAIM_TYPE = '"+INSTITUTIONAL_CLAIM_TYPE+"' "
                    + "AND STATUS_CODE = 'NEW' "
                    + "GROUP BY FILE_NUMBER");
            instClaimQuery.setParameter(1, vFileNumber);
            List voInstFileNbr = instClaimQuery.getResultList();
            Iterator voInstFileNbrIterator = voInstFileNbr.iterator();

            // For each file number from which new claims were loaded - log how many claims were
            // loaded from tha file
            while (voInstFileNbrIterator.hasNext()) {
              Object crInstFileNbr = voInstFileNbrIterator.next();
              List crInstFileNbrList = null;

              // TWP: FIX This could be buggy need lists of lists
              if (crInstFileNbr instanceof List) {
                crInstFileNbrList = (List) crInstFileNbr;
              } else {
                crInstFileNbrList = new ArrayList();
                crInstFileNbrList.add(crInstFileNbr);
              }

              vFileNbr = ((Object[]) crInstFileNbrList.get(0))[0].toString();
              vClaimCount = ((Object[]) crInstFileNbrList.get(0))[1].toString();
              vFileCountTotal = voInstFileNbr.size();
              LOGGER.info("Institutional File Count : " + vFileCountTotal);
              LOGGER.info("No. of Institutional Claims for File Number : " + vFileNbr + " are "
                  + vClaimCount);
              vTransDate = getIsa13Dt(vFilePath);
              dbEM.getTransaction().begin();
              insertProcessLog(vClaimCount, vTransDate, claimType, dbEM);
              LOGGER.info("Committing the database transaction with "
                      + "the process log insert for institutional claims");
              dbEM.getTransaction().commit();
              LOGGER.info("Commit successful");
            }
          }

          // This block of logic is run for professional claims
          if (claimType.equalsIgnoreCase(PROFESSIONAL_CLAIM_TYPE)) {

            // For each file number from which new claims were loaded - log how many claims were
            // loaded from tha file
            Query prfsnClaimQuery = dbEM.createNativeQuery(
                "SELECT DISTINCT FILE_NUMBER FileANbr, COUNT(*) "
                    + "FROM CLAIM "
                    + "WHERE FILE_NUMBER = ? "
                    + "AND CLAIM_TYPE = '"+PROFESSIONAL_CLAIM_TYPE+"' "
                    + "AND STATUS_CODE = 'NEW' "
                    + "GROUP BY FILE_NUMBER");
            prfsnClaimQuery.setParameter(1, vFileNumber);
            List voPrfsnFileNbr = prfsnClaimQuery.getResultList();
            Iterator voPrfsnFileNbrIterator = voPrfsnFileNbr.iterator();

            while (voPrfsnFileNbrIterator.hasNext()) {
              List crPrfsnFileNbrList = null;

              // TWP: FIX This could be buggy need lists of lists
              Object crPrfsnFileNbr = voPrfsnFileNbrIterator.next();
              if (crPrfsnFileNbr instanceof List) {
                crPrfsnFileNbrList = (List) crPrfsnFileNbr;
              } else {
                crPrfsnFileNbrList = new ArrayList();
                crPrfsnFileNbrList.add(crPrfsnFileNbr);
              }

              vFileNbr = ((Object[]) crPrfsnFileNbrList.get(0))[0].toString();
              vClaimCount = ((Object[]) crPrfsnFileNbrList.get(0))[1].toString();
              vFileCountTotal = voPrfsnFileNbr.size();
              LOGGER.info("Professional File Count : " + vFileCountTotal);
              LOGGER.info("Count of Professional Claims for File Number " + vFileNbr + ": "
                  + vClaimCount);
              vTransDate = getIsa13Dt(vFilePath);
              dbEM.getTransaction().begin();
              insertProcessLog(vClaimCount, vTransDate, claimType, dbEM);
              LOGGER.info("Committing the database transaction with "
                      + "the process log insert for professional claims");
              dbEM.getTransaction().commit();
              LOGGER.info("Commit successful");
            }
          }

          // This block of logic is run for dental claims
          if (claimType.equalsIgnoreCase(DENTAL_CLAIM_TYPE)) {
            // For each file number from which new claims were loaded - log how many claims were
            // loaded from tha file
            Query dentalClaimQuery = dbEM.createNativeQuery(
                "SELECT DISTINCT FILE_NUMBER FileANbr, COUNT(*) "
                    + "FROM CLAIM "
                    + "WHERE FILE_NUMBER = ? "
                    + "AND CLAIM_TYPE = '"+DENTAL_CLAIM_TYPE+"' "
                    + "AND STATUS_CODE = 'NEW' "
                    + "GROUP BY FILE_NUMBER");
            dentalClaimQuery.setParameter(1, vFileNumber);
            List voDentalFileNbr = dentalClaimQuery.getResultList();
            Iterator voDentalFileNbrIterator = voDentalFileNbr.iterator();
            while (voDentalFileNbrIterator.hasNext()) {

              Object crDentalFileNbr = voDentalFileNbrIterator.next();
              List crDentalFileNbrList = null;
              // TWP: FIX This could be buggy need lists of lists
              if (crDentalFileNbr instanceof List) {
                crDentalFileNbrList = (List) crDentalFileNbr;
              } else {
                crDentalFileNbrList = new ArrayList();
                crDentalFileNbrList.add(crDentalFileNbr);
              }

              vFileNbr = ((Object[]) crDentalFileNbrList.get(0))[0].toString();
              vClaimCount = ((Object[]) crDentalFileNbrList.get(0))[1].toString();
              vFileCountTotal = voDentalFileNbr.size();
              LOGGER.info("Dental File Count : " + vFileCountTotal);
              LOGGER.info("No. of Dental Claims for File Number        : " + vFileNbr + " are "
                  + vClaimCount);
              vTransDate = getIsa13Dt(vFilePath);
              dbEM.getTransaction().begin();
              insertProcessLog(vClaimCount, vTransDate, claimType, dbEM);
              LOGGER.info("Committing the database transaction with "
                  + "the process log insert for dental claims");
              dbEM.getTransaction().commit();
              LOGGER.info("Commit successful");
            }
          }
        }
        deleteFiles(vFilePath);
      }
    } catch (Exception e) {
      LOGGER.error("Rolling back the database transaction");
      if (dbEM != null && dbEM.getTransaction().isActive()) {
        dbEM.getTransaction().rollback();
        LOGGER.info("Rollback successful");
      } else {
        LOGGER.error("Transaction was not active - no rollback");
      }

      throw new DbTransactionEnderException(
          "Rolled Back the Transaction", e);
    }
  }

  private void clearCachedEntities(EntityManager dbEM) {
    dbEM.flush();
    dbEM.clear();
  }

  /*
   * deleteFiles deletes all the .DAT data files recognized by the x12 load from the directory
   * vFilePath. It is normally called after the data has been loaded. It cleans up the data
   * directory so the data will not be loaded again if the program is run successively.
   */
  private void deleteFiles(String vFilePath) {
    FileUtils.deleteFile(vFilePath + "INTCHG_CONTROL.DAT");
    FileUtils.deleteFile(vFilePath + "FUNCTIONAL_GROUP.DAT");
    FileUtils.deleteFile(vFilePath + "TXN_SET.DAT");
    FileUtils.deleteFile(vFilePath + "FOREIGN_CURRENCY.DAT");
    FileUtils.deleteFile(vFilePath + "HRCHL_LEVEL.DAT");
    FileUtils.deleteFile(vFilePath + "SUBSCRIBER_INFO.DAT");
    FileUtils.deleteFile(vFilePath + "PATIENT_INFO.DAT");
    FileUtils.deleteFile(vFilePath + "CLAIM.DAT");
    FileUtils.deleteFile(vFilePath + "HI_CODES.DAT");
    FileUtils.deleteFile(vFilePath + "MEDI_INPT_ADJDCT.DAT");
    FileUtils.deleteFile(vFilePath + "MEDI_OUTPT_ADJDCT.DAT");
    FileUtils.deleteFile(vFilePath + "OT_INSURANCE_COV.DAT");
    FileUtils.deleteFile(vFilePath + "HOME_HCARE_PLAN_IN.DAT");
    FileUtils.deleteFile(vFilePath + "CLAIM_QUANTITY.DAT");
    FileUtils.deleteFile(vFilePath + "CONTRACT_INFO.DAT");
    FileUtils.deleteFile(vFilePath + "ORTHODONTIC_INFO.DAT");
    FileUtils.deleteFile(vFilePath + "TOOTH_STATUS.DAT");
    FileUtils.deleteFile(vFilePath + "INST_CLAIM_CODE.DAT");
    FileUtils.deleteFile(vFilePath + "INST_SVC_LINE.DAT");
    FileUtils.deleteFile(vFilePath + "PRFSN_SVC_LINE.DAT");
    FileUtils.deleteFile(vFilePath + "DENTAL_SVC_LINE.DAT");
    FileUtils.deleteFile(vFilePath + "PRV_SPCL_INFO.DAT");
    FileUtils.deleteFile(vFilePath + "ENTITY_NAME.DAT");
    FileUtils.deleteFile(vFilePath + "ENTITY_ADTL_ID.DAT");
    FileUtils.deleteFile(vFilePath + "ENTITY_ADDRESS.DAT");
    FileUtils.deleteFile(vFilePath + "CNTC_INFO.DAT");
    FileUtils.deleteFile(vFilePath + "DEMOGRAPHIC_INFO.DAT");
    FileUtils.deleteFile(vFilePath + "SPNL_MPLT_SVC_INFO.DAT");
    FileUtils.deleteFile(vFilePath + "CLAIM_DT.DAT");
    FileUtils.deleteFile(vFilePath + "CLAIM_AMOUNT.DAT");
    FileUtils.deleteFile(vFilePath + "CLAIM_NOTE.DAT");
    FileUtils.deleteFile(vFilePath + "CASES.DAT");
    FileUtils.deleteFile(vFilePath + "CLAIM_SUPL_INFO.DAT");
    FileUtils.deleteFile(vFilePath + "FILE_INFORMATION.DAT");
    FileUtils.deleteFile(vFilePath + "HOME_HCARE_INFO.DAT");
    FileUtils.deleteFile(vFilePath + "CRC_SEGMENT.DAT");
    FileUtils.deleteFile(vFilePath + "HCARE_SERVICE_DLVR.DAT");
    FileUtils.deleteFile(vFilePath + "AMBLN_TRNPT_INFO.DAT");
    FileUtils.deleteFile(vFilePath + "CLAIM_PRC_REPRC.DAT");
    FileUtils.deleteFile(vFilePath + "DME_MED_EQP_SVC.DAT");  //TWP: We delete this file but there is no such table in DB ?
    FileUtils.deleteFile(vFilePath + "HOME_OXN_THPY_CERT.DAT");
    FileUtils.deleteFile(vFilePath + "DRUG_PRICING.DAT");
    FileUtils.deleteFile(vFilePath + "DRUG_ID.DAT");
    FileUtils.deleteFile(vFilePath + "TEST_RESULT_MEA.DAT");
    FileUtils.deleteFile(vFilePath + "PURCH_SERVICE_INFO.DAT");
    FileUtils.deleteFile(vFilePath + "DBL_MED_EQP_SVC.DAT");
    FileUtils.deleteFile(vFilePath + "DBL_MDCL_EQP_CERT.DAT");
    FileUtils.deleteFile(vFilePath + "SUPTNG_DOC.DAT");
    FileUtils.deleteFile(vFilePath + "SVC_LINE_ADJDCT.DAT");
    FileUtils.deleteFile(vFilePath + "TOOTH_INFORMATION.DAT");
    FileUtils.deleteFile(vFilePath + "FORM_LQ_ID.DAT");  //TWP This was not loaded?  Nor did we attempt to load it?
    FileUtils.deleteFile(vFilePath + "RMTTNC_DLVRY_MTHD.DAT");
  }

  /*
   * Insert a record into table EDI_PROCESS_LOG and record the types of claims loaded (claimType),
   * the count of loaded claims (vClmCnt), the status as "NEW" and the date the data was loaded.
   */
  private void insertProcessLog(String vClmCnt, String vFileDt,
      String claimType, EntityManager dbEM) throws SQLException {
    final EdiProcessLog crEdiProcessLog = new EdiProcessLog();
    crEdiProcessLog.setEdiProcessId(UUID.randomUUID().toString()
        .toUpperCase());
    crEdiProcessLog.setFileDate(DateUtil.getSqlDate(vFileDt));
    String vClmType = "";
    if (claimType.equalsIgnoreCase(INSTITUTIONAL_CLAIM_TYPE)) {
      vClmType = "INSTITUTIONAL";
    } else if (claimType.equalsIgnoreCase(PROFESSIONAL_CLAIM_TYPE)) {
      vClmType = "PROFESSIONAL";
    } else if (claimType.equalsIgnoreCase(DENTAL_CLAIM_TYPE)) {
      vClmType = "DENTAL";
    }
    crEdiProcessLog.setProcessName(vClmType + " LOAD TO EDI REPOSITORY");
    crEdiProcessLog.setMessage("NUM. OF " + vClmType + " CLAIMS LOADED INTO EDI REPOSITORY");
    crEdiProcessLog.setCounts(new Long(vClmCnt));
    crEdiProcessLog.setClaimType("837");
    crEdiProcessLog.setStatus("NEW");

    dbEM.persist(crEdiProcessLog);
  }

  /**
   * Get the interchange control date from the file INTCHG_CONTROL.DAT located in the directory
   * indicated by vFilePath.
   * 
   * @param vFilePath
   *          - the path to the file INTCHG_CONTROL.DAT file.
   * @return a string representing the date
   * @throws IOException
   */
  private String getIsa13Dt(String vFilePath) throws IOException {
    String vIsa13Dt = "";
    String lRec = "";
    final BufferedReader in = FileUtils.readFile(vFilePath + "INTCHG_CONTROL.DAT");
    if ((lRec = in.readLine()) != null) {
      final String[] tokens = lRec.split("\\^", -1);
      vIsa13Dt = tokens[10];
    }
    in.close();
    return vIsa13Dt;
  }

  /**
   * Returns true if this data has already been loaded - identified by the fileNumber in the
   * INTCHG_CONTROL file. TODO: Write a unit test for this. Change the SQL to SELECT COUNT (*)?
   * 
   * @param dbEM
   *          - the database connection / entity manager
   * @param fileNumber
   *          - the fileNumber that identifies a particular batch
   */
  private boolean isDataAlreadyLoaded(EntityManager dbEM, String fileNumber) {

    String fileNumberTruncated = fileNumber.substring(0, 13);

    // query the database to see if this batch was already loaded
    Query isBatchAlreadyLoadedQuery = dbEM.createQuery("SELECT i"
        + "  FROM gov.va.med.domain.erepos.IntchgControl i"
        + " WHERE i.fileNumber like :fileNumber");
    isBatchAlreadyLoadedQuery.setParameter("fileNumber", "%" + fileNumberTruncated + "%");
    try {
      if (isBatchAlreadyLoadedQuery.getSingleResult() != null) {
        return true;
      }
    } catch (NoResultException e) {
      // Just eat this -- it means the batch has not been loaded before.
    }
    return false;
  }
}