package gov.va.med.hac.edi.utils.jpa;

import java.io.*;

import org.slf4j.*;

// This class collects utility methods used in EDI processes surrounding file
//   manipulations
public class FileUtils 
{
  static final Logger LOGGER = LoggerFactory.getLogger(FileUtils.class);

  // Treats the given string as a file to be read
  public static BufferedReader readFile(String clmIdFilePath)
  throws IOException
  {
    return new BufferedReader( new FileReader(clmIdFilePath)); 
  }

  // Close the given Reader
  public static void closeFile(Reader in)
  throws IOException
  {
    in.close();
  }    
    
  // Delete the file given as a parameter, without logging
  public static boolean deleteFile(String inFileName)
  {
    boolean success = false;
    if ((new File(inFileName)).exists())
    {
      success = (new File(inFileName)).delete();
    }
    
    return success;
  }
  
  // Delete the file given as the second parameter, logging to the first parameter
  public static void deleteFile(final PrintWriter vLogFile, final String inFileName) {
      // Call the other version of deleteFile to carry out the deletion
      final boolean success = deleteFile(inFileName);

      // Log the result
      if (success) {
        vLogFile.println("The file " + inFileName + " has been successfully deleted");
      } else {
        vLogFile.println("Unable to delete " + inFileName);
      }
  }
    
  /**
   * Deletes all the files in the given directory.
   * 
   * @param targetDirectory the directory containing the files to be deleted.
   * 
   * @throws IOException
   */
  public static void deleteFilesInDirectory(File targetDirectory) throws IOException {

    if (targetDirectory.isDirectory()) {

      String[] children = targetDirectory.list();
      for (int i = 0; i < children.length; i++) {
        new File(targetDirectory, children[i]).delete();
      }
    }
  }

  /**
   * This method simply copies the files and folders in a source directory to
   * a target directory. Note: it copies the files recursively.
   * 
   * @param sourceLocation
   *          the directory to copy files from
   * @param targetLocation
   *          the directory to copu files to
   * 
   * @throws Exception
   */
  public static void copyDirectory(File sourceLocation, File targetLocation)
      throws IOException {

    if (sourceLocation.isDirectory()) {
      if (!targetLocation.exists()) {
        targetLocation.mkdir();
      }

      String[] children = sourceLocation.list();
      for (int i = 0; i < children.length; i++) {
        copyDirectory(new File(sourceLocation, children[i]), new File(
            targetLocation, children[i]));
      }
    } else {
      InputStream in = new FileInputStream(sourceLocation);
      OutputStream out = new FileOutputStream(targetLocation);

      // Copy the bits from instream to outstream
      byte[] buf = new byte[1024];
      int len;
      while ((len = in.read(buf)) > 0) {
        out.write(buf, 0, len);
      }
      in.close();
      out.close();
    }
  }

  // copy all Files (but not subdirectories) from the source to the dest
  public static void copyDirFiles(File sourceDir, File destDir)
          throws IOException {
    LOGGER.debug("In copyDirFiles with source " + sourceDir.getAbsolutePath()
            + " and destination " + destDir.getAbsolutePath());
    File[] dirFiles = sourceDir.listFiles();
    LOGGER.debug("Got all files in this directory: " + dirFiles.length);

    for (int i = 0; i < dirFiles.length; i++) {
      File nextFile = dirFiles[i];
      LOGGER.debug("Considering file: " + nextFile.getName());
      if (nextFile.isFile()) {
        LOGGER.debug("This is a file, not a directory, so OK to copy");
        File destFile = new File(destDir, dirFiles[i].getName());
        fastCopyFile(nextFile, destFile);
        LOGGER.debug("Back from fastCopyfile");
      } else {
        LOGGER.debug("This is a directory, not a file, so not copying it");
      }
    }
  }

  static final int BUFF_SIZE = 100000;
  static final byte[] buffer = new byte[BUFF_SIZE];

  public static void fastCopyFile(File sourceFile, File destFile) throws IOException{
     InputStream in = null;
     OutputStream out = null;
     try {
        in = new FileInputStream(sourceFile);
        out = new FileOutputStream(destFile);
        while (true) {
           synchronized (buffer) {
              int amountRead = in.read(buffer);
              if (amountRead == -1) {
                 break;
              }
              out.write(buffer, 0, amountRead);
           }
        }
     } finally {
        if (in != null) {
           in.close();
        }
        if (out != null) {
           out.close();
        }
     }
  }
}