package gov.va.med.hac.edi.utils.jpa;

import java.io.*;
import java.net.URL;
import java.sql.*;
import java.util.Properties;

import javax.naming.ConfigurationException;
import javax.persistence.*;

import org.hibernate.Session;
import org.slf4j.*;

/**
 * Helper class for JPA persistence.
 * 
 */
public class JPAUtil {

  static final Logger summaryLOGGER = LoggerFactory.getLogger(JPAUtil.class);

  // disable unnecessary instantiation
  private JPAUtil() {
  }

  /**
   * Helper method to retrieve the file. First looks for file in path, or assumes that relative or
   * absolute path is given. If not looks for file in class path.
   * 
   * @param filename
   *          the fully qualified (relative or absolute) path name of JPA configuration file.
   * @return the File object containing the configuration information.
   * @throws FileNotFoundException
   *           if the file cannot be found.
   * @throws IllegalArgumentException
   *           on null or empty file name.
   */
  public static File getJpaConfigFile(String filename) throws FileNotFoundException {
    if (filename == null || "".equals(filename)) {
      throw new IllegalArgumentException("null input: filename");
    }

    File configfile = new File(filename);
    if (!configfile.exists()) {
      URL url = ClassLoader.getSystemResource(filename);
      if (url != null) {
        configfile = new File(url.getFile());
      }
    }

    if (!configfile.exists()) {
      throw new FileNotFoundException(
               "Unable to find JPA configuration file: " + filename);
    }
    return configfile;
  }

  /**
   * Returns a JPA entity manager using the properties file name supplied (Note: the properties file
   * must be on the class path. The entity manager is based on the configuration set up in
   * persitence.xml identifed by the name entityManager. See the file for more information -- At the
   * very least, that persistence unit should be configured to store and retrieve the NCPDP<Entity>
   * objects. entityManager.
   * 
   * @param propertiesFileName
   * @param entityManager
   * @return
   * @throws FileNotFoundException
   * @throws IOException
   * @throws ConfigurationException
   */
  public static EntityManager getEntityManager(String propertiesFileName,
      String entityManager) throws FileNotFoundException, IOException,
      ConfigurationException {
    EntityManager em;
    File csvConfigFile = JPAUtil.getJpaConfigFile(propertiesFileName);
    em = JPAUtil.getEntityManager(csvConfigFile, entityManager);
    if (em == null) {
      throw new ConfigurationException(
          "Unable to configure Entity Manager for "
              + propertiesFileName);
    }
    return em;
  }

  /**
   * Creates an EntityManager From a persistence unit, "edi", defined in the persistence.xml file.
   * The config file is expected to contain the username, password, and connection url as these are
   * the most likely to change in each environment and throughout the life of the application.
   * 
   * @param configfile
   *          a properties file containing hibernate.connection.username,
   *          hibernate.connection.password, and hibernate.connection.url properties.
   * @return The entity manager created with the "edi" persistence unit.
   * @throws IOException
   *           on any error in obtaining the configfile.
   */
  public static EntityManager getEntityManager(File configfile) throws IOException {
    if (configfile == null) {
      throw new IllegalArgumentException("configfile was null.");
    }
    Properties p = new Properties();
    p.load(new FileInputStream(configfile));
    
    EntityManagerFactory emf = Persistence.createEntityManagerFactory("edi", p);
    return emf.createEntityManager();
  }

  /**
   * Creates an EntityManager From a persistence unit defined in the persistence.xml file. The
   * config file is expected to contain the username, password, and connection url as these are the
   * most likely to change in each environment and throughout the life of the application. The
   * configuration file can be used to override these values.
   * 
   * @param configfile
   *          a properties file containing hibernate.connection.username,
   *          hibernate.connection.password, and hibernate.connection.url properties.
   * @param String
   *          the name of the persistence unit.
   * @return The entity manager created with the passed in persistence unit.
   * @throws IOException
   *           on any error in obtaining the configfile.
   */
  public static EntityManager getEntityManager(File configfile, String entityManagerName)
      throws IOException {
    if (configfile == null) {
      throw new IllegalArgumentException("configfile was null.");
    }
    Properties p = new Properties();
    p.load(new FileInputStream(configfile));
    
    // Log connection information
    summaryLOGGER.info("Connecting to instance: " + p.getProperty("hibernate.connection.url")
        + ", schema: " + p.getProperty("hibernate.connection.username"));
    
    EntityManagerFactory emf = Persistence.createEntityManagerFactory(entityManagerName, p);
    return emf.createEntityManager();
  }

  /**
   * This method allows for the calling of Stored Procedures. This SQL statement should be in the
   * form of "{call &lt;procedure&gt;(&lt;parameters&gt;)}". For instance a procedure called A in
   * the FPPS_OWNER schema taking two parameters will look like this: "{call fpps_owner.a(?1, ?2)}"
   * with two parameters in the parameter section. Note that the parameters should be index as such,
   * the store procedure should not return any values and throw an exception on failure.
   * 
   * @param em
   *          The EntityManager
   * @param sqlStmt
   *          the statement calling the database stored procedure.
   * @param parameters
   *          the parameters to pass the stored procedure. There should be one parameter for every
   *          indexed (?x) parameter in the procedure.
   * @throws Exception
   *           on failure in calling or the execution of the stored procedure.
   */
  @SuppressWarnings("deprecation")
  public static void runDbProcedure (EntityManager em, String sqlStmt, Object... parameters) throws Exception {
	  
	boolean DEBUG = true;  
	  
    Session s = (Session) em.getDelegate();
    
    if(DEBUG) {
    	System.out.println("runDbProcedure() - sesssion connected is:  " + s.isConnected());
    }
    
    
    Connection conn = s.connection();
    
    if(DEBUG) {
    	System.out.println("runDbProcedure() conn is: " + conn.toString());
    }
    
    if(DEBUG) {
    	System.out.println("runDbProcedure() statement is: " + sqlStmt);
    }
    
    
    CallableStatement stmt = conn.prepareCall(sqlStmt);
    
/**
 * Caused by: java.sql.SQLException: ORA-01017: invalid username/password; logon denied    
 */
    if(DEBUG) {
    	System.out.println("runDbProcedure() stmt is: " + sqlStmt);
    }
    
    int count = 1;
    
    for (Object o : parameters) {
      stmt.setObject(count, o);
      count++;
    }
    
    if(DEBUG) {
    	System.out.println("runDbProcedure() - done parsing paramters");
    }
    
    
    try {
    	if(DEBUG) {
    		System.out.println("runDbProcedure(): running statement");
    	}
 

    	stmt.execute();	// takes forever to return from Oracle. If you are using the same login creds. with Toad, turn off Toad
    	
    	if(DEBUG) {
    		System.out.println("runDbProcedure(): DONE running statement");
    	}
    	
    	
    }catch(Exception e) {
    	System.out.println("runDbProcedure() ERROR - : " + e.getMessage() );
    	e.printStackTrace();
    }
    
    if(DEBUG) {
    	System.out.println("runDbProcedure() - done executing");
    }
    
  } // end runDbProcedure()

}	// end class
